#!/usr/bin/python
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""This script performs the node configuration
"""

import os
import re
import sys
import stat
import time
import traceback
import shutil
import shlex

import vi_utils
import vi_os
import vi_xml
import storage_tech

from optparse import OptionParser
from optparse import OptionGroup

####################################################################################################

class work_set_OBJ:

	def __init__( self, name, configFcn, requirementFailedFcn, fileInfo_DICT, requirements_LIST ):
		self.mName					= name
		self.mConfigFcn				= configFcn
		self.mRequirementFailedFcn	= requirementFailedFcn
		self.mFileInfo_DICT			= fileInfo_DICT
		self.mRequirements_LIST		= requirements_LIST
		self.mPulledSuccessFully	= False

	def __repr__(self):
		return "N: %s C: %s F: %s R: %s" % (
				self.mName
			,	self.mConfigFcn
			,	self.mFileInfo_DICT
			,	self.mRequirements_LIST )

####################################################################################################

global_errAndLog		= None		# VI_ErrAndLog_OBJ
global_nicCfgSuccessLog	= None		# VI_PrintRedirector_OBJ
global_aoeNIC_LIST		= []

global_dom0VersionXMLPath			= '/etc/node_configuration_version.xml'

global_nodeCfgLimitsAndMemoryPath	= '/etc/node_config_limits_and_memory'

####################################################################################################
def create_particular_NIC_config_log( vi_errorAndLog_obj, nic_state, configuration_type ):
	"""
	This routine
	"""
#
	particularNICcfgLogPath	= "/var/log/NIC_%s.log" % ( (nic_state.mMAC).replace( ':', '_' ) )

	particularNICcfgLog	= vi_utils.vi_create_simple_PrintRedirector( particularNICcfgLogPath, True )
	if not particularNICcfgLog:
		return None

	# Add this log to the vi_errorAndLog_obj :: THUS IT MUST BE REMOVED LATER !!!
	vi_errorAndLog_obj.augment_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog', particularNICcfgLog.get_dst( 'log' ) )

	particularNICcfgLog.vi_print( '<NIC_config Interface="%s" MAC="%s" Type="%s">' % \
				(	nic_state.mInterface
				,	nic_state.mMAC
				,	configuration_type ) )
	particularNICcfgLog.vi_print( '<NIC_config_run><![CDATA[\n' )

	return particularNICcfgLog

####################################################################################################
def get_trimmed_xml( vi_errorAndLog_obj, xmlFilePath, trimTag, asBufferVSasLIST ):
	"""
	This routine
	"""
#
	xmlLines_LIST, read_failed = vi_utils.vi_read_text_lines( vi_errorAndLog_obj, xmlFilePath )
	if read_failed:
		vi_errorAndLog_obj.vi_print_err( "Reading '%s' failed" % ( xmlFilePath ) )
		return None

	begTag_cmpldRegEx	= re.compile( "^<%s>\s*$" % ( trimTag ) )
	endTag_cmpldRegEx	= re.compile( "^</%s>\s*$" % ( trimTag ) )
	begIdx				= -1
	endIdx				= -1
	xmlLinesLen			= len(xmlLines_LIST)

	for idx in range( xmlLinesLen ):
		if begTag_cmpldRegEx.match( xmlLines_LIST[idx] ):
			begIdx	= idx
			break

	if begIdx < 0:
		vi_errorAndLog_obj.vi_print_dbg( 0, "XML file '%s' did not contain '%s'" % ( xmlFilePath, trimTag ) )
		return None

	for idx in range( begIdx, xmlLinesLen ):
		if endTag_cmpldRegEx.match( xmlLines_LIST[idx] ):
			endIdx	= idx
			break

	if endIdx < 0:
		vi_errorAndLog_obj.vi_print_err( "XML file '%s' contained beg '%s', but not end" % ( xmlFilePath, trimTag ) )
		return None

	if endIdx < xmlLinesLen:
		del( xmlLines_LIST[ endIdx+1 : ] )

	if begIdx > 0:
		del( xmlLines_LIST[ 0 : begIdx ] )

	if not asBufferVSasLIST:
		return xmlLines_LIST

	xmlLinesBuffer	= '\n'.join( xmlLines_LIST )

	return xmlLinesBuffer

####################################################################################################
def generate_iscsi_initiatorname( vi_errorAndLog_obj, primaryIface ):
	"""
	This routine
	"""
#
	if not primaryIface:
		vi_errorAndLog_obj.vi_print_err( "Bad primaryIface : '%s'" % ( primaryIface ) )
		return None

	interface_DICT	= vi_os.vi_get_NIC_dict( vi_errorAndLog_obj, False, True )
	nic_state		= interface_DICT.get( primaryIface, None )
	if not nic_state:
		vi_errorAndLog_obj.vi_print_err( "Unknown primaryIface : '%s'" % ( primaryIface ) )
		return None

	macMerged				= nic_state.mMAC.replace( ':', '' )

	unpackIPv4_cmpldRegEx	= re.compile( '^(\d+).(\d+).(\d+).(\d+)$' )
	unpackIPv4_match		= unpackIPv4_cmpldRegEx.match( nic_state.mIPv4_addr )
	lastAddrGrp				= unpackIPv4_match.group( 4 )

	initiatorDomain			= '2007-01.com.virtualiron'
	initiatorNum			= '01'
	uniqueInitVal			= "%s-%s" % ( macMerged, lastAddrGrp )

	return "iqn.%s:%s:%s" % ( initiatorDomain, initiatorNum, uniqueInitVal )

####################################################################################################
def consider_configuration_version( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	dom0VersionXMLbuffer	= get_trimmed_xml( vi_errorAndLog_obj, global_dom0VersionXMLPath, 'NodeConfigurationVersion', True )
	if not dom0VersionXMLbuffer:
		vi_errorAndLog_obj.vi_print_err( "failed to extract internal dom0VersionXMLPath : '%s'" % ( global_dom0VersionXMLPath ) )
		return True

	vi_errorAndLog_obj.vi_print_dbg( 5, "Dxml:", dom0VersionXMLbuffer )

	dom0VerExtracted_LIST	= vi_xml.vi_extract_configuration_version_list(	vi_errorAndLog_obj
																		,	dom0VersionXMLbuffer
																		,	vi_xml.vi_get_default_configuration_version_fcn_list()
																		)

	vi_errorAndLog_obj.vi_print_dbg( 4, "Ddir:", dom0VerExtracted_LIST )

	if len( dom0VerExtracted_LIST ) < vi_xml.VI_CONFIGURATION_VERSION__ModelVersion:
		vi_errorAndLog_obj.vi_print_err( "dom0VerExtracted_LIST is wicked short : %s : '%s'" % ( dom0VerExtracted_LIST, dom0VersionXMLbuffer ) )
		return True

	try:
		dom0majorVersion	= int( dom0VerExtracted_LIST[ vi_xml.VI_CONFIGURATION_VERSION__ReleaseVersion ].mMajorVersion )
		dom0minorVersion	= int( dom0VerExtracted_LIST[ vi_xml.VI_CONFIGURATION_VERSION__ReleaseVersion ].mMinorVersion )
	except:
		vi_errorAndLog_obj.vi_print_err( "unexpected elements in dom0VerExtracted_LIST : %s : '%s'" % ( dom0VerExtracted_LIST, dom0VersionXMLbuffer ) )
		return True

	vi_errorAndLog_obj.vi_print_log( "beg :: consider_configuration_version() : %s" % ( dom0VerExtracted_LIST ) )

	nodeCfgVerPath	= work_set.mFileInfo_DICT[ 'node_configuration_version.xml' ]
	if not nodeCfgVerPath:
		return False

	nodeCfgVerXMLbuffer	= get_trimmed_xml( vi_errorAndLog_obj, nodeCfgVerPath, 'NodeConfigurationVersion', True )
	if not nodeCfgVerXMLbuffer:
		return True

	vi_errorAndLog_obj.vi_print_dbg( 5, "Nxml:", nodeCfgVerXMLbuffer )

	nodeCfgVerExtracted_LIST	= vi_xml.vi_extract_configuration_version_list(	vi_errorAndLog_obj
																			,	nodeCfgVerXMLbuffer
																			,	vi_xml.vi_get_default_configuration_version_fcn_list()
																			)

	vi_errorAndLog_obj.vi_print_dbg( 4, "Ndir:", nodeCfgVerExtracted_LIST )

	if len( nodeCfgVerExtracted_LIST ) < vi_xml.VI_CONFIGURATION_VERSION__ModelVersion:
		vi_errorAndLog_obj.vi_print_err( "nodeCfgVerExtracted_LIST is wicked short : %s" % ( nodeCfgVerExtracted_LIST ) )
		return True

	modelVersion	= nodeCfgVerExtracted_LIST[ vi_xml.VI_CONFIGURATION_VERSION__ModelVersion ].mModelVersion
	if modelVersion != str(1):
		vi_errorAndLog_obj.vi_print_err( "modelVersion mismatch : %s : %s" % ( modelVersion, nodeCfgVerExtracted_LIST ) )
		return True

	if len( nodeCfgVerExtracted_LIST ) < vi_xml.VI_CONFIGURATION_VERSION__ReleaseVersion:
		vi_errorAndLog_obj.vi_print_err( "nodeCfgVerExtracted_LIST is short : %s" % ( nodeCfgVerExtracted_LIST ) )
		return True

	try:
		majorVersion	= int( nodeCfgVerExtracted_LIST[ vi_xml.VI_CONFIGURATION_VERSION__ReleaseVersion ].mMajorVersion )
		minorVersion	= int( nodeCfgVerExtracted_LIST[ vi_xml.VI_CONFIGURATION_VERSION__ReleaseVersion ].mMinorVersion )
	except:
		vi_errorAndLog_obj.vi_print_err( "unexpected elements in nodeCfgVerExtracted_LIST : %s" % ( nodeCfgVerExtracted_LIST ) )
		return True

	if majorVersion != dom0majorVersion and majorVersion != 0 and dom0majorVersion != 0:
		vi_errorAndLog_obj.vi_print_err( "majorVersion incompatible : %d : %d : %s" % ( dom0majorVersion, majorVersion, nodeCfgVerExtracted_LIST ) )
		return True

	if minorVersion != dom0minorVersion and minorVersion != 0 and dom0minorVersion != 0:
		vi_errorAndLog_obj.vi_print_err( "minorVersion incompatible : %d : %d : %s" % ( dom0minorVersion, minorVersion, nodeCfgVerExtracted_LIST ) )
		return True

	try:
		shutil.copy( nodeCfgVerPath, '/etc/' )
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/'\n\n%s" % ( nodeCfgVerPath, traceback.format_exc() ) )
		return True

	vi_errorAndLog_obj.vi_print_log( "end :: consider_configuration_version() : %s" % ( nodeCfgVerExtracted_LIST ) )

	return False

####################################################################################################
def get_limits_and_memory_default_list( vi_errorAndLog_obj, options_OBJ ):
	"""
	This routine
	"""
#
	limitsAndMemoryCfgDefault_LIST	= vi_xml.vi_get_limits_and_memory_cfg_each_list()

	theStorageConnectionLimits		= limitsAndMemoryCfgDefault_LIST[ vi_xml.VI_LIMITS_AND_MEMORY_CFG__StorageConnectionLimits ]
	theLVMlimits					= limitsAndMemoryCfgDefault_LIST[ vi_xml.VI_LIMITS_AND_MEMORY_CFG__LVMlimits ]
	theVirtualMachineLimits			= limitsAndMemoryCfgDefault_LIST[ vi_xml.VI_LIMITS_AND_MEMORY_CFG__VirtualMachineLimits ]
	theDom0MemoryCfg				= limitsAndMemoryCfgDefault_LIST[ vi_xml.VI_LIMITS_AND_MEMORY_CFG__Dom0MemoryCfg ]

	theStorageConnectionLimits.set_values(	1024	# mMaxDiskPaths
										,	512		# mMaxDisks
										)

	theLVMlimits.set_values(				200		# mMaxActiveVolumeGroups
										,	200		# mMaxActiveLogicalVolumes
										,	100		# mMaxActiveLogicalVolumesPerVG
										,	2		# mMaxActiveSnapshots
										)

	theVirtualMachineLimits.set_values(		32		# mMaxVirtualMachines
										,	32		# mMaxVNCconnections
										)

	theDom0MemoryCfg.set_values(			50		# mInitialDom0MemoryReserveInMB
										)

	return limitsAndMemoryCfgDefault_LIST

####################################################################################################
def configure_limits_and_memory( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	limitsAndMemoryConfigPath	= work_set.mFileInfo_DICT[ 'limits_and_memory_config.xml' ]
	if not limitsAndMemoryConfigPath:
		return False

	vi_errorAndLog_obj.vi_print_log( "beg :: configure_limits_and_memory()" )

	limitsAndMemoryCfg_LIST		= get_limits_and_memory_default_list( vi_errorAndLog_obj, options_OBJ )

	vi_errorAndLog_obj.vi_print_dbg( 6, "dflt:", limitsAndMemoryCfg_LIST )

	limitsAndMemoryCfgXMLbuffer	= get_trimmed_xml( vi_errorAndLog_obj, limitsAndMemoryConfigPath, 'LimitsAndMemoryCfg', True )
	if not limitsAndMemoryCfgXMLbuffer:
		return True

	vi_errorAndLog_obj.vi_print_dbg( 5, "Nxml:", limitsAndMemoryCfgXMLbuffer )

	limitsAndMemoryCfgExtracted_LIST	= vi_xml.vi_extract_limits_and_memory_cfg_list(	vi_errorAndLog_obj
																					,	limitsAndMemoryCfgXMLbuffer
																					,	vi_xml.vi_get_default_limits_and_memory_cfg_fcn_list()
																					)

	vi_errorAndLog_obj.vi_print_dbg( 4, "Ndir:", limitsAndMemoryCfgExtracted_LIST )

	for limAmemCfg in limitsAndMemoryCfgExtracted_LIST:
		limitsAndMemoryCfg_LIST[ limAmemCfg.GetDirectiveType() ]	= limAmemCfg

	del( limitsAndMemoryCfg_LIST[ 0 ] )	# Remove [0] == None

	vi_errorAndLog_obj.vi_print_dbg( 3, "finl:", limitsAndMemoryCfg_LIST )

	for limAmemCfg in limitsAndMemoryCfg_LIST:
		writePath	= "%s/nodeSetting__%s" % ( global_nodeCfgLimitsAndMemoryPath, limAmemCfg.GetName() )
		writeValue	= '[ ' + str( limAmemCfg ) + ' ]'
		if vi_utils.vi_write_text_value( vi_errorAndLog_obj, writePath, writeValue, False ):
			vi_errorAndLog_obj.vi_print_err( "Failed to write( %s, %s )" % ( writePath, writeValue ) )
			return True

	vi_errorAndLog_obj.vi_print_log( "end :: configure_limits_and_memory()" )

	return False

####################################################################################################
def default_limits_and_memory( vi_errorAndLog_obj, work_set, options_OBJ, un_met_requirements_LIST ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( "beg :: default_limits_and_memory()" )

	limitsAndMemoryCfg_LIST		= get_limits_and_memory_default_list( vi_errorAndLog_obj, options_OBJ )

	del( limitsAndMemoryCfg_LIST[ 0 ] )	# Remove [0] == None

	vi_errorAndLog_obj.vi_print_dbg( 3, "finl:", limitsAndMemoryCfg_LIST )

	for limAmemCfg in limitsAndMemoryCfg_LIST:
		writePath	= "%s/nodeSetting__%s" % ( global_nodeCfgLimitsAndMemoryPath, limAmemCfg.GetName() )
		writeValue	= '[ ' + str( limAmemCfg ) + ' ]'
		if vi_utils.vi_write_text_value( vi_errorAndLog_obj, writePath, writeValue, False ):
			vi_errorAndLog_obj.vi_print_err( "Failed to write( %s, %s )" % ( writePath, writeValue ) )
			return True

	vi_errorAndLog_obj.vi_print_log( "end :: default_limits_and_memory()" )

	return False

####################################################################################################
def do_configure__DHCP( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
	"""
	This routine
	"""
#
	nic_state	= vi_os.vi_get_NIC_state( vi_errorAndLog_obj, True, networkCfgDirective.mInterface )
	if not nic_state:
		vi_errorAndLog_obj.vi_print_err( "Bad interface :: %s" % ( networkCfgDirective ) )
		return True

	timeoutInSecs	= 120 + networkCfgDirective.mAttr_extraTime

	vi_errorAndLog_obj.vi_print_log( "DHCP: %s (%s) %d" % ( nic_state.mInterface, nic_state.mMAC, timeoutInSecs ) )

	particularNICcfgLog	= create_particular_NIC_config_log( vi_errorAndLog_obj, nic_state, 'DHCP' )
	if not particularNICcfgLog:
		vi_errorAndLog_obj.vi_print_err( "Was not able to create config log :: %s :: %s" % ( networkCfgDirective, nic_state ) )
		return True

	while True:

		if vi_os.vi_invoke_dhcp( vi_errorAndLog_obj, nic_state, timeoutInSecs ):
			break

		particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
		particularNICcfgLog.vi_print( '<NIC_config_result>SUCCESS</NIC_config_result>' )
		particularNICcfgLog.vi_print( '</NIC_config>' )

		vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
		del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

		global_nicCfgSuccessLog.vi_print( "configure_NIC__SUCCESS__%s[%s](dhcp)" % ( nic_state.mInterface, nic_state.mMAC ) )

		vi_errorAndLog_obj.vi_print_log( "Started DHCP successfully" )

		return False

	particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
	particularNICcfgLog.vi_print( '<NIC_config_result>FAILURE</NIC_config_result>' )
	particularNICcfgLog.vi_print( '</NIC_config>' )

	vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
	del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

	vi_errorAndLog_obj.vi_print_log( "FAILURE :: DHCP: %s (%s) %d" % ( nic_state.mInterface, nic_state.mMAC, timeoutInSecs ) )

	return True

####################################################################################################
def do_configure__STATIC_IP( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
	"""
	This routine
	"""
#
	nic_state	= vi_os.vi_get_NIC_state( vi_errorAndLog_obj, True, networkCfgDirective.mInterface )
	if not nic_state:
		vi_errorAndLog_obj.vi_print_err( "Bad interface :: %s" % ( networkCfgDirective ) )
		return True

	timeoutInSecs	= 120 + networkCfgDirective.mAttr_extraTime

	vi_errorAndLog_obj.vi_print_log( "StaticIP: %s (%s) %d :: %s" % ( nic_state.mInterface, nic_state.mMAC, timeoutInSecs, networkCfgDirective ) )

	particularNICcfgLog	= create_particular_NIC_config_log( vi_errorAndLog_obj, nic_state, 'StaticIP' )
	if not particularNICcfgLog:
		vi_errorAndLog_obj.vi_print_err( "Was not able to create config log :: %s :: %s" % ( networkCfgDirective, nic_state ) )
		return True

	while True:

		vi_errorAndLog_obj.vi_print_log( "Resetting interface: %s" % ( nic_state.mInterface ) )

		if vi_os.vi_reset_NIC_and_wait_for_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs ):
			break

		cmd_LIST	= [ 'ifconfig', nic_state.mInterface, networkCfgDirective.mIPv4_addr, 'netmask', networkCfgDirective.mIPv4_netmask ]

		vi_errorAndLog_obj.vi_print_log( "Setting StaticIP: [%d] %s" % ( timeoutInSecs, ' '.join( cmd_LIST ) ) )

		stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )
		if exitStatus != 0:

			particularNICcfgLog.vi_print( "ERROR: staticIP ifconfig failed for %s (%s) %s %s : %s" % \
					(	nic_state.mInterface
					,	nic_state.mMAC
					,	networkCfgDirective.mIPv4_addr
					,	networkCfgDirective.mIPv4_netmask
					,	exitStatus ) )
			if len( stdout_LIST ) > 0:
				particularNICcfgLog.vi_print( ''.join( stdout_LIST ) )
			if len( stderr_LIST ) > 0:
				particularNICcfgLog.vi_print( ''.join( stderr_LIST ) )

			# Leave the NIC UP and RUNNING, but clear the IP
			cmd_LIST	= [ 'ifconfig', nic_state.mInterface, 'inet', '0' ]
			vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )
			break

		if vi_os.vi_wait_for_NIC_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs ):
			break

		particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
		particularNICcfgLog.vi_print( '<NIC_config_result>SUCCESS</NIC_config_result>' )
		particularNICcfgLog.vi_print( '</NIC_config>' )

		vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
		del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

		global_nicCfgSuccessLog.vi_print( "configure_NIC__SUCCESS__%s[%s](staticIP)" % ( nic_state.mInterface, nic_state.mMAC ) )

		vi_errorAndLog_obj.vi_print_log( "Set StaticIP successfully" )

		return False

	particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
	particularNICcfgLog.vi_print( '<NIC_config_result>FAILURE</NIC_config_result>' )
	particularNICcfgLog.vi_print( '</NIC_config>' )

	vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
	del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

	vi_errorAndLog_obj.vi_print_log( "FAILURE :: StaticIP: %s (%s) %d :: %s" % ( nic_state.mInterface, nic_state.mMAC, timeoutInSecs, networkCfgDirective ) )

	return True

####################################################################################################
def do_configure__MTU( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
	"""
	This routine
	"""
#
	nic_state	= vi_os.vi_get_NIC_state( vi_errorAndLog_obj, True, networkCfgDirective.mInterface )
	if not nic_state:
		vi_errorAndLog_obj.vi_print_err( "Bad interface :: %s" % ( networkCfgDirective ) )
		return True

	timeoutInSecs	= 120 + networkCfgDirective.mAttr_extraTime

	vi_errorAndLog_obj.vi_print_log( "Setting MTU: %s (%s) %d :: %s" % ( nic_state.mInterface, nic_state.mMAC, timeoutInSecs, networkCfgDirective.mMTU ) )

	particularNICcfgLog	= create_particular_NIC_config_log( vi_errorAndLog_obj, nic_state, 'MTU' )
	if not particularNICcfgLog:
		vi_errorAndLog_obj.vi_print_err( "Was not able to create config log :: %s :: %s" % ( networkCfgDirective, nic_state ) )
		return True

	vi_errorAndLog_obj.vi_print_log( "Resetting interface: %s" % ( nic_state.mInterface ) )

	if not vi_os.vi_reset_NIC_and_wait_for_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs, networkCfgDirective.mMTU ):

		particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
		particularNICcfgLog.vi_print( '<NIC_config_result>SUCCESS</NIC_config_result>' )
		particularNICcfgLog.vi_print( '</NIC_config>' )

		vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
		del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

		global_nicCfgSuccessLog.vi_print( "configure_NIC__SUCCESS__%s[%s](mtu)" % ( nic_state.mInterface, nic_state.mMAC ) )

		vi_errorAndLog_obj.vi_print_log( "Set MTU successfully" )

		return False

	particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
	particularNICcfgLog.vi_print( '<NIC_config_result>FAILURE</NIC_config_result>' )
	particularNICcfgLog.vi_print( '</NIC_config>' )

	vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
	del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

	vi_errorAndLog_obj.vi_print_log( "FAILURE :: MTU: %s (%s) %d :: %s" % ( nic_state.mInterface, nic_state.mMAC, timeoutInSecs, networkCfgDirective.mMTU ) )

	return True

####################################################################################################
def do_configure__AOE( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
	"""
	This routine
	"""
#
	nic_state	= vi_os.vi_get_NIC_state( vi_errorAndLog_obj, True, networkCfgDirective.mInterface )
	if not nic_state:
		vi_errorAndLog_obj.vi_print_err( "Bad interface :: %s" % ( networkCfgDirective ) )
		return True

	timeoutInSecs	= 120 + networkCfgDirective.mAttr_extraTime

	vi_errorAndLog_obj.vi_print_log( "Commencing AOE: %s (%s) %d" % ( nic_state.mInterface, nic_state.mMAC, timeoutInSecs ) )

	particularNICcfgLog	= create_particular_NIC_config_log( vi_errorAndLog_obj, nic_state, 'AOE' )
	if not particularNICcfgLog:
		vi_errorAndLog_obj.vi_print_err( "Was not able to create config log :: %s :: %s" % ( networkCfgDirective, nic_state ) )
		return True

	vi_errorAndLog_obj.vi_print_log( "Resetting interface: %s" % ( nic_state.mInterface ) )

	if not vi_os.vi_reset_NIC_and_wait_for_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs ):

		particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
		particularNICcfgLog.vi_print( '<NIC_config_result>SUCCESS</NIC_config_result>' )
		particularNICcfgLog.vi_print( '</NIC_config>' )

		vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
		del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

		global_nicCfgSuccessLog.vi_print( "configure_NIC__SUCCESS__%s[%s](aoe)" % ( nic_state.mInterface, nic_state.mMAC ) )

		global global_aoeNIC_LIST
		global_aoeNIC_LIST.append( nic_state )

		vi_errorAndLog_obj.vi_print_log( "Commenced AOE successfully" )

		return False

	particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
	particularNICcfgLog.vi_print( '<NIC_config_result>FAILURE</NIC_config_result>' )
	particularNICcfgLog.vi_print( '</NIC_config>' )

	vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
	del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

	vi_errorAndLog_obj.vi_print_log( "FAILURE :: AOE: %s (%s) %d" % ( nic_state.mInterface, nic_state.mMAC, timeoutInSecs ) )

	return True

####################################################################################################
def do_configure__BOND( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
	"""
	This routine
	"""
#
	timeoutInSecs	= 120 + networkCfgDirective.mAttr_extraTime

	vi_errorAndLog_obj.vi_print_log( "Bond: %d :: %s" % ( timeoutInSecs, networkCfgDirective ) )

	nic_state_LIST	= []
	for interface in networkCfgDirective.mInterface_LIST:
		nic_state	= vi_os.vi_get_NIC_state( vi_errorAndLog_obj, True, interface )
		if not nic_state:
			vi_errorAndLog_obj.vi_print_err( "Bad interface :: '%s' :: %s" % ( interface, networkCfgDirective ) )
			return True
		nic_state_LIST.append( nic_state )

	if networkCfgDirective.mBondType == 'Failover':
		bondMode	= 0
	elif networkCfgDirective.mBondType == 'LinkAggregation':
		bondMode	= 1
	elif networkCfgDirective.mBondType == 'LoadBalance':
		bondMode	= 2
	else:
		vi_errorAndLog_obj.vi_print_err( "Bad mBondType :: %s" % ( interface, networkCfgDirective ) )
		return True

	if not networkCfgDirective.mBondTag.startswith( 'bond' ):
		vi_errorAndLog_obj.vi_print_err( "Bad mBondTag :: %s" % ( interface, networkCfgDirective ) )
		return True

	if len( nic_state_LIST ) != 2:
		vi_errorAndLog_obj.vi_print_err( "Bad number of interfaces :: %d :: %s" % ( len( nic_state_LIST ), networkCfgDirective ) )
		return True

	particularNICcfgLogPath	= "/var/log/NIC_BOND__%s.log" % ( networkCfgDirective.mBondTag )

	particularNICcfgLog	= vi_utils.vi_create_simple_PrintRedirector( particularNICcfgLogPath, True )
	if not particularNICcfgLog:
		return None

	# Add this log to the vi_errorAndLog_obj :: THUS IT MUST BE REMOVED LATER !!!
	vi_errorAndLog_obj.augment_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog', particularNICcfgLog.get_dst( 'log' ) )

	particularNICcfgLog.vi_print( '<NIC_BOND_config Bond="%s" Type="%s">' % \
				(	networkCfgDirective.mBondTag
				,	networkCfgDirective.mBondType ) )
	particularNICcfgLog.vi_print( '<NIC_config_run><![CDATA[\n' )

	while True:

		cmd_LIST	=	[	'/etc/nex/scripts/nex-bond-if'
						,	'start'
						,	"bondm=%s"		% ( networkCfgDirective.mBondTag )
						,	"primary=%s"	% ( nic_state_LIST[ 0 ].mInterface )
						,	"backup=%s"		% ( nic_state_LIST[ 1 ].mInterface )
						,	"bondmode=%s"	% ( bondMode )
						]

		vi_errorAndLog_obj.vi_print_log( "Creating Bond: [%d] %s" % ( timeoutInSecs, ' '.join( cmd_LIST ) ) )

		stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )

		if len( stdout_LIST ) > 0:
			particularNICcfgLog.vi_print( ''.join( stdout_LIST ) )
		if len( stderr_LIST ) > 0:
			particularNICcfgLog.vi_print( ''.join( stderr_LIST ) )

		if exitStatus != 0:
			break

		particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
		particularNICcfgLog.vi_print( '<NIC_config_result>SUCCESS</NIC_config_result>' )
		particularNICcfgLog.vi_print( '</NIC_config>' )

		vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
		del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

		vi_errorAndLog_obj.vi_print_log( "Created Bond successfully" )

		return False

	particularNICcfgLog.vi_print( '\n]]></NIC_config_run>' )
	particularNICcfgLog.vi_print( '<NIC_config_result>FAILURE</NIC_config_result>' )
	particularNICcfgLog.vi_print( '</NIC_config>' )

	vi_errorAndLog_obj.remove_dst( vi_utils.VI_ErrAndLog__FOR_FLAG__LOG, 'NICcfgLog' )
	del particularNICcfgLog	# Now that it has been removed from vi_errorAndLog_obj

	vi_errorAndLog_obj.vi_print_log( "FAILURE :: Bond: %d :: %s" % ( timeoutInSecs, networkCfgDirective ) )

	return True

####################################################################################################
def do_configure__ROUTE( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
	"""
	This routine
	"""
#
	cmd_LIST	= [ 'route' ]
	cmd_LIST.extend( shlex.split( networkCfgDirective.mRouteCmd ) )

	timeoutInSecs	= 120 + networkCfgDirective.mAttr_extraTime

	vi_errorAndLog_obj.vi_print_log( "Route: [%d] %s" % ( timeoutInSecs, cmd_LIST ) )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime	= vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )

	if exitStatus == 0:
		vi_errorAndLog_obj.vi_print_log( "ROUTE(sucess):0: %s" % ( ' '.join( cmd_LIST ) ) )
	else:
		vi_errorAndLog_obj.vi_print_log( "ROUTE(failure):%s: %s" % ( exitStatus, ' '.join( cmd_LIST ) ) )

	if len( stdout_LIST ) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ) )

	if len( stderr_LIST ) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ) )

	return False

####################################################################################################
def do_configure__IFCONFIG( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
	"""
	This routine
	"""
#
	cmd_LIST	= [ 'ifconfig' ]
	cmd_LIST.extend( shlex.split( networkCfgDirective.mIfconfigCmd ) )

	timeoutInSecs	= 120 + networkCfgDirective.mAttr_extraTime

	vi_errorAndLog_obj.vi_print_log( "Ifconfig: [%d] %s" % ( timeoutInSecs, cmd_LIST ) )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime	= vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )

	if exitStatus == 0:
		vi_errorAndLog_obj.vi_print_log( "IFCONFIG(sucess):0: %s" % ( ' '.join( cmd_LIST ) ) )
	else:
		vi_errorAndLog_obj.vi_print_log( "IFCONFIG(failure):%s: %s" % ( exitStatus, ' '.join( cmd_LIST ) ) )

	if len( stdout_LIST ) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ) )

	if len( stderr_LIST ) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ) )

	return False

####################################################################################################
def do_configure__ETHTOOL( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
	"""
	This routine
	"""
#
	cmd_LIST	= [ 'ethtool' ]
	cmd_LIST.extend( shlex.split( networkCfgDirective.mEthtoolCmd ) )

	timeoutInSecs	= 120 + networkCfgDirective.mAttr_extraTime

	vi_errorAndLog_obj.vi_print_log( "Ethtool: [%d] %s" % ( timeoutInSecs, cmd_LIST ) )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime	= vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )

	if exitStatus == 0:
		vi_errorAndLog_obj.vi_print_log( "ETHTOOL(sucess):0: %s" % ( ' '.join( cmd_LIST ) ) )
	else:
		vi_errorAndLog_obj.vi_print_log( "ETHTOOL(failure):%s: %s" % ( exitStatus, ' '.join( cmd_LIST ) ) )

	if len( stdout_LIST ) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ) )

	if len( stderr_LIST ) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ) )

	return False

####################################################################################################
def configure_core( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( "beg :: configure_core()" )

	try:
		srcPath	= 'scsi_id.config'
		srcPath	= work_set.mFileInfo_DICT[ srcPath ]
		shutil.copy( srcPath, '/etc/' )
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
		return True

	try:
		srcPath			= 'multipath.conf'
		srcPath			= work_set.mFileInfo_DICT[ srcPath ]
		write_failed	= vi_utils.vi_read_write_text_lines( vi_errorAndLog_obj, srcPath, '/etc/multipath.conf', False, False, True )
		if write_failed:
			return True
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
		return True

	vi_errorAndLog_obj.vi_print_log( "end :: configure_core()" )

	return False

####################################################################################################
def default_core( vi_errorAndLog_obj, work_set, options_OBJ, un_met_requirements_LIST ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( "beg :: default_core()" )

	try:
		srcPath	= '/etc/init.d/default_node_config/scsi_id.config'
		shutil.copy( srcPath, '/etc/' )
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
		return True

	try:
		srcPath	= '/etc/init.d/default_node_config/multipath.conf'
		shutil.copy( srcPath, '/etc/' )
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
		return True

	vi_errorAndLog_obj.vi_print_log( "end :: default_core()" )

	return True	# Prevent further configuration

####################################################################################################
def configure_network( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( "beg :: configure_network()" )

	networkCfgDirectiveFcn_LIST	= vi_xml.vi_get_default_network_directive_fcn_list()
	networkCfgDirectiveFcn_LIST[	vi_xml.VI_NETWORK_DIRECTIVE__DHCP		]	=	do_configure__DHCP
	networkCfgDirectiveFcn_LIST[	vi_xml.VI_NETWORK_DIRECTIVE__STATIC_IP	]	=	do_configure__STATIC_IP
	networkCfgDirectiveFcn_LIST[	vi_xml.VI_NETWORK_DIRECTIVE__MTU		]	=	do_configure__MTU
	networkCfgDirectiveFcn_LIST[	vi_xml.VI_NETWORK_DIRECTIVE__AOE		]	=	do_configure__AOE
	networkCfgDirectiveFcn_LIST[	vi_xml.VI_NETWORK_DIRECTIVE__BOND		]	=	do_configure__BOND
	networkCfgDirectiveFcn_LIST[	vi_xml.VI_NETWORK_DIRECTIVE__ROUTE		]	=	do_configure__ROUTE
	networkCfgDirectiveFcn_LIST[	vi_xml.VI_NETWORK_DIRECTIVE__IFCONFIG	]	=	do_configure__IFCONFIG
	networkCfgDirectiveFcn_LIST[	vi_xml.VI_NETWORK_DIRECTIVE__ETHTOOL	]	=	do_configure__ETHTOOL

	try:
		srcPath	= 'network_config_directives.xml'
		srcPath	= work_set.mFileInfo_DICT[ srcPath ]
		shutil.copy( srcPath, '/etc/' )
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
		return True

	networkCfgDirectivesXMLbuffer	= get_trimmed_xml( vi_errorAndLog_obj, srcPath, 'NetworkCfgDirectives', True )
	if not networkCfgDirectivesXMLbuffer:
		return True

	vi_errorAndLog_obj.vi_print_dbg( 5, "Nxml:", networkCfgDirectivesXMLbuffer )

	networkCfgDirective_LIST	= vi_xml.vi_extract_network_directive_list( vi_errorAndLog_obj, networkCfgDirectivesXMLbuffer, networkCfgDirectiveFcn_LIST )

	vi_errorAndLog_obj.vi_print_dbg( 5, "Ndir:", networkCfgDirective_LIST )

	for idx, networkCfgDirective in enumerate( networkCfgDirective_LIST ):
		if networkCfgDirective.mFcn( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
			# Network configuration is in trouble, but bring up the remaining simple networks anyway
			for idx in range( idx+1, len( networkCfgDirective_LIST ) ):
				networkCfgDirective	= networkCfgDirective_LIST[ idx ]
				directiveType		= networkCfgDirective.GetDirectiveType()
				if directiveType == vi_xml.VI_NETWORK_DIRECTIVE__MTU or directiveType == vi_xml.VI_NETWORK_DIRECTIVE__DHCP or directiveType == vi_xml.VI_NETWORK_DIRECTIVE__STATIC_IP:
					networkCfgDirective.mFcn( vi_errorAndLog_obj, networkCfgDirective, options_OBJ )
			return True

	try:
		srcPath	= 'alt_network_config_directives.xml'
		srcPath	= work_set.mFileInfo_DICT[ srcPath ]

		if not srcPath:		# Was optional and did NOT exist
#			vi_errorAndLog_obj.vi_print_log( "alt_network_config_directives.xml does not exist" )
			vi_errorAndLog_obj.vi_print_log( "end :: configure_network()" )
			return False

		shutil.copy( srcPath, '/etc/' )
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
		return False		# ALT can not prevent configuration from succeeding

	networkCfgDirectivesXMLbuffer	= get_trimmed_xml( vi_errorAndLog_obj, srcPath, 'NetworkCfgDirectives', True )
	if not networkCfgDirectivesXMLbuffer:
		return False		# ALT can not prevent configuration from succeeding

	vi_errorAndLog_obj.vi_print_dbg( 5, "NAxml:", networkCfgDirectivesXMLbuffer )

	networkCfgDirective_LIST	= vi_xml.vi_extract_network_directive_list( vi_errorAndLog_obj, networkCfgDirectivesXMLbuffer, networkCfgDirectiveFcn_LIST )

	vi_errorAndLog_obj.vi_print_dbg( 5, "NAdir:", networkCfgDirective_LIST )

	for networkCfgDirective in networkCfgDirective_LIST:
		if networkCfgDirective.mFcn( vi_errorAndLog_obj, networkCfgDirective, options_OBJ ):
			return False	# ALT can not prevent configuration from succeeding

	vi_errorAndLog_obj.vi_print_log( "end :: configure_network()" )

	return False

####################################################################################################
def configure_iscsi( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( "beg :: configure_iscsi()" )

	srcPath	= 'iscsi_portal_list.xml'
	srcPath	= work_set.mFileInfo_DICT[ srcPath ]

	altSrcPath	= 'alt_iscsi_portal_list.xml'
	altSrcPath	= work_set.mFileInfo_DICT[ altSrcPath ]

	if not altSrcPath:		# Was optional and did NOT exist
		try:
			shutil.copy( srcPath, '/etc/' )
		except:
			vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
			return True
	else:
		vi_errorAndLog_obj.vi_print_log( "Merging iscsi_portal lists" )

		srcXML_LIST		= get_trimmed_xml( vi_errorAndLog_obj, srcPath, 'ISCSIportalList', False )
		if srcXML_LIST and len( srcXML_LIST ) >= 2:
			srcXML_LIST.pop(0)
			srcXML_LIST.pop()
		else:
			srcXML_LIST	= []

		altSrcXML_LIST		= get_trimmed_xml( vi_errorAndLog_obj, altSrcPath, 'ISCSIportalList', False )
		if altSrcXML_LIST and len( altSrcXML_LIST ) >= 2:
			altSrcXML_LIST.pop(0)
			altSrcXML_LIST.pop()
		else:
			altSrcXML_LIST	= []

		mergedISCSIportalList	= vi_utils.vi_create_simple_PrintRedirector( '/etc/iscsi_portal_list.xml', False )
		mergedISCSIportalList.vi_print( '<ISCSIportalList>' )
		mergedISCSIportalList.vi_print( '\n'.join( srcXML_LIST ) )
		mergedISCSIportalList.vi_print( '\n'.join( altSrcXML_LIST ) )
		mergedISCSIportalList.vi_print( '</ISCSIportalList>' )

		mergedISCSIportalList.close_dst_files()
		del mergedISCSIportalList

	try:
		srcPath	= 'iscsid.conf'
		srcPath	= work_set.mFileInfo_DICT[ srcPath ]
		shutil.copy( srcPath, '/etc/iscsi/' )
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/iscsi/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
		return True

	srcPath	= 'initiatorname.iscsi'
	srcPath	= work_set.mFileInfo_DICT[ srcPath ]

	initiatorLines_LIST, read_failed = vi_utils.vi_read_text_lines( vi_errorAndLog_obj, srcPath )
	if read_failed:
		vi_errorAndLog_obj.vi_print_err( "Reading '%s' failed" % ( srcPath ) )
		return True

	iqn_cmpldRegEx			= re.compile( '^iqn.(\S+)$' )
	genRequest_cmpldRegEx	= re.compile( '^<<<<GENERATE__MAC_plus_IPend>>>>$' )
	initiator_cmpldRegEx	= re.compile( '^InitiatorName\s*=\s*(\S+)$' )
	initiator_match			= None
	idx						= 0
	for initLine in initiatorLines_LIST:
		initiator_match		= initiator_cmpldRegEx.match( initLine )
		if initiator_match:
			break
		idx					+= 1

	if not initiator_match:
		vi_errorAndLog_obj.vi_print_err( "Bad initiatorname.iscsi :: No proper InitiatorName" )
		return True

	if not iqn_cmpldRegEx.match( initiator_match.group(1) ):
		if not genRequest_cmpldRegEx.match( initiator_match.group(1) ):
			vi_errorAndLog_obj.vi_print_err( "Bad initiatorname.iscsi :: Bad InitiatorName '%s'" % ( initiator_match.group(1) ) )
			return True

		initiatorLines_LIST			= map( str.strip, initiatorLines_LIST )
		iscsi_initiatorname			= generate_iscsi_initiatorname( vi_errorAndLog_obj, options_OBJ.primaryIface )
		initiatorLines_LIST[ idx ]	= "InitiatorName=%s" % ( iscsi_initiatorname )

		write_failed	= vi_utils.vi_write_text_lines( vi_errorAndLog_obj, srcPath, initiatorLines_LIST, True, False )
		if write_failed:
			vi_errorAndLog_obj.vi_print_err( "Writing '%s' failed" % ( srcPath ) )
			return True
	else:
		iscsi_initiatorname			= initiator_match.group(1)

	vi_errorAndLog_obj.vi_print_log( "iSCSI Initiator Name = %s" % ( iscsi_initiatorname ) )

	try:
		shutil.copy( srcPath, '/etc/iscsi/' )
	except:
		vi_errorAndLog_obj.vi_print_err( "cp '%s' '/etc/iscsi/'\n\n%s" % ( srcPath, traceback.format_exc() ) )
		return True

	vi_errorAndLog_obj.vi_print_log( "end :: configure_iscsi()" )

	return False

####################################################################################################
def configure_aoe( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( "beg :: configure_aoe()" )

	while len(global_aoeNIC_LIST) > 0:

		stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'lsmod' ], 30 )
		if exitStatus != 0:
			vi_errorAndLog_obj.vi_print_err( "lsmod failed !?!\n%s\n%s", ''.join( stdout_LIST ), ''.join( stderr_LIST ) )
			break

		aoe_driver_cmpldRegEx	= re.compile( '^aoe\s+\d+' )
		for line in stdout_LIST:
			if aoe_driver_cmpldRegEx.match( line ):
				vi_errorAndLog_obj.vi_print_wrn( "AOE currently running : skipping further AOE configuration" )
				break
		else:
			aoe_iflist_str	= 'aoe_iflist="%s' % ( global_aoeNIC_LIST[0].mInterface )
			for idx in range( 1, len(global_aoeNIC_LIST) ):
				aoe_iflist_str	+= ' ' + global_aoeNIC_LIST[ idx ].mInterface
			aoe_iflist_str	+= '"'

			timeoutInSecs	= 10 + len(global_aoeNIC_LIST) * 30

			vi_errorAndLog_obj.vi_print_log( "Starting AOE [%d] %s" % ( timeoutInSecs, aoe_iflist_str ) )

			exitStatus	= 0
			stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'modprobe', 'aoe', aoe_iflist_str ], timeoutInSecs )
			if exitStatus != 0:
				vi_errorAndLog_obj.vi_print_err( "modprobe aoe %s failed !?!\n%s\n%s", aoe_iflist_str, ''.join( stdout_LIST ), ''.join( stderr_LIST ) )
				break

			vi_errorAndLog_obj.vi_print_log( "waiting for AOE idle" )
			if storage_tech.udev_idle( vi_errorAndLog_obj, 4, 120 ):
				vi_errorAndLog_obj.vi_print_wrn( "configure_aoe() udev_idle( 4, 120 ) timed out" )

			stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'aoe-interfaces' ], timeoutInSecs )
			if len( stdout_LIST ) > 0:
				vi_errorAndLog_obj.vi_print_log( 'aoe-interfaces: %s' % ( exitStatus ) )
				vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
			else:
				vi_errorAndLog_obj.vi_print_log( 'aoe-interfaces: %s :: Nothing ?!?' % ( exitStatus ) )

			stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'aoe-stat' ], timeoutInSecs )
			if len( stdout_LIST ) > 0:
				vi_errorAndLog_obj.vi_print_log( 'aoe-stat: %s' % ( exitStatus ) )
				vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
			else:
				vi_errorAndLog_obj.vi_print_log( 'aoe-stat: %s :: Nothing ?!?' % ( exitStatus ) )

		break

	vi_errorAndLog_obj.vi_print_log( "end :: configure_aoe()" )

	return False

####################################################################################################
def configure_cleanup_default( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( "beg :: configure_cleanup_default()" )

	if 1:
		try:
			shutil.rmtree( '/etc/init.d/default_node_config', True )
		except:
			vi_errorAndLog_obj.vi_print_err( "rm -rf /etc/init.d/default_node_config Failed\n\n%s" % ( traceback.format_exc() ) )

		try:
			shutil.rmtree( '/etc/init.d/optional_node_config', True )
		except:
			vi_errorAndLog_obj.vi_print_err( "rm -rf /etc/init.d/optional_node_config Failed\n\n%s" % ( traceback.format_exc() ) )

		try:
			shutil.rmtree( '/etc/init.d/http_node_config', True )
		except:
			vi_errorAndLog_obj.vi_print_err( "rm -rf /etc/init.d/http_node_config Failed\n\n%s" % ( traceback.format_exc() ) )

	vi_errorAndLog_obj.vi_print_log( "end :: configure_cleanup_default()" )

	return False

####################################################################################################
def configure_dump_network_state( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( "beg :: configure_dump_network_state()" )

	interface_DICT	= vi_os.vi_get_NIC_dict(   vi_errorAndLog_obj, False, True )
	route_LIST		= vi_os.vi_get_route_list( vi_errorAndLog_obj, True )

	vi_errorAndLog_obj.vi_print_log( "Network Interfaces:" )
	for interface, nic_state in interface_DICT.iteritems():
		vi_errorAndLog_obj.vi_print_log( "    %-8s -> %s: %-16s %-16s" % ( \
												nic_state.mInterface
											,	nic_state.mMAC
											,	nic_state.mIPv4_addr
											,	nic_state.mIPv4_netmask ) )

	vi_errorAndLog_obj.vi_print_log( "Routes:" )
	for route in route_LIST:
		vi_errorAndLog_obj.vi_print_log( '   ', route )

	vi_errorAndLog_obj.vi_print_log( "end :: configure_dump_network_state()" )

	return False

####################################################################################################
def execute_debug_script( vi_errorAndLog_obj, work_set, options_OBJ ):
	"""
	This routine
	"""
#
	debugScriptPath	= work_set.mFileInfo_DICT[ 'debug_script.sh' ]
	if not debugScriptPath:		# Was optional and did NOT exist
#		vi_errorAndLog_obj.vi_print_log( "beg :: execute_debug_script()" )
#		vi_errorAndLog_obj.vi_print_log( "debug_script.sh does not exist" )
#		vi_errorAndLog_obj.vi_print_log( "end :: execute_debug_script()" )
		return False

	vi_errorAndLog_obj.vi_print_log( "beg :: execute_debug_script()" )

	os.chmod( debugScriptPath, ( stat.S_IRWXU | stat.S_IRWXG | stat.S_IROTH ) )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, debugScriptPath )
	if len(stdout_LIST) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
	if len(stderr_LIST) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

	vi_errorAndLog_obj.vi_print_log( "end :: execute_debug_script()" )

	return False

####################################################################################################
def do_node_configuration( vi_errorAndLog_obj, nicCfgSuccessLog, options_OBJ ):
	"""
	This routine
	"""
#
	work_set_LIST	=	[	work_set_OBJ(	'configuration_version'
										,	consider_configuration_version
										,	None
										,	{	'node_configuration_version.xml'	:	False
											}
										,	[]
										)
						,	work_set_OBJ(	'limits_and_memory'
										,	configure_limits_and_memory
										,	  default_limits_and_memory
										,	{	'limits_and_memory_config.xml'	:	True
											}
										,	[	'configuration_version'
											]
										)
						,	work_set_OBJ(	'core'
										,	configure_core
										,	  default_core
										,	{	'scsi_id.config'		:	False
											,	'multipath.conf'		:	False
											}
										,	[	'configuration_version'
											,	'limits_and_memory'
											]
										)
						,	work_set_OBJ(	'network'
										,	configure_network
										,	None
										,	{	'network_config_directives.xml'		:	False
											,	'alt_network_config_directives.xml'	:	True
											}
										,	[	'configuration_version'
											]
										)
						,	work_set_OBJ(	'iscsi'
										,	configure_iscsi
										,	None
										,	{	'iscsi_portal_list.xml'		:	False
											,	'alt_iscsi_portal_list.xml'	:	True
											,	'iscsid.conf'				:	False
											,	'initiatorname.iscsi'		:	False
											}
										,	[	'core'
											#,	'network'	Let iSCSI portion figure out if it has working NICs.
											]
										)
						,	work_set_OBJ(	'aoe'
										,	configure_aoe
										,	None
										,	{}
										,	[	'core'
											,	'network'
											]
										)
						,	work_set_OBJ(	'cleanup_default'
										,	configure_cleanup_default
										,	None
										,	{}
										,	[	'core'	]
										)
						,	work_set_OBJ(	'dump_network_state'
										,	configure_dump_network_state
										,	None
										,	{}
										,	[]
										)
						]

	if options_OBJ.doDebugScript:
		work_set_LIST.append(	work_set_OBJ(	'debugScript'
										,	execute_debug_script
										,	None
										,	{	'debug_script.sh'		:	True
											}
										,	[]
										) )

	for work_set in work_set_LIST:
		work_set.mPulledSuccessFully	= True
		for fileName, isOptional in work_set.mFileInfo_DICT.iteritems():
			if not isOptional:
				work_set.mPulledSuccessFully	= False
				break

	numOfPullErrs	= 0
	while True:
		if options_OBJ.pullURL == '':
			numOfPullErrs	= -1
			vi_errorAndLog_obj.vi_print_log( "No where to pull from : Defaulting" )
			break

		final_pullURL	= vi_os.vi_sanitize_pull_URL( vi_errorAndLog_obj, options_OBJ.pullURL, options_OBJ.primaryIface )
		if not final_pullURL:
			numOfPullErrs	= -2
			vi_errorAndLog_obj.vi_print_log( "Bad URL '%s' : Defaulting" % ( options_OBJ.pullURL ) )
			break

		for work_set in work_set_LIST:

			work_set.mPulledSuccessFully	= True
			for fileName, isOptional in work_set.mFileInfo_DICT.iteritems():

				localFilePath	= vi_os.vi_pull_node_configuration_file(	vi_errorAndLog_obj
																		,	final_pullURL
																		,	options_OBJ.localConfigDirPath
																		,	fileName
																		,	isOptional )

				# Switch the fileInfo from isOptional to the filePath of the local file
				work_set.mFileInfo_DICT[fileName]	= localFilePath

				if not isOptional and not localFilePath:
					vi_errorAndLog_obj.vi_print_err( "Failed to pull required file for %s : %s" % ( work_set.mName, fileName ) )
					work_set.mPulledSuccessFully	 = False
					numOfPullErrs					+= 1
					break

		vi_errorAndLog_obj.vi_print_log( "Pull Complete: %d" % ( numOfPullErrs ) )
		break

	if numOfPullErrs < 0:
		for work_set in work_set_LIST:
			for fileName, isOptional in work_set.mFileInfo_DICT.iteritems():
				work_set.mFileInfo_DICT[fileName]	= None

	configuredSuccessfully_LIST	= []

	numOfConfigErrs	= 0
	for work_set in work_set_LIST:

		un_met_requirements_LIST	= []
		for requirement in work_set.mRequirements_LIST:
			if not requirement in configuredSuccessfully_LIST:
				un_met_requirements_LIST.append( requirement )

		if len(un_met_requirements_LIST) > 0:
			vi_errorAndLog_obj.vi_print_err( "Unexpected unmet requirement for %s : %s" % ( work_set.mName, un_met_requirements_LIST ) )
			if work_set.mRequirementFailedFcn:
				failed	= work_set.mRequirementFailedFcn( vi_errorAndLog_obj, work_set, options_OBJ, un_met_requirements_LIST )
				if not failed:
					vi_errorAndLog_obj.vi_print_dbg( 2, "Work set '%s' now considered successfully configured even though un_met_requirements" % ( work_set.mName ) )
					configuredSuccessfully_LIST.append( work_set.mName )
			continue

		if work_set.mPulledSuccessFully:

			failed	= work_set.mConfigFcn( vi_errorAndLog_obj, work_set, options_OBJ )
			if not failed:
				configuredSuccessfully_LIST.append( work_set.mName )
			else:
				numOfConfigErrs	+= 1

		elif work_set.mRequirementFailedFcn:

			failed	= work_set.mRequirementFailedFcn( vi_errorAndLog_obj, work_set, options_OBJ, [] )
			if not failed:
				vi_errorAndLog_obj.vi_print_dbg( 2, "Work set '%s' now considered successfully configured even though pullFailed" % ( work_set.mName ) )
				configuredSuccessfully_LIST.append( work_set.mName )
			else:
				numOfConfigErrs	+= 1

		else:
			vi_errorAndLog_obj.vi_print_log( "Pull failed for work set '%s' :: configuration impossible" % ( work_set.mName ) )

	vi_errorAndLog_obj.vi_print_log( "Configuration Complete [%d]" % ( numOfConfigErrs ) )

	return numOfConfigErrs

####################################################################################################
def parse_given_args():
	"""
	This routine parses the arguments given to this script
	"""
#
	parser = OptionParser()

	parser.add_option("", "--debugLvl", dest="debugLvl", default=0, type="int",
					help="debugLvl message printing to stderr" )

	parser.add_option("", "--primaryIface", dest="primaryIface", default=None, type="str",
					help="primary network Interface (ie. NIC connected to the MS)" )

	parser.add_option("", "--pullURL", dest="pullURL", default=None, type="str",
					help="URL location to pull configuration files from" )

	parser.add_option("", "--localConfigDirPath", dest="localConfigDirPath", default=None, type="str",
					help="local directory path to perform to configuration from" )

	parser.add_option("", "--extraNICtimeInSecs", dest="extraNICtimeInSecs", default=0, type="int",
					help="extra NIC timeout time in seconds" )

	parser.add_option( "", "--doDebugScript", dest="doDebugScript", default=False, action="store_true",
					help="pull optional debug_script.sh and execute it" )

	options_OBJ, args_LIST = parser.parse_args()

	if 0:
		print >> sys.stderr, options_OBJ
		print >> sys.stderr, args_LIST
		print >> sys.stderr, dir(options_OBJ)

	if options_OBJ.localConfigDirPath and options_OBJ.localConfigDirPath.endswith( '/' ):
		options_OBJ.localConfigDirPath	= options_OBJ.localConfigDirPath[:-1]

	if options_OBJ.pullURL == None:
		print >> sys.stderr, '\nno pullURL\n'
		parser.print_help(sys.stderr)
		print >> sys.stderr, ''
		vi_utils.vi_exit( 1 )

	if not options_OBJ.localConfigDirPath:
		print >> sys.stderr, '\nno localConfigDirPath\n'
		parser.print_help(sys.stderr)
		print >> sys.stderr, ''
		vi_utils.vi_exit( 1 )

	return ( options_OBJ, args_LIST )

####################################################################################################
def init_complex_globals( options_OBJ, args_LIST ):
	"""
	This routine initializes and complex global variables
	"""
#
	global global_errAndLog
	global global_nicCfgSuccessLog

	newLog_FILE = None
	logFilePath = '/var/log/configure_node.log'
	try:
		if os.path.exists( logFilePath ):
			os.unlink( logFilePath )
		newLog_FILE = open( logFilePath, 'a', 0 )		# UNbuffered
	except:
		pass

	errAndLog_dst_DICT =	{	'stdErr'			:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__ERR
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__WRN
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__DBG
														,	sys.stderr
														)
							,	'stdout'			:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__STD
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__LOG
														,	sys.stdout
														)
							}

	if newLog_FILE:
		errAndLog_dst_DICT[ 'log' ]	=	(	vi_utils.VI_ErrAndLog__FOR_FLAG__ALL
										,	newLog_FILE
										)

	global_errAndLog = vi_utils.VI_ErrAndLogRedirector_OBJ( errAndLog_dst_DICT, options_OBJ.debugLvl )

	if newLog_FILE:
		global_errAndLog.vi_print_log( "LOG: %s" % ( logFilePath ) )
	else:
		global_errAndLog.vi_print_wrn( "could not create global log FILE '%s'" % ( logFilePath ) )

	global_nicCfgSuccessLog	= vi_utils.vi_create_simple_PrintRedirector( '/var/log/NIC_config_success.log', False )

####################################################################################################
def main():
	"""
	This routine is the script entry point
	"""
#
	options_OBJ, args_LIST = parse_given_args()

	init_complex_globals( options_OBJ, args_LIST )

	removeConfigDirOnSuccess	= False

	if not os.path.exists( options_OBJ.localConfigDirPath ):
		try:
			os.mkdir( options_OBJ.localConfigDirPath )
		except:
			global_errAndLog.vi_print_err( "mkdir( %s ) failed" % ( options_OBJ.localConfigDirPath ) )
			return 1
		removeConfigDirOnSuccess	= True

	numOfConfigErrs	= do_node_configuration( global_errAndLog, global_nicCfgSuccessLog, options_OBJ )

	if numOfConfigErrs > 0:
		return 1

	if removeConfigDirOnSuccess:
		if vi_utils.vi_delete_directory_including_file_contents( options_OBJ.localConfigDirPath ):
			global_errAndLog.vi_print_wrn( "could not clean configDir '%s'" % ( options_OBJ.localConfigDirPath ) )

	return 0

####################################################################################################
### root of all evil

sys.stdout = os.fdopen( sys.stdout.fileno(), sys.stdout.mode, 0 )	# UNbuffer stdout

exitValue  = main()

sys.exit( exitValue )

# END
####################################################################################################
