#!/usr/bin/python
####################################################################################################
#
#	Author:		Joshua Nicholas, Konrad Rzeszutek
#
"""Script for disconnecting devices from the kernel and cleaning up their many dependencies
"""

import sys
import os
import re
import traceback
import errno
import time

import vi_utils
import vi_os
import storage_tech

from optparse import OptionParser
from optparse import OptionGroup

####################################################################################################

global_logDirectory		= '/var/log/disconnect_device'
global_baseLogFileName	= 'disconnect'
global_pid_file			= '/var/run/disconnect_device.pid'
global_pickle_file		= '/var/tmp/disconnect_device.pkl'

global_errAndLog		= None		# VI_ErrAndLog_OBJ
global_logFilePath		= ''

####################################################################################################

class VG_LV_OBJ:

	def __init__( self, theMapName, theVGname, theLVname, theSpecialType=None ):
		self.mMapName		= theMapName
		self.mVGname		= theVGname
		self.mLVname		= theLVname
		self.mSpecialType	= theSpecialType

	def __repr__( self ):
		return "M: %s VG: %s LV: %s ST: %s" % ( self.mMapName, self.mVGname, self.mLVname, self.mSpecialType )

class DisconnectionSet_OBJ:

	def __init__( self, theDMmap ):

		self.mDMmap			= theDMmap	# DeviceMapper_map_OBJ
		self.mVGLV_LIST		= []		# VG_LV_OBJ
		self.mBlkDev_LIST	= []		# BlkDevInfo_OBJ

	def __repr__( self ):

		return "dm: %s" % ( self.mDMmap )

####################################################################################################
def disconnect_device_mapper_map( vi_errorAndLog_obj, dm_map_name, forceDisconnection, retryCount ):
	"""
	This routine disconnects the device map. You can set forceDisconnection (optional) to
	True which will force the removal of the map (if it is used).
	"""
#
	extraIfZeroCount	= True
	while True:
		if not forceDisconnection:
			cmd_LIST	= [ 'dmsetup', 'remove', dm_map_name ]
		else:
			cmd_LIST	= [ 'dmsetup', 'remove', '--force', dm_map_name ]

		vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST )

		if forceDisconnection:
			itrEnd_ex	= 150
		else:
			itrEnd_ex	=  60

		itr	= 0
		while True:
			time.sleep( 0.1 )
			open_count = storage_tech.get_dm_map_open_count( vi_errorAndLog_obj, dm_map_name )
			if open_count == -1:
				return 0

			if open_count == 0 and extraIfZeroCount:	# In the force case, a "always error" map is 
				extraIfZeroCount	= False				# inserted so that all client I/Os will fail.
				retryCount	+= 1						# However, we still end up waiting for those
				break									# clients to close the file.

			itr	+= 1
			if itr > itrEnd_ex:
				break

		if retryCount < 1:
			break

		vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'sync' ], 60 )
		vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'sleep', '1' ], 3 )	# Force descheduling
		retryCount	-= 1

	if forceDisconnection:
		cmd_LIST	= [ 'dmsetup', 'status', dm_map_name ]
		stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, 20 )
		if len( stdout_LIST ) == 0:
			return 0	# Gone now
		dmMapType	= stdout_LIST[0].split()[ 2 ]
		if dmMapType == 'error':
			return -1	# Users still exist but safe to remove underlying block device

	return 1

####################################################################################################
def find_and_dump_device_mapper_map_users( vi_errorAndLog_obj, dm_map_name ):
	"""
	This routine
	"""
#
	cmd_LIST	= [ 'lsof', '-n', '-P', '-Fp', '--', "/dev/mapper/%s" % ( dm_map_name ) ]

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, 60 )

	pidsUsingDMmap_str	= ' ' + ' '.join( stdout_LIST ) + ' '
	del stdout_LIST
	del stderr_LIST
	pidsUsingDMmap_LIST	= map( str.strip, pidsUsingDMmap_str.split( 'p' ) )
	del pidsUsingDMmap_str

	vi_utils.vi_sort_uniq( pidsUsingDMmap_LIST, True )

	cmd_LIST	= [ 'lsof', '+c', '16', '-n', '-P', '-p', '0' ]
	for pid in pidsUsingDMmap_LIST:
		cmd_LIST[ 6 ]	= pid

		stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, 30 )

		vi_errorAndLog_obj.vi_print_log( "======== BEG %s using '%s'" % ( pid, dm_map_name ) )
		if len(stdout_LIST) > 0:
			vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
			vi_errorAndLog_obj.vi_print_log( "======== END %s" % ( pid ) )

		del stdout_LIST
		del stderr_LIST

	return len(pidsUsingDMmap_LIST)

####################################################################################################
def disconnect_disconnection_set( vi_errorAndLog_obj, disconnectionSet, forceDisconnection ):
	"""
	This routine
	"""
#
	dm_map	= disconnectionSet.mDMmap

	if len( disconnectionSet.mVGLV_LIST ) > 0 and not forceDisconnection:
		for vglv in disconnectionSet.mVGLV_LIST:
			if len( vglv.mSpecialType ) > 0:
				break
			open_count	= storage_tech.get_dm_map_open_count( vi_errorAndLog_obj, vglv.mMapName )
			if open_count > 0:
				vi_errorAndLog_obj.vi_print_log( "Device '%s' has open references : use --force to disconnect" % ( dm_map.mMapName ) )
				return 1

	for vglv in disconnectionSet.mVGLV_LIST:
		vi_errorAndLog_obj.vi_print_dbg( 2, "Attempting to remove vglvMapName '%s'" % ( vglv.mMapName ) )
		retVal	= disconnect_device_mapper_map( vi_errorAndLog_obj, vglv.mMapName, forceDisconnection, 1 )
		if retVal == 0:
			vi_errorAndLog_obj.vi_print_log( "Device '%s' LVM dependency %s removed" % ( dm_map.mMapName, vglv.mMapName ) )
		elif retVal < 0:
			vi_errorAndLog_obj.vi_print_log( "Device '%s' LVM dependency %s replaced with error map but users still exist" % ( dm_map.mMapName, vglv.mMapName ) )
			find_and_dump_device_mapper_map_users( vi_errorAndLog_obj, vglv.mMapName )
		else: # retVal > 0:
			vi_errorAndLog_obj.vi_print_log( "Device '%s' LVM dependency %s could not be removed" % ( dm_map.mMapName, vglv.mMapName ) )
			find_and_dump_device_mapper_map_users( vi_errorAndLog_obj, vglv.mMapName )
			if not forceDisconnection:
				return 1

	for vglv in disconnectionSet.mVGLV_LIST:	# Wipe the vgchange -aly tendrils
		vg_dev_path	= "/dev/%s" % ( vglv.mVGname )
		if not os.path.exists( vg_dev_path ):
			continue
		vi_errorAndLog_obj.vi_print_dbg( 2, "Attempting to remove vg_dev_path '%s'" % ( vg_dev_path ) )
		vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ '/bin/rm', '-fr', vg_dev_path ] )
		vi_utils.vi_exe_sync( vi_errorAndLog_obj, 'sync' )
		if os.path.exists( vg_dev_path ):
			vi_errorAndLog_obj.vi_print_log( "Device '%s' VG dev dependencies could not be removed" % ( dm_map.mMapName ) )
			return 1
		else:
			vi_errorAndLog_obj.vi_print_log( "Device '%s' VG dev dependency %s removed" % ( dm_map.mMapName, vg_dev_path ) )

	vi_errorAndLog_obj.vi_print_dbg( 2, "Attempting to remove disk dm_map '%s'" % ( dm_map.mMapName ) )
	retVal	= disconnect_device_mapper_map( vi_errorAndLog_obj, dm_map.mMapName, forceDisconnection, 2 )
	if retVal == 0:
		vi_errorAndLog_obj.vi_print_log( "Device '%s' removed" % ( dm_map.mMapName ) )
	elif retVal < 0:
		vi_errorAndLog_obj.vi_print_log( "Device '%s' replaced with error map but users still exist" % ( dm_map.mMapName ) )
		find_and_dump_device_mapper_map_users( vi_errorAndLog_obj, dm_map.mMapName )
	else: # retVal > 0:
		vi_errorAndLog_obj.vi_print_log( "Device '%s' could not be removed" % ( dm_map.mMapName ) )
		find_and_dump_device_mapper_map_users( vi_errorAndLog_obj, dm_map.mMapName )
		return 1	# Not safe to remove underlying device in this case

	retVal	= 0
	for blkDevInfo in disconnectionSet.mBlkDev_LIST:

		vi_errorAndLog_obj.vi_print_dbg( 2, "Attempting to disconnect %s blkDev %s" % ( blkDevInfo.mWWID, blkDevInfo.mBlkDev ) )

		if blkDevInfo.mOperations.disconnect_block_device( vi_errorAndLog_obj, forceDisconnection ):
			vi_errorAndLog_obj.vi_print_log( "Device '%s' blkDev %s FAILED to disconnect" % ( blkDevInfo.mWWID, blkDevInfo.mBlkDev ) )
			if retVal == 0:
				retVal	= 2
		else:
			vi_errorAndLog_obj.vi_print_log( "Device '%s' blkDev %s disconnected" % ( blkDevInfo.mWWID, blkDevInfo.mBlkDev ) )

	if retVal:
		return retVal

	return 0

####################################################################################################
def figure_out_disconnection_set_for_WWID( vi_errorAndLog_obj, wwid ):
	"""
	This routine assembles a DisconnectionSet_OBJ which consists of:
		- object.mBlkDev_LIST - list of BlockDevice_OBJ
		- object.mVGLV_LIST   - list of VG_LV_OBJ 
	"""
#
	dm_maps_DICT	= storage_tech.get_device_mapper_maps( vi_errorAndLog_obj, False )
	dm_map			= dm_maps_DICT.get( wwid, None )
	if not dm_map:
		return None

	disconnectionSet	= DisconnectionSet_OBJ( dm_map )

	for dm_map_dep in dm_map.mDep_LIST:

		blkDevInfo	= storage_tech.get_blkDevInfo( vi_errorAndLog_obj, dm_map_dep.mBlkDevName, dm_map.mMapName, False, False )
		if not blkDevInfo:
			vi_errorAndLog_obj.vi_print_log( "Device '%s' blkDev %s FAILED to acquire info" % ( dm_map.mMapName, dm_map_dep.mBlkDevName ) )
			continue

		disconnectionSet.mBlkDev_LIST.append( blkDevInfo )

	lvm_dm_map_uuid_cmpldRegEx  = re.compile( "^([^-]+)-([^-]+)(.*)$" )

	for lvm_dm_map in dm_map.mLVMusing_LIST:

		lvm_dm_map_uuid_match	= lvm_dm_map_uuid_cmpldRegEx.match( lvm_dm_map.mMapName )
		if not lvm_dm_map_uuid_match:
			vi_errorAndLog_obj.vi_print_err( "Unexpected LVM dm map '%s'" % ( lvm_dm_map.mMapName ) )
			return None

		lvm_dm_map_uuid_match_TPL	= lvm_dm_map_uuid_match.groups()
		vgName						= lvm_dm_map_uuid_match_TPL[ 0 ]
		lvName						= lvm_dm_map_uuid_match_TPL[ 1 ]

		if len( lvm_dm_map_uuid_match_TPL ) > 2:
			spType	= ( lvm_dm_map_uuid_match_TPL[ 2 ] )[1:]
		else:
			spType	= ''

		if len( spType ) == 0:
			# ALWAYS keep the non-special types at the front of the list
			disconnectionSet.mVGLV_LIST.insert( 0, VG_LV_OBJ( lvm_dm_map.mMapName, vgName, lvName, spType ) )
		else:
			disconnectionSet.mVGLV_LIST.append( VG_LV_OBJ( lvm_dm_map.mMapName, vgName, lvName, spType ) )

	return disconnectionSet

####################################################################################################
def dump_disconnection_set_info( vi_errorAndLog_obj, disconnectionSet ):
	"""
	This routine
	"""
#
	dm_map	= disconnectionSet.mDMmap
	vi_errorAndLog_obj.vi_print_dbg( 0, "n: %s    #: %s:%s u: %s\n    d: %s" % \
			(	dm_map.mMapName
			,	dm_map.mMajorNum
			,	dm_map.mMinorNum
			,	dm_map.mUUID
			,	dm_map.mDep_LIST ) )

#	for lvm_dm_map in dm_map.mLVMusing_LIST:
#		vi_errorAndLog_obj.vi_print_dbg( 0, "    l: n: %s    #: %s:%s u: %s" % \
#				(	lvm_dm_map.mMapName
#				,	lvm_dm_map.mMajorNum
#				,	lvm_dm_map.mMinorNum
#				,	lvm_dm_map.mUUID ) )

	for vglv in disconnectionSet.mVGLV_LIST:
		vi_errorAndLog_obj.vi_print_dbg( 0, "    vglv: %s" % ( vglv ) )

####################################################################################################
def disconnect_based_on_WWID( vi_errorAndLog_obj, options_OBJ, wwid_LIST ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( 'Considering for disconnect:' )
	disconnectionSet_LIST	= []
	for wwid in wwid_LIST:
		disconnectionSet	= figure_out_disconnection_set_for_WWID( vi_errorAndLog_obj, wwid )
		if disconnectionSet:
			disconnectionSet_LIST.append( disconnectionSet )
			vi_errorAndLog_obj.vi_print_log( "    %s" % ( wwid ) )
		else:
			vi_errorAndLog_obj.vi_print_log( "    %s :: Skipping :: does not exist" % ( wwid ) )

	vi_errorAndLog_obj.vi_print_log( '' )

	if len(disconnectionSet_LIST) == 0:
		return 0, 0

	if options_OBJ.onlyDumpDependencies:
		for disconnectionSet in disconnectionSet_LIST:
			dump_disconnection_set_info( vi_errorAndLog_obj, disconnectionSet )
		return 0, 0

	retVal					= 0
	numOfWWIDdisconnected	= 0
	for disconnectionSet in disconnectionSet_LIST:
		altRetVal	= disconnect_disconnection_set( vi_errorAndLog_obj, disconnectionSet, options_OBJ.theForce )
		if altRetVal:
			if retVal == 0:
				retVal	=  altRetVal
		else:
			numOfWWIDdisconnected	+= 1

	return retVal, numOfWWIDdisconnected

####################################################################################################
def get_wwid_list_from_iqn( vi_errorAndLog_obj, given_iqn, refresh ):
	"""
	This routine
	"""
#
	openiscsi_IQN_DICT	= storage_tech.get_openiscsi_iqn_to_blk_dev_list_dict( vi_errorAndLog_obj, refresh )
	blkdev_LIST			= openiscsi_IQN_DICT.get( given_iqn, None )
	if not blkdev_LIST:
		return []

	blk_dev_to_dm_map_lookup_DICT	= storage_tech.get_blk_dev_to_dm_map_lookup_dict( vi_errorAndLog_obj, refresh )

	wwid_DICT	= {}
	for blkdev in blkdev_LIST:
		wwid	= blk_dev_to_dm_map_lookup_DICT.get( blkdev, None )
		if wwid:
			wwid_DICT[ wwid ]	= True

	wwid_LIST	= wwid_DICT.keys()
	wwid_LIST.sort()

	return wwid_LIST

####################################################################################################
def disconnect_based_on_IQN( vi_errorAndLog_obj, options_OBJ, iqn_LIST ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( 'Considering for disconnect:' )
	disconnectionSet_LIST	= []
	for iqn in iqn_LIST[:]:
		if iqn not in storage_tech.get_iqn_list( vi_errorAndLog_obj, False ):
			iqn_LIST.remove(iqn)
			vi_errorAndLog_obj.vi_print_log( "    %s :: Skipping :: does not exist" % ( iqn ) )
		else:
			vi_errorAndLog_obj.vi_print_log( "    %s" % ( iqn ) )

	vi_errorAndLog_obj.vi_print_log( '' )

	if len(iqn_LIST) == 0:
		return 0, 0

	for iqn in iqn_LIST:
		wwid_LIST	= get_wwid_list_from_iqn( vi_errorAndLog_obj, iqn, False )
		for wwid in wwid_LIST:
			disconnectionSet	= figure_out_disconnection_set_for_WWID( vi_errorAndLog_obj, wwid )
			if disconnectionSet:
				disconnectionSet_LIST.append( disconnectionSet )

	if len(disconnectionSet_LIST) == 0:
		return 0, 0

	if options_OBJ.onlyDumpDependencies:
		for disconnectionSet in disconnectionSet_LIST:
			dump_disconnection_set_info( vi_errorAndLog_obj, disconnectionSet )
		return 0, 0

	retVal					= 0
	numOfIQNdisconnected	= 0
	for disconnectionSet in disconnectionSet_LIST:
		altRetVal	= disconnect_disconnection_set( vi_errorAndLog_obj, disconnectionSet, options_OBJ.theForce )
		if altRetVal:
			if retVal == 0:
				retVal	=  altRetVal
		else:
			numOfIQNdisconnected	+= 1

	for iqn in iqn_LIST:
		storage_tech.iscsiadm_exe_sync( vi_errorAndLog_obj, [ 'iscsiadm', '-m', 'node', '--logout', '-T', iqn ] )

	for iqn in iqn_LIST:
		storage_tech.iscsiadm_exe_sync( vi_errorAndLog_obj, [ 'iscsiadm', '-m', 'node', '-o', 'delete', '-T', iqn ] )

	return retVal, numOfIQNdisconnected

####################################################################################################
def list_available_devices( vi_errorAndLog_obj, options_OBJ ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_std( '' )

	dm_maps_DICT	= storage_tech.get_device_mapper_maps( vi_errorAndLog_obj, False )
	if len( dm_maps_DICT ) > 0:
		vi_errorAndLog_obj.vi_print_std( "Disks: [%d]" % ( len( dm_maps_DICT ) ) )
		for dm_map_name in dm_maps_DICT.keys():
			vi_errorAndLog_obj.vi_print_std( "    %s" % ( dm_map_name ) )
		vi_errorAndLog_obj.vi_print_std( '' )

	openiscsi_IQN_DICT	= storage_tech.get_openiscsi_iqn_to_blk_dev_list_dict( vi_errorAndLog_obj, False )
	if len( openiscsi_IQN_DICT ) > 0:
		vi_errorAndLog_obj.vi_print_std( "IQNs: [%d]" % ( len( openiscsi_IQN_DICT ) ) )
		for openiscsi_IQN, blkdev_LIST in openiscsi_IQN_DICT.iteritems():
			wwid_LIST	= get_wwid_list_from_iqn( vi_errorAndLog_obj, openiscsi_IQN, False )
			vi_errorAndLog_obj.vi_print_std( "    %s %s" % ( openiscsi_IQN, blkdev_LIST ) )
			for wwid in wwid_LIST:
				vi_errorAndLog_obj.vi_print_std( "        %s" % ( wwid ) )
		vi_errorAndLog_obj.vi_print_std( '' )

####################################################################################################
def update_WWID_workfile(errAndLog, wwid_LIST):
	"""
	Update the work file with new WWIDs to analayze. The new combined
	list of WWIDs is returned.
	"""

	# Get fileLock on the work list.
	work_lock = vi_utils.vi_fileLock('work file', global_pickle_file)
	work_lock.CreateLockFile()
	while (True):
		if work_lock.AcquireLock(True) == 0:
				break

	# Read the existing pickle file.

	wwid_pickle_LIST = work_lock.LoadPickle()
	if not wwid_pickle_LIST:
		wwid_pickle_LIST = []

	wwid_pickle_LIST.extend(wwid_LIST)
	vi_utils.vi_sort_uniq(wwid_pickle_LIST, False)

	# Dump the list.
	work_lock.SavePickle(wwid_pickle_LIST)
	work_lock.ReleaseLock()
	del work_lock

	return wwid_pickle_LIST

####################################################################################################
def expunge_WWID_from_workfile(errAndLog, wwid_LIST):
	"""
	Expunge from the work file the WWIDs that have been analyzed.
	"""

	# Get fileLock on the work list.
	work_lock = vi_utils.vi_fileLock('work file', global_pickle_file)
	work_lock.CreateLockFile()
	while (True):
		if work_lock.AcquireLock(True) == 0:
			break

	# Read the existing pickle file.
	wwid_pickle_LIST = work_lock.LoadPickle()
	if not wwid_pickle_LIST:
		wwid_pickle_LIST = []

	# Remove from the www_pickle_LIST items that we have tried to disconnect.
	for wwid in wwid_pickle_LIST[:]:
		if wwid in wwid_LIST:
			wwid_pickle_LIST.remove(wwid)

	# Sort it
	vi_utils.vi_sort_uniq(wwid_pickle_LIST, False)

	# update it
	work_lock.SavePickle(wwid_pickle_LIST)
	work_lock.ReleaseLock()

	del work_lock

	return wwid_pickle_LIST

####################################################################################################
def check_multipath_status(errAndLog, work_wwid_LIST):
	"""
	Check the health status of the passed WWIDS.

	The return value is the list of WWIDs which are unhealthy - have no
	active paths and are failed.
	"""

	m_DICT=storage_tech.get_multipath_info( errAndLog, True  )
	unhealthy_wwid_LIST = []
	for wwid_name in work_wwid_LIST:
		blkDev_DICT = m_DICT.get(wwid_name, None)
		if not blkDev_DICT:
			errAndLog.vi_print_log("%s: not present. Skipping." % (wwid_name),)
			continue
		else:
			failed_cnt = 0
			active_cnt = 0
			healthy_LIST = []
			sick_LIST = []
			for blkDev, info in blkDev_DICT.iteritems():
				if info.mPathDMmapState == "failed" or info.mCheckerState == "failed":
					failed_cnt += 1
					sick_LIST.append(blkDev)
				else:
					active_cnt += 1
					healthy_LIST.append(blkDev)

			if len(blkDev_DICT) == failed_cnt:
				errAndLog.vi_print_log("%s: all disks (%s) for it are down. Recommending removal." % (wwid_name, sick_LIST))
				unhealthy_wwid_LIST.append(wwid_name)
			else:
				if len(blkDev_DICT) == active_cnt:
					errAndLog.vi_print_log("%s: OK. Skipping " % (wwid_name))
				else:
					errAndLog.vi_print_log("%s: semi-OK. failed disks (%d): %s, healthy disks (%d): %s. Skipping. " % \
													(wwid_name, failed_cnt, str(sick_LIST), active_cnt, str(healthy_LIST)))

	return unhealthy_wwid_LIST

####################################################################################################
def parse_given_args():
	"""
	This routine
	"""
#
	parser = OptionParser()

	parser.add_option("", "--debugLvl", dest="debugLvl", default=0, type="int",
					help="debugLvl message printing to stderr" )

	parser.add_option( "", "--force", dest="theForce", default=False, action="store_true",
					help="WARNING: Trys to make the node usable BUT reboot is REQUIRED before further disk adds are safe" )

	parser.add_option( "", "--reaper", dest="reaper", default=False, action="store_true",
					help="Reaper. Takes care of sick wwids..")

	parser.add_option( "", "--reap_all", dest="reap_all", default=False, action="store_true",
					help="Check all of the wwids when reaping.")

	parser.add_option("", "--wwidList", dest="wwidList", default=None, type="string",
					help="wwidList:  wwid1,...,wwidN  OR  <<STDIN>>  to read in the list from stdin. This is a list of WWID device-mapper names" )

	parser.add_option( "", "--onlyDumpDependencies", dest="onlyDumpDependencies", default=False, action="store_true",
					help="do NOT disconnect only dump the dependencies" )

	parser.add_option("", "--iqnList", dest="iqnList", default=None, type="string",
					help="iqnList:  iqn1,...,iqnN  OR  <<STDIN>>  to read in the list from stdin" )

	parser.add_option("", "--listDevices", dest="listDevices", default=False, action="store_true",
					help="List the devices available for disconnect")

	options_OBJ, args_LIST = parser.parse_args()

	wwid_LIST = []
	if options_OBJ.wwidList == '<<STDIN>>':
		if options_OBJ.debugLvl >= 4:
			print >> sys.stderr, 'Reading dist_LIST from stdin'
		try:
			while True:
				stdinInputLine = raw_input()
				if stdinInputLine:
					stdinInputLine	= stdinInputLine.strip()
					if len(stdinInputLine):
						wwid_LIST.append( stdinInputLine )
		except EOFError:
			pass
	else:
		if options_OBJ.wwidList:
			for givenItem in options_OBJ.wwidList.split(','):
				givenItem	= givenItem.strip()
				if len(givenItem):
					wwid_LIST.append( givenItem )

	vi_utils.vi_sort_uniq( wwid_LIST, True )

	iqn_LIST = []
	if options_OBJ.iqnList == '<<STDIN>>':
		if options_OBJ.debugLvl >= 4:
			print >> sys.stderr, 'Reading iqn_LIST from stdin'
		try:
			while True:
				stdinInputLine = raw_input()
				if stdinInputLine:
					stdinInputLine	= stdinInputLine.strip()
					if len(stdinInputLine):
						iqn_LIST.append( stdinInputLine )
		except EOFError:
			pass
	else:
		if options_OBJ.iqnList:
			for givenItem in options_OBJ.iqnList.split(','):
				givenItem	= givenItem.strip()
				if len(givenItem):
					iqn_LIST.append( givenItem )

	vi_utils.vi_sort_uniq( iqn_LIST, True )

	if 0:
		print >> sys.stderr, options_OBJ
		print >> sys.stderr, args_LIST
		print >> sys.stderr, dir(options_OBJ)
		print >> sys.stderr, parser.values
		print >> sys.stderr, task_LIST
		vi_utils.vi_dbg_stop()

	if options_OBJ.reap_all:
		options_OBJ.reaper = True

	if len( wwid_LIST ) == 0 and len( iqn_LIST ) == 0 and not (options_OBJ.listDevices or options_OBJ.reaper):
		print >> sys.stderr, '\nno work requested\n'
		parser.print_help(sys.stderr)
		print >> sys.stderr, ''
		vi_utils.vi_exit( 1 )

	return ( options_OBJ, args_LIST, wwid_LIST, iqn_LIST )

####################################################################################################
def init_complex_globals( options_OBJ, args_LIST ):
	"""
	This routine initializes and complex global variables
	"""
#
	global global_errAndLog
	global global_logFilePath

	newLog_FILE = None
	if not options_OBJ.listDevices:
		vi_utils.vi_rototillLogs( global_logDirectory, global_baseLogFileName, 40, 10 )
	
		logFileName = vi_utils.vi_generate_log_filename( global_baseLogFileName )
		global_logFilePath = "%s/%s" % ( global_logDirectory, logFileName )
		try:
			newLog_FILE = open( global_logFilePath, 'a', 0 )		# UNbuffered
		except:
			pass

	errAndLog_dst_DICT =	{	'stdErr'			:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__ERR
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__WRN
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__DBG
														,	sys.stderr
														)
							,	'stdout'			:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__STD
														,	sys.stdout
														)
							,	'log_to_stdout'		:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__LOG
														,	sys.stdout
														)
							}

	if newLog_FILE:
		errAndLog_dst_DICT[ 'log' ]	=	(	vi_utils.VI_ErrAndLog__FOR_FLAG__LOG
										,	newLog_FILE
										)

	global_errAndLog = vi_utils.VI_ErrAndLogRedirector_OBJ( errAndLog_dst_DICT, options_OBJ.debugLvl )

	if newLog_FILE:
		global_errAndLog.vi_print_log( "LOG: %s\n" % ( global_logFilePath ) )
	elif not options_OBJ.listDevices:
		global_errAndLog.vi_print_wrn( "could not create global log FILE '%s'" % ( global_logFilePath ) )

####################################################################################################
def main():
	"""
	This routine is the script entry point
	"""
#
	options_OBJ, args_LIST, wwid_LIST, iqn_LIST = parse_given_args()
	work_wwid_LIST = []

	init_complex_globals( options_OBJ, args_LIST )

	if options_OBJ.reap_all:
		m_DICT=storage_tech.get_multipath_info( global_errAndLog, False  )
		if (m_DICT):
			wwid_LIST = m_DICT.keys()

	if options_OBJ.reaper:
		work_wwid_LIST = update_WWID_workfile(global_errAndLog, wwid_LIST)
		if len(wwid_LIST) > 0:
			global_errAndLog.vi_print_log("Updated work list with %s->%s\n" % (wwid_LIST ,work_wwid_LIST))

		wwid_LIST = iqn_LIST = [] # MUST set to null. This list of WWIDs to delete will be set by 'check_multipath_status'

		if len(work_wwid_LIST) == 0:
			global_errAndLog.vi_print_log("No unhealthy multipath devices to check. Quiting.")
		else:
			wwid_LIST = check_multipath_status(global_errAndLog, work_wwid_LIST)

	if options_OBJ.listDevices:
		list_available_devices( global_errAndLog, options_OBJ )
		wwid_LIST = iqn_LIST = [] # MUST set to null. Will make parent quit peacefully.

	retVal_WWID				= 0
	numOfWWIDdisconnected	= 0
	try:
		if len( wwid_LIST ) > 0:
			retVal_WWID, numOfWWIDdisconnected = disconnect_based_on_WWID( global_errAndLog, options_OBJ, wwid_LIST )
	except:
		global_errAndLog.vi_croak_exception( traceback.format_exc() )
		retVal_WWID	= errno.EFAULT
		global_errAndLog.vi_print_log( '\ncontinuing ...\n\n' )

	retVal_IQN				= 0
	numOfIQNdisconnected	= 0
	try:
		if len( iqn_LIST ) > 0:
			retVal_IQN, numOfIQNdisconnected = disconnect_based_on_IQN( global_errAndLog, options_OBJ, iqn_LIST )
	except:
		global_errAndLog.vi_croak_exception( traceback.format_exc() )
		retVal_IQN	= errno.EFAULT

	if numOfWWIDdisconnected > 0 or numOfIQNdisconnected > 0:
		global_errAndLog.vi_print_log( '\nDisconnect successful' )

	if options_OBJ.reaper:
		expunge_WWID_from_workfile(global_errAndLog, work_wwid_LIST)

	global_errAndLog.close_dst_files()
	vi_utils.vi_exe_sync( global_errAndLog, [ 'gzip', '--best', global_logFilePath ], 30 )

	global_errAndLog.vi_print_log( '' )

	if retVal_WWID:
		return retVal_WWID

	if retVal_IQN:
		return retVal_IQN

	return 0

####################################################################################################
### main

sys.stdout = os.fdopen( sys.stdout.fileno(), sys.stdout.mode, 0 )	# UNbuffer stdout

exitValue  = main()

sys.exit( exitValue )

# END
####################################################################################################
