#!/usr/bin/python
####################################################################################################
#
#	Author:		Joshua Nicholas, Konrad Rzeszutek
#
"""Script for flushing inflight io that may still be in dom0
"""

import sys
import os
import re
import traceback
import errno
import time

import vi_utils
import storage_tech

from optparse import OptionParser
from optparse import OptionGroup

####################################################################################################

global_logDirectory		= '/var/log/flush_inflight_io'
global_baseLogFileName	= 'flush_ifio'

global_errAndLog		= None		# VI_ErrAndLog_OBJ
global_logFilePath		= ''

####################################################################################################

class flushInfo_OBJ:

	def __init__( self ):

		self.mDiskInfo_LIST				= []	
		self.mRootDMgraphComp_LIST		= []	# DeviceMapper_graph_component_OBJ
		self.mAffectedDMgraphComp_LIST	= []	# DeviceMapper_graph_component_OBJ

	def __repr__( self ):

		finalStr	 = "D: %s" % ( self.mDiskInfo_LIST )
		finalStr	+= " R: ["
		if len(self.mRootDMgraphComp_LIST) > 0:
			for rootDMgraphComp in self.mRootDMgraphComp_LIST:
				finalStr	+= "%s," % ( rootDMgraphComp.mMapName )
			finalStr	 = finalStr.rstrip(',')
		finalStr	+= "]"
		finalStr	+= " A: ["
		if len(self.mAffectedDMgraphComp_LIST) > 0:
			for affectedDMgraphComp in self.mAffectedDMgraphComp_LIST:
				finalStr	+= "%s," % ( affectedDMgraphComp.mMapName )
			finalStr	 = finalStr.rstrip(',')
		finalStr	+= "]"
		return finalStr

class diskInfo_OBJ:

	def __init__( self, theWWID, theDMmapBlkDev ):

		self.mWWID					= theWWID
		self.mDMmapBlkDev			= theDMmapBlkDev
		self.mDMmapBlkDevStatIFcnt	= -1
		self.mIsBlkDevPathStateGood	= False
		self.mBlkDevSGinfo_LIST		= []	# blkDevSGinfo_OBJ

	def refresh_inflight_count( self, vi_errorAndLog_obj ):

		mpDMmapBlkDevStats	= storage_tech.get_stats_for_blkDev( vi_errorAndLog_obj, self.mDMmapBlkDev )
		if mpDMmapBlkDevStats:
			self.mDMmapBlkDevStatIFcnt	= mpDMmapBlkDevStats.mInFlight
		else:
			self.mDMmapBlkDevStatIFcnt	= -1
		return self.mDMmapBlkDevStatIFcnt

	def __cmp__( self, other ):

		if not isinstance( other, diskInfo_OBJ ):
			return -1
		return cmp( self.mWWID, other.mWWID )

	def __repr__( self ):

		return "%s (%d)" % \
					(	self.mWWID
					,	self.mDMmapBlkDevStatIFcnt )

class blkDevSGinfo_OBJ:

	def __init__( self, theBlkDev, theBlkDevPathState, theBlkDevCheckerState ):

		self.mBlkDev				= theBlkDev
		self.mBlkDevPathState		= theBlkDevPathState
		self.mBlkDevCheckerState	= theBlkDevCheckerState
		self.mBlkDevStats			= None	# BlkDevStats_OBJ
		self.mSGdev					= None

	def is_state_good( self ):

		if self.mBlkDevPathState == 'active' and self.mBlkDevCheckerState == 'ready':
			return True
		return False

	def get_str( self ):

		return "mBlkDev: %s mBlkDevPathState: %s mBlkDevCheckerState: %s mBlkDevStats: %s mSGdev: %s" % \
					(	self.mBlkDev
					,	self.mBlkDevPathState
					,	self.mBlkDevCheckerState
					,	self.mBlkDevStats
					,	self.mSGdev
					)

	def __repr__( self ):

		if self.mBlkDevStats:
			ifCnt	= self.mBlkDevStats.mInFlight
		else:
			ifCnt	= -1
		return "( bd: %s sg: %s ps: %s ckr: %s if: %s )" % \
					(	self.mBlkDev
					,	self.mSGdev
					,	self.mBlkDevPathState[ : 2 ]
					,	self.mBlkDevCheckerState[ : 2 ]
					,	ifCnt
					)

####################################################################################################
def recurse_walk_for_intermediates( dmGraphComp, cbArgTPL, cbAccumulation ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj	= cbArgTPL

	if dmGraphComp.mMapType == storage_tech.VI_DM_GRAPH_COMPONENT_TYPE__MULTIPATH:
		return True, cbAccumulation

	intermediates_LIST	= cbAccumulation

	found	= False
	for knownDMgraphComp in intermediates_LIST:
		if knownDMgraphComp.mMapName == dmGraphComp.mMapName:
			found	= True
			break
	if not found:
		intermediates_LIST.append( dmGraphComp )

	return False, intermediates_LIST

####################################################################################################
def flush_disk( vi_errorAndLog_obj, diskInfo, forceDeviceReset, forceBusReset ):
	"""
	This routine
	"""
#
	for blkDevSGinfo in diskInfo.mBlkDevSGinfo_LIST:
		blkDevSGinfo.mBlkDevStats	= storage_tech.get_stats_for_blkDev( vi_errorAndLog_obj, blkDevSGinfo.mBlkDev )

	for blkDevSGinfo in diskInfo.mBlkDevSGinfo_LIST:
		if not blkDevSGinfo.mBlkDevStats:
			vi_errorAndLog_obj.vi_print_log( "NoStats  SG [%s]  ????  : %s" % ( diskInfo, blkDevSGinfo ) )
#			continue

		didSGreset	= False
		sgDevPath	= "/dev/%s" % ( blkDevSGinfo.mSGdev )
		exitStatus	= 0

		if blkDevSGinfo.mBlkDevStats.mInFlight > 0 or forceDeviceReset:
			didSGreset	= True
			vi_errorAndLog_obj.vi_print_log( "Reseting SG [%s]   dev  : %s" % ( diskInfo, blkDevSGinfo ) )
			cmd_LIST	= [ 'sg_reset', '-d', sgDevPath ]
			stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST )
			if stdout_LIST != None and len(stdout_LIST) > 0:
				vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
			if stderr_LIST != None and len(stderr_LIST) > 0:
				vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

			vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'sync' ], 10 )
			blkDevSGinfo.mBlkDevStats	= storage_tech.get_stats_for_blkDev( vi_errorAndLog_obj, blkDevSGinfo.mBlkDev )

		if exitStatus != 0 or ( blkDevSGinfo.mBlkDevStats and blkDevSGinfo.mBlkDevStats.mInFlight > 0 ) or forceBusReset:
			didSGreset	= True
			vi_errorAndLog_obj.vi_print_log( "Reseting SG [%s]   bus  : %s" % ( diskInfo, blkDevSGinfo ) )
			cmd_LIST	= [ 'sg_reset', '-b', sgDevPath ]
			stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST )
			if stdout_LIST != None and len(stdout_LIST) > 0:
				vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
			if stderr_LIST != None and len(stderr_LIST) > 0:
				vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

			vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'sync' ], 10 )
			blkDevSGinfo.mBlkDevStats	= storage_tech.get_stats_for_blkDev( vi_errorAndLog_obj, blkDevSGinfo.mBlkDev )

		if didSGreset:
			vi_errorAndLog_obj.vi_print_log( "Reset    SG [%s] result : %s" % ( diskInfo, blkDevSGinfo ) )
		else:
			vi_errorAndLog_obj.vi_print_log( "NoInFlit SG [%s]  skip  : %s" % ( diskInfo, blkDevSGinfo ) )

	return 0

####################################################################################################
def suspend_disk( vi_errorAndLog_obj, diskInfo ):
	"""
	This routine
	"""
#
	diskInfo.refresh_inflight_count( vi_errorAndLog_obj )

	vi_errorAndLog_obj.vi_print_log( "Suspending  [%s] %s : %s" % ( diskInfo, diskInfo.mDMmapBlkDev, diskInfo.mBlkDevSGinfo_LIST ) )

	dmSuspendProc = vi_utils.vi_exe_async( vi_errorAndLog_obj, [ 'dmsetup', 'suspend', diskInfo.mWWID ] )

	##### Wait for suspend to complete
	resetCnt	= 0
	itrCnt		= 0
	while True:
		itrCnt	+= 1

		done_LIST, wasSomeProgressMade	= vi_utils.vi_monitor_async( vi_errorAndLog_obj, dmSuspendProc )
		if len( done_LIST ) > 0:
			break

		if dmSuspendProc.get_elapsed_time() < 20 or itrCnt < 20:
			time.sleep( 0.2 )
			continue

		if resetCnt != 2:
			flush_disk( vi_errorAndLog_obj, diskInfo, True, False )
		else:
			flush_disk( vi_errorAndLog_obj, diskInfo, True, True )

		resetCnt	+= 1
		if resetCnt >= 3:
			diskInfo.refresh_inflight_count( vi_errorAndLog_obj )

			vi_errorAndLog_obj.vi_print_log( "Suspend FAILED [%s] %s : %s" % ( diskInfo, diskInfo.mDMmapBlkDev, diskInfo.mBlkDevSGinfo_LIST ) )

			dmSuspendProc.terminate_child()
			stdout_LIST, stderr_LIST, exitStatus, deltaTime	= dmSuspendProc.get_child_results()
			if stdout_LIST != None and len(stdout_LIST) > 0:
				vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
			if stderr_LIST != None and len(stderr_LIST) > 0:
				vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )
			return errno.EBUSY

		time.sleep( resetCnt )

	##### Suspend completed

	stdout_LIST, stderr_LIST, exitStatus, deltaTime	= dmSuspendProc.get_child_results()
	if stdout_LIST != None and len(stdout_LIST) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
	if stderr_LIST != None and len(stderr_LIST) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

	vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'sync' ], 10 )
	diskInfo.refresh_inflight_count( vi_errorAndLog_obj )

	vi_errorAndLog_obj.vi_print_log( "Suspended   [%s] %s : %s" % ( diskInfo, diskInfo.mDMmapBlkDev, diskInfo.mBlkDevSGinfo_LIST ) )

	return 0

####################################################################################################
def resume_disk( vi_errorAndLog_obj, diskInfo ):
	"""
	This routine
	"""
#
	if not diskInfo.mIsBlkDevPathStateGood and False:	# This is only as good as the check, which should be good, but may not always be

		vi_errorAndLog_obj.vi_print_log( "Deleting    [%s]" % ( diskInfo ) )

		cmd_LIST	= [ 'dmsetup', 'remove', '--force', diskInfo.mWWID ]

		stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST )
#		if stdout_LIST != None and len(stdout_LIST) > 0:
#			vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
#		if stderr_LIST != None and len(stderr_LIST) > 0:
#			vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

		vi_errorAndLog_obj.vi_print_log( "Deleted     [%s] disk removed" % ( diskInfo ) )

		return 0

	vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'sync' ], 10 )
	diskInfo.refresh_inflight_count( vi_errorAndLog_obj )

	vi_errorAndLog_obj.vi_print_log( "Resuming    [%s] %s : %s" % ( diskInfo, diskInfo.mDMmapBlkDev, diskInfo.mBlkDevSGinfo_LIST ) )

	cmd_LIST	= [ 'dmsetup', 'resume', diskInfo.mWWID ]

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST )
	if stdout_LIST != None and len(stdout_LIST) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
	if stderr_LIST != None and len(stderr_LIST) > 0:
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

	vi_errorAndLog_obj.vi_print_log( "Resumed     [%s] flush complete" % ( diskInfo ) )

	return 0

####################################################################################################
def flush_inflight_io( vi_errorAndLog_obj, options_OBJ, flushInfo ):
	"""
	This routine
	"""
#
	for diskInfo in flushInfo.mDiskInfo_LIST:
		retVal	= suspend_disk( vi_errorAndLog_obj, diskInfo )
		if retVal:
			for alt_diskInfo in flushInfo.mDiskInfo_LIST:
				if alt_diskInfo == diskInfo:
					break
				resume_disk( vi_errorAndLog_obj, alt_diskInfo )
			return retVal

	for diskInfo in flushInfo.mDiskInfo_LIST:
		flush_disk( vi_errorAndLog_obj, diskInfo, options_OBJ.forceDeviceReset, options_OBJ.forceBusReset )

	##### Reset complete now, dump more info  ...

	for dmGraphComp in flushInfo.mAffectedDMgraphComp_LIST:
		dmGraphCompBlkDevStats	= storage_tech.get_stats_for_blkDev( vi_errorAndLog_obj, dmGraphComp.mBlkDevName )
		if dmGraphCompBlkDevStats:
			dmGraphCompBlkDevStatIFcnt	= dmGraphCompBlkDevStats.mInFlight
		else:
			dmGraphCompBlkDevStatIFcnt	= -1
		vi_errorAndLog_obj.vi_print_log( "Affected DM : lvl %d if %d : %s" % ( dmGraphComp.mDepLvl, dmGraphCompBlkDevStatIFcnt, dmGraphComp.mMapName ) )

	##### Should this multipath be re-enabled ...

	for diskInfo in flushInfo.mDiskInfo_LIST:
		resume_disk( vi_errorAndLog_obj, diskInfo )

	return 0

####################################################################################################
def flush_inflight_ios( vi_errorAndLog_obj, options_OBJ, flushInfo_LIST ):
	"""
	This routine
	"""
#
	dmMapGraph_DICT, blkDevParentDMmap_DICT	= storage_tech.get_device_mapper_map_graph( vi_errorAndLog_obj, False )
	multipathInfo_DICT						= storage_tech.get_multipath_info( vi_errorAndLog_obj, False )

	sgMajorDevNum	= storage_tech.get_character_device_major_num( vi_errorAndLog_obj, 'sg', False )
	sgPrefix		= "%d:" % ( sgMajorDevNum )

	#
	# PREP for flush
	#
	for flushInfo in flushInfo_LIST:

		for diskInfo in flushInfo.mDiskInfo_LIST:

			blkDev_DICT	= multipathInfo_DICT.get( diskInfo.mWWID, None )
			if not blkDev_DICT:
				vi_errorAndLog_obj.vi_print_log( "Disk: %s : not in multipath - skipping" % ( diskInfo.mWWID ) )
				continue

			if len( blkDev_DICT ) < 1:
				vi_errorAndLog_obj.vi_print_log( "Disk: %s : no blkDevs - skipping" % ( diskInfo.mWWID ) )
				continue

			for blkDevName, mpBlkDevInfo in blkDev_DICT.iteritems():

				blkDevSGinfo				= blkDevSGinfo_OBJ( blkDevName, mpBlkDevInfo.mPathDMmapState, mpBlkDevInfo.mCheckerState )
				if not blkDevSGinfo.is_state_good():
					vi_errorAndLog_obj.vi_print_log( "Disk: %s : path %s not good : %s : %s" % \
														(	diskInfo.mWWID
														,	blkDevSGinfo.mBlkDev
														,	blkDevSGinfo.mBlkDevPathState
														,	blkDevSGinfo.mBlkDevCheckerState ) )

				blkDevSGinfo.mBlkDevStats	= storage_tech.get_stats_for_blkDev( vi_errorAndLog_obj, blkDevSGinfo.mBlkDev )
				if not blkDevSGinfo.mBlkDevStats:
					vi_errorAndLog_obj.vi_print_log( "Disk: %s : path %s stats not available !?!" % ( diskInfo.mWWID, blkDevSGinfo.mBlkDev ) )

				sgDevNumsPath			= "/sys/block/%s/device/generic/dev" % ( blkDevSGinfo.mBlkDev )
				sgDevNums, read_failed	= vi_utils.vi_read_text_value( vi_errorAndLog_obj, sgDevNumsPath )
				if not read_failed and sgDevNums.startswith( sgPrefix ):
					blkDevSGinfo.mSGdev	= "sg%s" % ( sgDevNums[ len(sgPrefix) : ] )
				else:
					vi_errorAndLog_obj.vi_print_log( "Disk: %s : path %s SG available !?!" % ( diskInfo.mWWID, blkDevSGinfo.mBlkDev ) )

				diskInfo.mBlkDevSGinfo_LIST.append( blkDevSGinfo )

				if blkDevSGinfo.is_state_good():
					diskInfo.mIsBlkDevPathStateGood	= True

			blkDevParentDMmap	= blkDevParentDMmap_DICT.get( diskInfo.mWWID , None )
			if not blkDevParentDMmap:
				vi_errorAndLog_obj.vi_print_log( "Given wwid %s : dmMap does not exist - skipping" % ( diskInfo.mWWID ) )

			for lvmUsing in blkDevParentDMmap.mLVMusing_LIST:
				dmGraphComp	= dmMapGraph_DICT.get( lvmUsing.mMapName, None )
				if not dmGraphComp:
					vi_errorAndLog_obj.vi_print_wrn( "Disk: %s : lvmComp %s missing" % ( diskInfo.mWWID, lvmUsing.mMapName ) )
					continue

				dmGraphComp.recursive_walk_pre( recurse_walk_for_intermediates, vi_errorAndLog_obj, flushInfo.mAffectedDMgraphComp_LIST )

	if 0:
		for flushInfo in flushInfo_LIST:
			for diskInfo in flushInfo.mDiskInfo_LIST:
				diskInfo.refresh_inflight_count( vi_errorAndLog_obj )
			vi_errorAndLog_obj.vi_print_log( "fI: %s" % ( flushInfo ) )
		vi_errorAndLog_obj.vi_stop()

	#
	# Perform flush
	#
	retVal	= 0
	for flushInfo in flushInfo_LIST:
		alt_retVal	= flush_inflight_io( vi_errorAndLog_obj, options_OBJ, flushInfo )
		if alt_retVal and not retVal:
			retVal	= alt_retVal

		vi_errorAndLog_obj.vi_print_log( '' )

	if retVal:
		return retVal

	return 0

####################################################################################################
def flush_inflight_io_based_on_WWIDs( vi_errorAndLog_obj, options_OBJ, wwid_LIST ):
	"""
	This routine
	"""
#
	dmMapGraph_DICT, blkDevParentDMmap_DICT	= storage_tech.get_device_mapper_map_graph( vi_errorAndLog_obj, False )

	flushInfo_LIST		= []
	for wwid in wwid_LIST:
		if not os.path.exists( "/dev/mapper/%s" % ( wwid ) ):
			vi_errorAndLog_obj.vi_print_wrn( "Given wwid %s : does NOT exist - skipping" % ( wwid ) )
			continue
		blkDevParentDMmap	= blkDevParentDMmap_DICT.get( wwid, None )
		if not blkDevParentDMmap:
			vi_errorAndLog_obj.vi_print_log( "Given wwid %s : dmMap does not exist - skipping" % ( wwid ) )
			continue

		diskInfo	= diskInfo_OBJ( blkDevParentDMmap.mMapName, "dm-%d" % ( blkDevParentDMmap.mMinorNum ) )
		flushInfo	= flushInfo_OBJ()
		flushInfo.mDiskInfo_LIST.append( diskInfo )
		flushInfo_LIST.append( flushInfo )

	if 0:
		for flushInfo in flushInfo_LIST:
			vi_errorAndLog_obj.vi_print_log( "fI: %s" % ( flushInfo ) )

	return flush_inflight_ios( vi_errorAndLog_obj, options_OBJ, flushInfo_LIST )

####################################################################################################
def flush_inflight_io_based_on_backends( vi_errorAndLog_obj, options_OBJ, backendPath_LIST ):
	"""
	This routine
	"""
#
	backend_LIST	= []
	for backendPath in backendPath_LIST:
		if not os.path.exists( backendPath ):
			vi_errorAndLog_obj.vi_print_wrn( "Given backendPath %s : does NOT exist - skipping" % ( backendPath ) )
			continue
		if not backendPath.startswith( '/dev/mapper/' ):
			vi_errorAndLog_obj.vi_print_wrn( "Given backendPath %s : is NOT in /dev/mapper - skipping" % ( backendPath ) )
			continue
		backend_LIST.append( backendPath[ 12 : ] )

	dmMapGraph_DICT, blkDevParentDMmap_DICT	= storage_tech.get_device_mapper_map_graph( vi_errorAndLog_obj, False )
	dm_maps_DICT	= storage_tech.get_device_mapper_maps( vi_errorAndLog_obj, False )

	flushInfo_LIST		= []
	for backend in backend_LIST:

		# Is a raw disk
		blkDevParentDMmap	= blkDevParentDMmap_DICT.get( backend, None )
		if blkDevParentDMmap:
			diskInfo	= diskInfo_OBJ( blkDevParentDMmap.mMapName, "dm-%d" % ( blkDevParentDMmap.mMinorNum ) )
			found		= False
			for flushInfo in flushInfo_LIST:
				if diskInfo in flushInfo.mDiskInfo_LIST:
					found	= True
					break
			if not found:
				flushInfo	= flushInfo_OBJ()
				flushInfo.mDiskInfo_LIST.append( diskInfo )
				flushInfo_LIST.append( flushInfo )
			continue

		# Get the root comp for this LV
		rootDMgraphComp	= dmMapGraph_DICT.get( backend, None )
		if not rootDMgraphComp:
			vi_errorAndLog_obj.vi_print_wrn( "Given backend %s : NOT a root ?!?" % ( backend ) )	# survivable

		# Search for disks that use backend
		alreadyInIdx_LIST		= []
		blkDevParentDMmap_LIST	= []
		for blkDevParentDMmapName, blkDevParentDMmap in blkDevParentDMmap_DICT.iteritems():
			for lvmUsing in blkDevParentDMmap.mLVMusing_LIST:
				if lvmUsing.mMapName == backend:
					blkDevParentDMmap_LIST.append( blkDevParentDMmap )
					diskInfo	= diskInfo_OBJ( blkDevParentDMmap.mMapName, "dm-%d" % ( blkDevParentDMmap.mMinorNum ) )
					for flushInfoIdx, flushInfo in enumerate(flushInfo_LIST):
						if diskInfo in flushInfo.mDiskInfo_LIST:
							alreadyInIdx_LIST.append( flushInfoIdx )
					break

		# Is there anything ?
		if len(blkDevParentDMmap_LIST) == 0:
			vi_errorAndLog_obj.vi_print_wrn( "No disks found for '%s' - skipping" % ( backend ) )
			continue

		if len( alreadyInIdx_LIST ) == 0:
			flushInfo	= None
		elif len( alreadyInIdx_LIST ) == 1:
			flushInfo	= flushInfo_LIST[ alreadyInIdx_LIST.pop(0) ]
		else:
			flushInfo	= flushInfo_LIST[ alreadyInIdx_LIST.pop(0) ]
			for alreadyInIdx in alreadyInIdx_LIST:
				alreadyInFlushInfo	= flushInfo_LIST[ alreadyInIdx ]
				flushInfo.mDiskInfo_LIST.extend( alreadyInFlushInfo.mDiskInfo_LIST )
				flushInfo.mRootDMgraphComp_LIST.extend( alreadyInFlushInfo.mRootDMgraphComp_LIST )
			for alreadyInIdx in reversed(alreadyInIdx_LIST):
				del( flushInfo_LIST[ alreadyInIdx ] )

		if not flushInfo:
			flushInfo	= flushInfo_OBJ()
			for blkDevParentDMmap in blkDevParentDMmap_LIST:
				diskInfo	= diskInfo_OBJ( blkDevParentDMmap.mMapName, "dm-%d" % ( blkDevParentDMmap.mMinorNum ) )
				flushInfo.mDiskInfo_LIST.append( diskInfo )
			if rootDMgraphComp:
				flushInfo.mRootDMgraphComp_LIST.append( rootDMgraphComp )
			flushInfo_LIST.append( flushInfo )
		else:
			for blkDevParentDMmap in blkDevParentDMmap_LIST:
				diskInfo	= diskInfo_OBJ( blkDevParentDMmap.mMapName, "dm-%d" % ( blkDevParentDMmap.mMinorNum ) )
				if not diskInfo in flushInfo.mDiskInfo_LIST:
					flushInfo.mDiskInfo_LIST.append( diskInfo )
			if rootDMgraphComp:
				if not rootDMgraphComp in flushInfo.mRootDMgraphComp_LIST:
					flushInfo.mRootDMgraphComp_LIST.append( rootDMgraphComp )

	if 0:
		for flushInfo in flushInfo_LIST:
			vi_errorAndLog_obj.vi_print_log( "fI: %s" % ( flushInfo ) )

	return flush_inflight_ios( vi_errorAndLog_obj, options_OBJ, flushInfo_LIST )

####################################################################################################
def parse_given_args():
	"""
	This routine
	"""
#
	parser = OptionParser()

	parser.add_option("", "--debugLvl", dest="debugLvl", default=0, type="int",
					help="debugLvl message printing to stderr" )

	parser.add_option( "", "--forceDeviceReset", dest="forceDeviceReset", default=False, action="store_true",
					help="Do a device reset anyway, even if otherwise")

	parser.add_option( "", "--forceBusReset", dest="forceBusReset", default=False, action="store_true",
					help="Do a bus reset anyway, even if the device reset succeeds or is otherwise unnecessary")

	parser.add_option("", "--wwidList", dest="wwidList", default=None, type="string",
					help="wwidList:  wwid1,...,wwidN  OR  <<STDIN>>  to read in the list from stdin. This is a list of WWID device-mapper names" )

	parser.add_option("", "--backendPathList", dest="backendPathList", default=None, type="string",
					help="backendPathList:  backend1,...,backendN  OR  <<STDIN>>  to read in the list from stdin. This is a list of backend names" )

	options_OBJ, args_LIST = parser.parse_args()

	wwid_LIST = []
	if options_OBJ.wwidList == '<<STDIN>>':
		if options_OBJ.debugLvl >= 4:
			print >> sys.stderr, 'Reading dist_LIST from stdin'
		try:
			while True:
				stdinInputLine = raw_input()
				if stdinInputLine:
					stdinInputLine	= stdinInputLine.strip()
					if len(stdinInputLine):
						wwid_LIST.append( stdinInputLine )
		except EOFError:
			pass
	else:
		if options_OBJ.wwidList:
			for givenItem in options_OBJ.wwidList.split(','):
				givenItem	= givenItem.strip()
				if len(givenItem):
					wwid_LIST.append( givenItem )

	vi_utils.vi_sort_uniq( wwid_LIST, True )

	backendPath_LIST = []
	if options_OBJ.backendPathList == '<<STDIN>>':
		if options_OBJ.debugLvl >= 4:
			print >> sys.stderr, 'Reading backendPath_LIST from stdin'
		try:
			while True:
				stdinInputLine = raw_input()
				if stdinInputLine:
					stdinInputLine	= stdinInputLine.strip()
					if len(stdinInputLine):
						backendPath_LIST.append( stdinInputLine )
		except EOFError:
			pass
	else:
		if options_OBJ.backendPathList:
			for givenItem in options_OBJ.backendPathList.split(','):
				givenItem	= givenItem.strip()
				if len(givenItem):
					backendPath_LIST.append( givenItem )

	vi_utils.vi_sort_uniq( backendPath_LIST, True )

	if 0:
		print >> sys.stderr, options_OBJ
		print >> sys.stderr, args_LIST
		print >> sys.stderr, dir(options_OBJ)
		print >> sys.stderr, parser.values
		print >> sys.stderr, task_LIST
		vi_utils.vi_dbg_stop()

	if len( wwid_LIST ) == 0 and len( backendPath_LIST ) == 0:
		print >> sys.stderr, '\nno work requested\n'
		parser.print_help(sys.stderr)
		print >> sys.stderr, ''
		vi_utils.vi_exit( 1 )

	return ( options_OBJ, args_LIST, wwid_LIST, backendPath_LIST )

####################################################################################################
def init_complex_globals( options_OBJ, args_LIST ):
	"""
	This routine initializes and complex global variables
	"""
#
	global global_errAndLog
	global global_logFilePath

	newLog_FILE = None
	vi_utils.vi_rototillLogs( global_logDirectory, global_baseLogFileName, 40, 10 )

	logFileName = vi_utils.vi_generate_log_filename( global_baseLogFileName )
	global_logFilePath = "%s/%s" % ( global_logDirectory, logFileName )
	try:
		newLog_FILE = open( global_logFilePath, 'a', 0 )		# UNbuffered
	except:
		pass

	errAndLog_dst_DICT =	{	'stdErr'			:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__ERR
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__WRN
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__DBG
														,	sys.stderr
														)
							,	'stdout'			:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__STD
														,	sys.stdout
														)
							,	'log_to_stdout'		:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__LOG
														,	sys.stdout
														)
							}

	if newLog_FILE:
		errAndLog_dst_DICT[ 'log' ]	=	(	vi_utils.VI_ErrAndLog__FOR_FLAG__LOG
										,	newLog_FILE
										)

	global_errAndLog = vi_utils.VI_ErrAndLogRedirector_OBJ( errAndLog_dst_DICT, options_OBJ.debugLvl )

	if newLog_FILE:
		global_errAndLog.vi_print_log( "LOG: %s\n" % ( global_logFilePath ) )
	else:
		global_errAndLog.vi_print_wrn( "could not create global log FILE '%s'" % ( global_logFilePath ) )

####################################################################################################
def main():
	"""
	This routine is the script entry point
	"""
#
	options_OBJ, args_LIST, wwid_LIST, backendPath_LIST = parse_given_args()

	init_complex_globals( options_OBJ, args_LIST )

	retVal	= 0

	if len(wwid_LIST) > 0:
		try:
			alt_retVal = flush_inflight_io_based_on_WWIDs( global_errAndLog, options_OBJ, wwid_LIST )
		except:
			global_errAndLog.vi_croak_exception( traceback.format_exc() )
			alt_retVal	= errno.EFAULT
		if alt_retVal and not retVal:
			retVal	= alt_retVal

	if len(backendPath_LIST) > 0:
		try:
			alt_retVal = flush_inflight_io_based_on_backends( global_errAndLog, options_OBJ, backendPath_LIST )
		except:
			global_errAndLog.vi_croak_exception( traceback.format_exc() )
			alt_retVal	= errno.EFAULT
		if alt_retVal and not retVal:
			retVal	= alt_retVal

	if not retVal:
		global_errAndLog.vi_print_log( 'Flush successful' )

	global_errAndLog.close_dst_files()
	vi_utils.vi_exe_sync( global_errAndLog, [ 'gzip', '--best', global_logFilePath ], 30 )

	global_errAndLog.vi_print_log( '' )

	if retVal:
		return retVal

	return 0

####################################################################################################
### main

sys.stdout = os.fdopen( sys.stdout.fileno(), sys.stdout.mode, 0 )	# UNbuffer stdout

exitValue  = main()

sys.exit( exitValue )

# END
####################################################################################################
