#!/usr/bin/python
####################################################################################################
#
#	Author:		Joshua Nicholas, Konrad Rzeszutek
#
"""Script for rescanning disks
"""

import sys
import os
import re
import traceback
import difflib

import vi_utils
import vi_xml
import vi_os
import storage_tech

from optparse import OptionParser
from optparse import OptionGroup

####################################################################################################

global_logDirectory			= '/var/log/rescan_for_disks'
global_baseLogFileName		= 'rescan'
global_iscsiPortalListPath	= '/etc/iscsi_portal_list.xml'
global_sectionPreable		= '========'

global_errAndLog			= None		# VI_ErrAndLog_OBJ
global_logFilePath			= ''

####################################################################################################

class Task_OBJ:
	def __init__(self, theArgNameStr, theHelpStr, defaultDoScan, theFcn):
		self.mArgNameStr     = theArgNameStr
		self.mHelpStr        = theHelpStr
		self.mDoScan         = defaultDoScan
		self.mFcn            = theFcn
	def __repr__(self):
		return "mArgNameStr: %s mHelpStr: %s mDoScan: %s mFcn: %s" % (
						self.mArgNameStr
					,	self.mHelpStr
					,	self.mDoScan
					,	self.mFcn )

class Qualifier_OBJ:
	def __init__(self, theArgNameStr, theHelpStr, defaultDoScan):
		self.mArgNameStr     = theArgNameStr
		self.mHelpStr        = theHelpStr
		self.mDoScan         = defaultDoScan
	def __repr__(self):
		return "mArgNameStr: %s mHelpStr: %s mDoScan: %s" % (
						self.mArgNameStr
					,	self.mHelpStr
					,	self.mDoScan )

####################################################################################################

global_scan_scsi_spacer                    = '                          '

####################################################################################################
def get_portal_list(vi_errorAndLog_obj):
	"""
	Scan the Open iSCSI (and QLA4xxx) XML file for the list of: a) unbound to a MAC
	portals, and b) for a list of portals bound to specified MAC.

	The portals bound to specified MACs are 'bounded' to said MAC by adding a host a route.
	If the XML file specifies a new host route, the old one would be deleted and added
	on the new MAC. Both types of portals are checked for IP address sanity.

	Returns:
	 A list of portals. No check is made whether they are 'ping' reachable, thought
	they are filter to see if a subnet-destination route exists for the portal IP address.
	"""
#
	portal_LIST = vi_xml.vi_iscsi_get_portal_list(vi_errorAndLog_obj, global_iscsiPortalListPath, True)
	portal_list_bound_to_MAC_addr_DICT = vi_xml.vi_iscsi_get_portal_list_assoc_to_MAC_addr(vi_errorAndLog_obj, global_iscsiPortalListPath, False)

	vi_errorAndLog_obj.vi_print_log("Considering bound to MAC addresses iSCSI portals [%d]." % (len(portal_list_bound_to_MAC_addr_DICT)))

	for bound_MAC_addr, bound_portal_LIST in portal_list_bound_to_MAC_addr_DICT.iteritems():
		for bound_portal_IP_and_port_str in bound_portal_LIST:
			portal_IP, portal_Port = vi_os.vi_extract_IP_addr(bound_portal_IP_and_port_str)
			# ensure_route_for_IP takes care of checking the subnet, validaty of MAC, and adding/deleting
			# the host routes.
			if vi_os.vi_ensure_route_for_IP(vi_errorAndLog_obj, True, portal_IP, bound_MAC_addr) == 0:
				portal_LIST.append(bound_portal_IP_and_port_str)
			else:
				vi_errorAndLog_obj.vi_print_wrn("Could not map %s to %s!" % (portal_IP, bound_MAC_addr))

	return portal_LIST

####################################################################################################
def task_scan_openiscsi( vi_errorAndLog_obj, options_OBJ, qualifier_DICT ):
	"""
	This routine scans the Open-iSCSI devices.
	"""
#
	pid_LIST = vi_utils.vi_pidof(vi_errorAndLog_obj, 'iscsid')
	if len(pid_LIST) == 0:
		vi_errorAndLog_obj.vi_print_log("No iSCSI daemon. Skipping Open-iSCSI rescan.\n")
		return

	vi_errorAndLog_obj.vi_print_log(global_sectionPreable, "Rescanning iSCSI [%s]." % (global_iscsiPortalListPath))

	portal_LIST = get_portal_list(vi_errorAndLog_obj)
	vi_errorAndLog_obj.vi_print_log("Considering iSCSI portals [%d]." % (len(portal_LIST)))
	for portal in portal_LIST:
		if storage_tech.discover_portal(vi_errorAndLog_obj, portal) != 0:
			vi_errorAndLog_obj.vi_print_wrn("Failed to discover portal [%s]." % (portal))

	# Extract the session data ("connected")
	session_DICT = storage_tech.get_openiscsi_session_info_dict(vi_errorAndLog_obj, True)
	session_LIST = session_DICT.values()

	vi_errorAndLog_obj.vi_print_log("Currently logged in SW iSCSI targets [%d]:" % (len(session_LIST)))
	for session in session_LIST:
		vi_errorAndLog_obj.vi_print_log( "%s" % (session) )

	# .. nodes for all of the portals Open-iSCSI discovered (on some of them we might have logged in)
	node_LIST = storage_tech.get_openiscsi_node_list(vi_errorAndLog_obj, True)
	vi_errorAndLog_obj.vi_print_log("Considering SW iSCSI targets: [%d]" % (len(node_LIST)))
	for node in node_LIST:
		for session in session_LIST:
			if session == node:
				vi_errorAndLog_obj.vi_print_log( "Already logged into SW iSCSI target %s @ %s (%s). Re-scanning" % ( node.mIQN, node.mTargetIP, node.mTargetPort ) )
				storage_tech.ensure_openiscsi_session(vi_errorAndLog_obj, session)
				break
		else:
			vi_errorAndLog_obj.vi_print_log( "Logging in to %s @ %s (%s)." % ( node.mIQN, node.mTargetIP, node.mTargetPort ) )
			storage_tech.login_on_node(vi_errorAndLog_obj, node)

	return 0

####################################################################################################
def get_aoe_dm_maps( vi_errorAndLog_obj, refresh ):
	"""
	This routine returns a dictionary of device-mapper maps which depend on the aoe device.
	"""
#
	aoeBlkDevMajor	= storage_tech.get_block_device_major_num( vi_errorAndLog_obj, 'aoe', False )
	if aoeBlkDevMajor < 0:
		vi_errorAndLog_obj.vi_croak( 'aoe block device not in /proc/devices' )
		return {}

	aoe_dm_map_name_cmpldRegEx = re.compile( "^%s:\d+$" % ( aoeBlkDevMajor ) )
	aoe_dm_maps_DICT           = {}
	dm_maps_DICT               = storage_tech.get_device_mapper_maps( vi_errorAndLog_obj, refresh )

	for dm_map_name in dm_maps_DICT:
		dm_map = dm_maps_DICT[ dm_map_name ]
		for dm_map_dep in dm_map.mDep_LIST:
			if aoe_dm_map_name_cmpldRegEx.match( dm_map_dep.mBlkDevNums ):
				aoe_dm_maps_DICT[ dm_map_name ] = dm_map
				break

	return aoe_dm_maps_DICT

####################################################################################################
def ensure_aoe_disk_map_uptodate( vi_errorAndLog_obj, aoe_disk, aoe_interface ):
	"""
	This routine ensures the AOE maps exists for the passed configurtion. If not it
	creates a device mapper map of an AOE disk.
	"""
#
	aoe_devPath	= "/dev/etherd/%s" % ( aoe_disk )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'ata_id', '--export', aoe_devPath ], 10 )

	serial_cmpldRegEx	= re.compile( '^ID_SERIAL=\s*' )

	dm_map_name = None
	for ata_id_line in stdout_LIST:
		ata_id_line = ata_id_line.strip()
		if serial_cmpldRegEx.match( ata_id_line ):
			dm_map_name	= "AOE_%s" % ( serial_cmpldRegEx.sub( '', ata_id_line ) )
			break
	if not dm_map_name:
		return None

	# Allright, it exists.. we can ignore it.
	if os.path.exists( "/dev/mapper/%s" % ( dm_map_name ) ):
		return dm_map_name

	# Need to create it.
	size_in_sects, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/block/etherd!%s/size" % ( aoe_disk ) )
	if read_failed:
		return None

	aoe_disk_table = "0 %s linear %s 0" % ( size_in_sects, aoe_devPath )

	vi_errorAndLog_obj.vi_print_log( "Create AOE dm map :: %s %s" % ( aoe_disk, dm_map_name ) )

	cmdToExe = [ 'dmsetup', 'create', dm_map_name, '--table', aoe_disk_table ]
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmdToExe )
	if len( stdout_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
	if len( stderr_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

	if not os.path.exists( "/dev/mapper/%s" % ( dm_map_name ) ):
		vi_errorAndLog_obj.vi_print_wrn( "Error: AOE disk map creation failed : %s %s : %s :: %s" % ( aoe_disk, aoe_interface, dm_map_name, aoe_disk_table ) )
		return dm_map_name

	cmdToExe = [ 'logger', '-s', '-p', 'kern.info', '-t', 'VirtualIron', '--', '{AOE_DISK_FOUND} disk="%s" interface="%s" wwid="%s"' % ( aoe_disk, aoe_interface, dm_map_name ) ]
	vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmdToExe )

####################################################################################################
def task_scan_aoe( vi_errorAndLog_obj, options_OBJ, qualifier_DICT ):
	"""
	This routine
	"""
#
	aoeBlkDevMajor	= storage_tech.get_block_device_major_num( vi_errorAndLog_obj, 'aoe', False )
	if aoeBlkDevMajor < 0:
		vi_errorAndLog_obj.vi_print_log( 'AOE not present. Skipping AOE rescan.' )
		return

	vi_errorAndLog_obj.vi_print_log( global_sectionPreable, "Rescanning AOE" )

	ethFilter_cmpldRegEx = re.compile( '^(eth\d+)$' )
	aoeInterface_LIST    = []

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, 'aoe-interfaces' )
	if len(stdout_LIST) == 0:
		vi_errorAndLog_obj.vi_print_log( "No AOE interfaces.")
		return

	for possibleEth in stdout_LIST[0].strip().split():
		if ethFilter_cmpldRegEx.match( possibleEth ):
			aoeInterface_LIST.append( possibleEth )

	if len( aoeInterface_LIST ) == 0:
		return

	pre_aoe_dm_maps_DICT = get_aoe_dm_maps( vi_errorAndLog_obj, False )

	vi_errorAndLog_obj.vi_print_log( "Rescanning AOE [ %s ]" % ( ' '.join( aoeInterface_LIST ) ) )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, 'aoe-stat' )
	if len( stdout_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ).rstrip() + '\n', None )
	if len( stderr_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ).rstrip() + '\n', None )

	vi_errorAndLog_obj.vi_print_log( 'Flushing all downed AOE disks' )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, 'aoe-flush' )
	if len( stdout_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
	if len( stderr_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

	if storage_tech.udev_idle( vi_errorAndLog_obj, 3, 120 ):
		vi_errorAndLog_obj.vi_print_wrn( "Failed to timeout on udev idle.")

	vi_errorAndLog_obj.vi_print_log( 'Discovering AOE disks' )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, 'aoe-discover' )
	if len( stdout_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
	if len( stderr_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

	if storage_tech.udev_idle( vi_errorAndLog_obj, 6, 120 ):
		vi_errorAndLog_obj.vi_print_wrn( "Failed to timeout on udev idle.")

	vi_errorAndLog_obj.vi_print_log( 'Ensuring AOE disk maps are up to date.' )

	pst_aoe_dm_maps_DICT = {}

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, 'aoe-stat' )
	if len( stdout_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ).rstrip() + '\n', None )
	if len( stderr_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ).rstrip() + '\n', None )
	for line in stdout_LIST:
		aoeDiskInfo_TPL = line.strip().split()
		vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'aoe-revalidate',   aoeDiskInfo_TPL[0] ] )
		dm_map_name = ensure_aoe_disk_map_uptodate( vi_errorAndLog_obj, aoeDiskInfo_TPL[0], aoeDiskInfo_TPL[2] )
		if dm_map_name:
			pst_aoe_dm_maps_DICT[ dm_map_name ] = True

	vi_errorAndLog_obj.vi_print_log( 'Verifying AOE disk maps' )

	for dm_map_name in pre_aoe_dm_maps_DICT:
		if pst_aoe_dm_maps_DICT.has_key( dm_map_name ):
			continue

		vi_errorAndLog_obj.vi_print_log( "Removing AOE dm map :: %s" % ( dm_map_name ) )

		cmdToExe = [ 'logger', '-s', '-p', 'kern.info', '-t', 'VirtualIron', '--', '{AOE_DISK_LOST} wwid="%s"' % ( dm_map_name ) ]
		vi_exe_sync( vi_errorAndLog_obj, cmdToExe )

		cmdToExe = [ 'dmsetup', 'remove', dm_map_name ]
		stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmdToExe )

		if os.path.exists( "/dev/mapper/%s" % ( dm_map_name ) ):
			if len( stdout_LIST ):
				vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ).rstrip() + '\n', None )
			if len( stderr_LIST ):
				vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ).rstrip() + '\n', None )
			vi_errorAndLog_obj.vi_print_wrn( "Error: AOE disk map remove failed : %s" % ( dm_map_name ) )

####################################################################################################
def acquire_driver_name_from_device_link( vi_errorAndLog_obj, devLink_path ):
	"""
	This routine returns the name of the driver that is responsible for the passed in path
	of the device. The driver usually is 'sd' for SATA/SCSI/SAS, 'ata_piix' for IDE,
	'iscsi_tcp' for iSCSI, and so on.
	"""
#
	try:
		devLink          = os.readlink( devLink_path )
		relPCIdevicePath = re.sub( '/host\d+/?.*$', '', devLink )
		absPCIdevicePath = re.sub( '^[../]*', '/sys/', relPCIdevicePath )
		drvrLink         = os.readlink( "%s/driver" % ( absPCIdevicePath ) )
		driverName       = re.sub( '^.*/', '', drvrLink )
	except:
		return ''

	return driverName

####################################################################################################
def extract_qla4xxx_scsiHostInfo( vi_errorAndLog_obj, scsi_host_num ):
	"""
	This routine
	"""
#
	qla4xxx_state_lines, read_failed = vi_utils.vi_read_text_lines( vi_errorAndLog_obj, "/proc/scsi/qla4xxx/%d" % ( scsi_host_num ) )
	if read_failed or len( qla4xxx_state_lines ) < 3:
		return 'QLogic HW ISCSI Initiator'

	scsiHostInfo = "%s\n%s : %s\n%s : FWrev %s\n%s : %s" % ( 'QLogic HW ISCSI Initiator', global_scan_scsi_spacer, qla4xxx_state_lines[0], global_scan_scsi_spacer, qla4xxx_state_lines[1], global_scan_scsi_spacer, qla4xxx_state_lines[2] )

	return scsiHostInfo

####################################################################################################
def scan_scsi_hosts__work_TLP__compare( work_TLP__A, work_TLP__B ):
	"""
	This routine
	"""
#
	if work_TLP__A[0] > work_TLP__B[0]:
		return 1
	if work_TLP__A[0] < work_TLP__B[0]:
		return -1
	return 0

####################################################################################################
def task_scan_scsi_hosts( vi_errorAndLog_obj, options_OBJ, qualifier_DICT ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( global_sectionPreable, "Rescanning SCSI Hosts" )

	work_TLP_LIST = []
	sg_dev_DICT = {}
	for scsi_host in os.listdir( '/sys/class/scsi_host' ):

		if not os.path.exists( "/sys/class/scsi_host/%s/scan" % ( scsi_host ) ):
			continue

		scsi_host_num = int( re.sub( '^host', '', scsi_host ) )

		driverName = acquire_driver_name_from_device_link( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/device" % ( scsi_host ) )
		vi_errorAndLog_obj.vi_print_dbg( 2, "%s :: driverName : %s" % ( scsi_host, driverName ) )

		procName,          read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/proc_name" % ( scsi_host ) )
		vi_errorAndLog_obj.vi_print_dbg( 2, "%s :: procName   : %s" % ( scsi_host, procName ) )

		modelName,         read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/model_name" % ( scsi_host ) )
		if read_failed:
			modelName,     read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/modelname" % ( scsi_host ) )
		vi_errorAndLog_obj.vi_print_dbg( 2, "%s :: modelName  : %s" % ( scsi_host, modelName ) )

		modelDesc,         read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/model_desc" % ( scsi_host ) )
		if read_failed:
			modelDesc,     read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/modeldesc" % ( scsi_host ) )
		vi_errorAndLog_obj.vi_print_dbg( 2, "%s :: modelDesc  : %s" % ( scsi_host, modelDesc ) )

		pciInfo,           read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/pci_info" % ( scsi_host ) )
		if read_failed:
			pciInfo,       read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/info" % ( scsi_host ) )
		vi_errorAndLog_obj.vi_print_dbg( 2, "%s :: pciInfo    : %s" % ( scsi_host, pciInfo ) )

		fwRev,             read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/fw_version" % ( scsi_host ) )
		if read_failed:
			fwRev,         read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/fwrev" % ( scsi_host ) )
		vi_errorAndLog_obj.vi_print_dbg( 2, "%s :: fwRev      : %s" % ( scsi_host, fwRev ) )

		driverVer,         read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/driver_version" % ( scsi_host ) )
		if read_failed:
			driverVer,     read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/%s/lpfc_drvr_version" % ( scsi_host ) )
		vi_errorAndLog_obj.vi_print_dbg( 2, "%s :: driverVer  : %s" % ( scsi_host, driverVer ) )

		knownType = ''
		while True:
			specific_qualifier = qualifier_DICT.get(driverName, None)
			if not specific_qualifier:
				specific_qualifier = qualifier_DICT.get("all_other", None)

			if driverName == 'lpfc':
				knownType          = 'Emulex'
				scsiHostInfo       = "%s\n%s : %s\n%s : FWrev %s\n%s : %s" % ( knownType, global_scan_scsi_spacer, pciInfo, global_scan_scsi_spacer, fwRev, global_scan_scsi_spacer, driverVer )
				break

			if driverName == 'qla2xxx':
				knownType          = 'QLogic'
				scsiHostInfo       = "%s\n%s : %s : %s\n%s : FWrev %s\n%s : %s" % ( knownType, global_scan_scsi_spacer, modelName, pciInfo, global_scan_scsi_spacer, fwRev, global_scan_scsi_spacer, driverVer )
				break

			if driverName == 'qla4xxx':
				knownType          = 'QLogic'
				scsiHostInfo       = extract_qla4xxx_scsiHostInfo( vi_errorAndLog_obj, scsi_host_num )
				break

			if procName == 'iscsi_tcp':
				knownType          = 'OpenISCSI'
				scsiHostInfo       = 'OpenISCSI SW Initiator'
				break

			# and for the rest of the universe.
			if driverName:
				if driverName == procName:
					scsiHostInfo   = "%s" % ( driverName )
				else:
					scsiHostInfo   = "%s :: %s" % ( driverName, procName )
			else:
				if procName:
					scsiHostInfo   = "%s" % ( procName )
				else:
					scsiHostInfo   = "Unknown %s %s" % ( modelName, modelDesc )
			break

		# And now put the three tuples on the work list
		if specific_qualifier:
			work_TLP_LIST.append( ( scsi_host_num, specific_qualifier.mDoScan, scsiHostInfo ) )
		else:
			# Uhhh.. This is bad.
			vi_errorAndLog_obj.vi_print_wrn( "Error: No qualifer for %d" % ( scsi_host_num ) )

	work_TLP_LIST.sort( scan_scsi_hosts__work_TLP__compare )

	for work_TLP in work_TLP_LIST:

		scsi_host_num = work_TLP[ 0 ]
		doScan        = work_TLP[ 1 ]
		scsiHostInfo  = work_TLP[ 2 ]
		if not doScan:
			vi_errorAndLog_obj.vi_print_log( "     Skipping host adapter %2d : %s." % ( scsi_host_num , scsiHostInfo ) )
			continue

		vi_errorAndLog_obj.vi_print_log( "Rescanning host adapter %2d SCSI devices" % ( scsi_host_num ) )
		disconnected_cnt, rescanned_cnt = storage_tech.disconnect_stale_sg_devices_on_host(vi_errorAndLog_obj, scsi_host_num)
		if disconnected_cnt > 0 or rescanned_cnt > 0:
			if storage_tech.udev_idle( vi_errorAndLog_obj, 3, 120 ):
				vi_errorAndLog_obj.vi_print_wrn( "Failed to udev timeout on host adapter's %2d SCSI devices" % ( scsi_host_num ) )

		vi_errorAndLog_obj.vi_print_log( "Rescanning host adapter %2d : %s" % ( scsi_host_num , scsiHostInfo ) )
		vi_utils.vi_write_text_value( vi_errorAndLog_obj, "/sys/class/scsi_host/host%d/scan" % ( scsi_host_num ), '- - -\n', True )

####################################################################################################
def task_perform_udevIdle( vi_errorAndLog_obj, options_OBJ, qualifier_LIST ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( global_sectionPreable, 'Wait for udev idle : 6 300' )
	if storage_tech.udev_idle( vi_errorAndLog_obj, 6, 300 ):
		vi_errorAndLog_obj.vi_print_wrn( "Failed to timeout on udev idle.")

####################################################################################################
def task_perform_lvm( vi_errorAndLog_obj, options_OBJ, qualifier_LIST ):
	"""
	This routine
	"""
#
	vi_errorAndLog_obj.vi_print_log( global_sectionPreable, 'Rescanning LVM' )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'pvscan' ] )
	if len( stdout_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
	if len( stderr_LIST ):
		vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )

	if options_OBJ.skipDom0memoryCushionGuarding:
		stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, [ 'vgchange', '--available', 'ly' ] )
		if len( stdout_LIST ):
			vi_errorAndLog_obj.vi_print_log( ''.join( stdout_LIST ), None )
		if len( stderr_LIST ):
			vi_errorAndLog_obj.vi_print_log( ''.join( stderr_LIST ), None )
	else:
		storage_tech.do_dom0_memory_safe_vgchange( vi_errorAndLog_obj, True, None, False )

####################################################################################################
def task_perform_multipath( vi_errorAndLog_obj, options_OBJ, qualifier_LIST ):
	"""
	This routine checks multipath maps for size discreprancies and if so
	resizes the multipath maps.
	"""
#
	vi_errorAndLog_obj.vi_print_log( global_sectionPreable, 'Rescanning device-mapper multipath maps' )
	if storage_tech.ensure_multipath_maps( vi_errorAndLog_obj):
		vi_errorAndLog_obj.vi_print_wrn( "Failed to ensure multipath maps are up-to-date.")

####################################################################################################
def parse_given_args():
	"""
	This routine
	"""
#
	task_LIST = [	Task_OBJ(	'ScanOpenISCSI'
							,	'Scan the SW open-iscsi initiator'
							,	True
							,	task_scan_openiscsi
							)
				,	Task_OBJ(	'ScanAOE'
							,	'Scan the ATA over ethernet initiator'
							,	True
							,	task_scan_aoe
							)
				,	Task_OBJ(	'ScanSCSI'
							,	'Scan the scsi hosts'
							,	True
							,	task_scan_scsi_hosts
							)
				,	Task_OBJ(	'UdevIdle'
							,	'Perform udev idle before exiting'
							,	True
							,	task_perform_udevIdle
							)
				,	Task_OBJ(	'ScanMultipath'
							,	'Perform multipath refresh before exiting'
							,	True
							,	task_perform_multipath
							)
				,	Task_OBJ(	'ScanLVM'
							,	'Perform LVM refresh before exiting'
							,	True
							,	task_perform_lvm
							)
				]

	qualifier_DICT =	{	'lpfc'		:	Qualifier_OBJ(	'EmulexFC'
														,	'scan of the Emulex FC HBA'
														,	True
														)
						,	'qla2xxx'	:	Qualifier_OBJ(	'QLogicFC'
														,	'scan of the QLogic FC HBA'
														,	True
														)
						,	'qla4xxx'	:	Qualifier_OBJ(	'QLogiciSCSI'
														,	'scan of the QLogic iSCSI HBA'
														,	True
														)
						,	"all_other"	:	Qualifier_OBJ(	'OtherSCSI'
														,	'scan of all other HBAs'
														,	True
														)
						}

	parser = OptionParser()

	parser.add_option("", "--debugLvl", dest="debugLvl", default=0, type="int",
					help="debugLvl message printing to stderr" )

	parser.add_option( "", "--skipDom0memoryCushionGuarding", dest="skipDom0memoryCushionGuarding", default=False, action="store_true",
					help="Skip dom0 memory cushion guarding" )

	def store_value____doALL( option, opt_str, value, parser ):
		for task in task_LIST:
			task.mDoScan = True

	def store_value____skipALL( option, opt_str, value, parser ):
		for task in task_LIST:
			task.mDoScan = False

	parser.add_option( "", "--doALL", dest=None, action="callback", callback=store_value____doALL,
					help="Enables all of the Tasks" )

	parser.add_option( "", "--skipALL", dest=None, action="callback", callback=store_value____skipALL,
					help="Skips all of the Tasks" )

	def store_value____task_do( option, opt_str, value, parser ):
		taskIdx = int(option.dest)
		task_LIST[ taskIdx ].mDoScan = True

	def store_value____task_skip( option, opt_str, value, parser ):
		taskIdx = int(option.dest)
		task_LIST[ taskIdx ].mDoScan = False

	taskIdx = -1
	for task in task_LIST:
		taskIdx = taskIdx + 1

		parser.add_option(	"", "--do%s" % ( task.mArgNameStr ), dest=str(taskIdx), action="callback",
							callback=store_value____task_do,  help=" Enable Task :: %s" % ( task.mHelpStr ) )

		parser.add_option(	"", "--skip%s" % ( task.mArgNameStr ), dest=str(taskIdx), action="callback",
							callback=store_value____task_skip, help="Disable Task :: %s" % ( task.mHelpStr ) )

	def store_value____qualifier_do( option, opt_str, value, parser ):
		qualifier = qualifier_DICT.get(option.dest, None)
		if qualifier:
			qualifier.mDoScan = True

	def store_value____qualifier_skip( option, opt_str, value, parser ):
		qualifier = qualifier_DICT.get(option.dest, None)
		if qualifier:
			qualifier.mDoScan = False

	for key, qualifier in qualifier_DICT.iteritems():
		parser.add_option(	"", "--do%s" % ( qualifier.mArgNameStr ), dest=str(key), action="callback",
							callback=store_value____qualifier_do,  help=" Enable :: %s" % ( qualifier.mHelpStr ) )

		parser.add_option(	"", "--skip%s" % ( qualifier.mArgNameStr ), dest=str(key), action="callback",
							callback=store_value____qualifier_skip, help="Disable :: %s" % ( qualifier.mHelpStr ) )

	options_OBJ, args_LIST = parser.parse_args()

	if 0:
		print >> sys.stderr, options_OBJ
		print >> sys.stderr, args_LIST
		print >> sys.stderr, dir(options_OBJ)
		print >> sys.stderr, parser.values
		print >> sys.stderr, task_LIST
		vi_utils.vi_dbg_stop()

	tasksToPerform_LIST = []
	for task in task_LIST:
		if task.mDoScan:
			tasksToPerform_LIST.append( task )

	if len( tasksToPerform_LIST ) == 0:
		print >> sys.stderr, 'no work requested\n'
		parser.print_help(sys.stderr)
		print >> sys.stderr, ''
		vi_utils.vi_exit( 2 )

	return ( options_OBJ, args_LIST, tasksToPerform_LIST, qualifier_DICT )

####################################################################################################
def init_complex_globals( options_OBJ, args_LIST ):
	"""
	This routine initializes and complex global variables
	"""
#
	global global_errAndLog
	global global_logFilePath

	vi_utils.vi_rototillLogs( global_logDirectory, global_baseLogFileName, 20, 10 )

	newLog_FILE = None
	logFileName = vi_utils.vi_generate_log_filename( global_baseLogFileName )
	global_logFilePath = "%s/%s" % ( global_logDirectory, logFileName )
	try:
		newLog_FILE = open( global_logFilePath, 'a', 0 )		# UNbuffered
	except:
		pass

	errAndLog_dst_DICT =	{	'stdErr'			:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__ERR
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__WRN
														|	vi_utils.VI_ErrAndLog__FOR_FLAG__DBG
														,	sys.stderr
														)
							,	'stdout'			:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__STD
														,	sys.stdout
														)
							,	'log_to_stdout'		:	(	vi_utils.VI_ErrAndLog__FOR_FLAG__LOG
														,	sys.stdout
														)
							}

	if newLog_FILE:
		errAndLog_dst_DICT[ 'log' ]	=	(	vi_utils.VI_ErrAndLog__FOR_FLAG__ERR
										|	vi_utils.VI_ErrAndLog__FOR_FLAG__WRN
										|	vi_utils.VI_ErrAndLog__FOR_FLAG__DBG
										|	vi_utils.VI_ErrAndLog__FOR_FLAG__LOG
										,	newLog_FILE
										)
		errAndLog_dst_DICT[ 'alt' ]	=	(	vi_utils.VI_ErrAndLog__FOR_FLAG__ALT
										,	newLog_FILE
										)

	global_errAndLog = vi_utils.VI_ErrAndLogRedirector_OBJ( errAndLog_dst_DICT, options_OBJ.debugLvl )

	if newLog_FILE:
		global_errAndLog.vi_print_log( "LOG: %s" % ( global_logFilePath ) )
	else:
		global_errAndLog.vi_print_wrn( "could not create global log FILE '%s'" % ( global_logFilePath ) )

####################################################################################################
def main():
	"""
	This routine is the script entry point
	"""
#
	options_OBJ, args_LIST, tasksToPerform_LIST, qualifier_DICT = parse_given_args()

	init_complex_globals( options_OBJ, args_LIST )

	pre_lsscsi_stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( global_errAndLog, 'lsscsi' )
	if len( pre_lsscsi_stdout_LIST ) > 0:
		global_errAndLog.vi_print_alt( ''.join( pre_lsscsi_stdout_LIST ), None )

	for task in tasksToPerform_LIST:
		try:
			task.mFcn( global_errAndLog, options_OBJ, qualifier_DICT )
		except:
#			global_errAndLog.vi_croak_exception( traceback.format_exc() )
			global_errAndLog.vi_print_log( 'Failed to perform: %s \ncontinuing ...\n\n%s' % (task.mHelpStr, traceback.format_exc()) )

	pst_lsscsi_stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( global_errAndLog, 'lsscsi' )
	if len( pst_lsscsi_stdout_LIST ) > 0:
		theDiffer				= difflib.Differ()
		diffResult_LIST			= list( theDiffer.compare( pre_lsscsi_stdout_LIST, pst_lsscsi_stdout_LIST ) )
		undesirableIndexes_LIST	= []
		for idx, diffResult_line in enumerate(diffResult_LIST):
			if diffResult_line[0] != '-' and diffResult_line[0] != '+':
				undesirableIndexes_LIST.append( idx )
		for idx in reversed(undesirableIndexes_LIST):
			del diffResult_LIST[ idx ]
		if len(diffResult_LIST) > 0:
			global_errAndLog.vi_print_log( global_sectionPreable, 'lsscsi DIFFs' )
			global_errAndLog.vi_print_log( ''.join( diffResult_LIST ), None )
	else:
		if len(pre_lsscsi_stdout_LIST) > 0:
			global_errAndLog.vi_print_log( global_sectionPreable, 'lsscsi DIFFs' )
		for line in pre_lsscsi_stdout_LIST:
			global_errAndLog.vi_print_log( '-', line, None )

	global_errAndLog.vi_print_std( global_sectionPreable )
	global_errAndLog.vi_print_std( 'rescan_for_disks Done.' )

	global_errAndLog.close_dst_files()
	vi_utils.vi_exe_sync( global_errAndLog, [ 'gzip', '--best', global_logFilePath ], 30 )

	return 0

####################################################################################################
### main

sys.stdout = os.fdopen( sys.stdout.fileno(), sys.stdout.mode, 0 )	# UNbuffer stdout

exitValue  = main()

sys.exit( exitValue )

# END
####################################################################################################
