#!/usr/bin/python
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for dealing with block devices
"""

import vi_utils

import os
import re
import traceback

from st_ATAoverEthernet			import *
from st_blockDeviceOperations	import *
from st_CCISS					import *
from st_openISCSI				import *
from st_QLogicHBA				import *
from st_scsiDeviceID			import *
from st_loopback				import *
from st_SG_device				import *

####################################################################################################

__all__ = [	'BlkDev_CurState_OBJ'
		,	'BlkDevInfo_OBJ'
		,	'BlkDevStats_OBJ'
		,	'fill_blkDevInfo_AOE_device'
		,	'fill_blkDevInfo_CCISS_device'
		,	'fill_blkDevInfo_non_scsi_device'
		,	'fill_blkDevInfo_scsi_device_basics'
		,	'fill_blkDevInfo_scsi_device'
		,	'get_empty_blkDevInfo'
		,	'get_blkDevInfo'
		,	'get_stats_for_blkDev' ]

####################################################################################################

class BlkDev_CurState_OBJ:

	def __init__( self ):
		self.mBlkDev            = None
		self.mSizeInBytes       = None
		self.mPathDMmapState    = None
		self.mCheckerState      = None

	def is_valid( self ):
		if type( self.mBlkDev )            == None: return False
		if type( self.mSizeInBytes )       == None: return False
		if type( self.mPathDMmapState )    == None: return False
		if type( self.mCheckerState )      == None: return False
		return True

	def __repr__( self ):
		return "mBlkDev: %s mSizeInBytes: %s mPathDMmapState: %s mCheckerState: %s" % \
					(	self.mBlkDev
					,	self.mSizeInBytes
					,	self.mPathDMmapState
					,	self.mCheckerState )

class BlkDevInfo_OBJ:

	def __init__( self ):
		self.mBlkDev               = None
		self.mDevPath              = None
		self.mWWID                 = None
		self.mOldWWID              = None	# ONLY set if the WWID has change in this release !!!
		self.mVendor               = None
		self.mProduct              = None
		self.mRev                  = None
		self.mHasExplicitMPconf    = False
		self.mIsWWIDgenerated      = False
		self.mIsSCSIdevice         = None
		self.mDevNums              = None
		self.mSCSIdevID            = None # optional SCSI_device_ID_OBJ()
		self.mDriver               = None
		self.mKind                 = None
		self.mTransport            = None
		self.mTarget               = None # for FC this is the HBA nr: '0x12...' for iSCSI this is the target IQN
		self.mTargetIP			   = None
		self.mTargetPort           = None # for FC this is the HBA's port ID. For iSCSI target's port (usually 3260)
		self.mTargetIQN			   = None # the same as the mTarget for iSCSI
		self.mTargetTPGT		   = -1 # 'int' value.
		self.mSpecificTransport    = None
		self.mCurState             = None # BlkDev_CurState_OBJ()
		self.mOperations           = None # optional BlkDev_Operations subclass
		self.mSGOperations		   = None # optional SG_device_Operations class.
		self.mDevMapNum		   = None # 

	def is_valid( self ):
		if type( self.mBlkDev )               == None: return False
		if type( self.mDevPath )              == None: return False
		if type( self.mDevNums )              == None: return False
		if type( self.mWWID )                 == None: return False
		if type( self.mVendor )               == None: return False
		if type( self.mProduct )              == None: return False
		if type( self.mRev )                  == None: return False
		if type( self.mIsSCSIdevice )         == None: return False
		if self.mIsSCSIdevice and not self.mSCSIdevID.is_valid(): return False
		if type( self.mDriver )               == None: return False
		if type( self.mKind )                 == None: return False
		if type( self.mTransport )            == None: return False
		if type( self.mTarget )               == None: return False
		if type( self.mTargetIP )             == None: return False
		if type( self.mTargetPort )           == None: return False
		if type( self.mTargetIQN )            == None: return False
		if type( self.mTargetTPGT )           == None: return False
		if type( self.mSpecificTransport )    == None: return False
		if type( self.mCurState )             == None: return False
		if not self.mCurState.is_valid(): return False
		return True

	def __repr__( self ):
		return "mBlkDev: %s mDevPath: %s mWWID: %s mDevNums: %s mVendor: %s mProduct: %s mRev: %s mIsSCSIdevice: %s mSCSIdevID: %s mDriver: %s mKind: %s mTransport: %s mTarget: %s mTargetIP: %s mTargetPort: %s mTargetIQN: %s mTargetTPGT: %d mSpecificTransport: %s mCurState: %s mOperations: %s mSGOperations: %s" % \
					(	self.mBlkDev
					,	self.mDevPath
					,	self.mWWID
					,	self.mDevNums
					,	self.mVendor
					,	self.mProduct
					,	self.mRev
					,	self.mIsSCSIdevice
					,	self.mSCSIdevID
					,	self.mDriver
					,	self.mKind
					,	self.mTransport
					,	self.mTarget
					,	self.mTargetIP
					,	self.mTargetPort
					,	self.mTargetIQN
					,	self.mTargetTPGT
					,	self.mSpecificTransport
					,	self.mCurState
					,	self.mOperations
					,	self.mSGOperations )

	def __cmp__( self, other ):
		if not isinstance( other, BlkDevInfo_OBJ ):
			return -1

		if not self.mSCSIdevID or not other.mSCSIdevID:
			if self.mSCSIdevID:
				return 1
			if other.mSCSIdevID:
				return -1
			return cmp( self.mBlkDev, other.mBlkDev )

		if self.mSCSIdevID.mHost > other.mSCSIdevID.mHost:
			return  1
		if self.mSCSIdevID.mHost < other.mSCSIdevID.mHost:
			return  -1
		if self.mSCSIdevID.mChannel > other.mSCSIdevID.mChannel:
			return  1
		if self.mSCSIdevID.mChannel < other.mSCSIdevID.mChannel:
			return  -1
		if self.mSCSIdevID.mSCSIid > other.mSCSIdevID.mSCSIid:
			return  1
		if self.mSCSIdevID.mSCSIid < other.mSCSIdevID.mSCSIid:
			return  -1
		if self.mSCSIdevID.mLUN > other.mSCSIdevID.mLUN:
			return  1
		if self.mSCSIdevID.mLUN < other.mSCSIdevID.mLUN:
			return  -1

		return 0

class BlkDevStats_OBJ:

	def __init__( self, statDataLine ):
		statData_LIST		= statDataLine.split()
		self.mReadIOs		= int(statData_LIST[ 0])
		self.mReadMerges	= int(statData_LIST[ 1])
		self.mReadSectors	= int(statData_LIST[ 2])
		self.mReadTicks		= int(statData_LIST[ 3])
		self.mWriteIOs		= int(statData_LIST[ 4])
		self.mWriteMerges	= int(statData_LIST[ 5])
		self.mWriteSectors	= int(statData_LIST[ 6])
		self.mWriteTicks	= int(statData_LIST[ 7])
		self.mInFlight		= int(statData_LIST[ 8])
		self.mIOtisks		= int(statData_LIST[ 9])
		self.mTimeInQueue	= int(statData_LIST[10])

	def get_tagVal_TPL_LIST( self ):
		return	[	(	'Read_IOs',			mReadIOs		)
				,	(	'Read_Merges',		mReadMerges		)
				,	(	'Read_Sectors',		mReadSectors	)
				,	(	'Read_Ticks',		mReadTicks		)
				,	(	'Write_IOs',		mWriteIOs		)
				,	(	'Write_Merges',		mWriteMerges	)
				,	(	'Write_Sectors',	mWriteSectors	)
				,	(	'Write_Ticks',		mWriteTicks		)
				,	(	'In_Flight',		mInFlight		)
				,	(	'IO_Ticks',			mIOtisks		)
				,	(	'Time_In_Queue',	mTimeInQueue	)
				]

	def __repr__( self ):
		return str(self.get_tagVal_TPL_LIST())

####################################################################################################
def fill_blkDevInfo_AOE_device( vi_errorAndLog_obj, blkDevInfo, aoe_disk ):
	"""
	This routine
	"""
#
	pathDMmapState		= 'failed'
	if os.path.exists( "/dev/mapper/%s" % ( aoe_disk.mWWID ) ):
		pathDMmapState	= 'active'
	checkerState		= 'faulty'
	if aoe_disk.mCurState.mInterfaceStatus == 'up':
		checkerState	= 'ready'

	blkDevInfo.mDevPath						= aoe_disk.mDevPath
	blkDevInfo.mVendor						= aoe_disk.mModel
	blkDevInfo.mProduct						= 'ATAoverEthernet'
	blkDevInfo.mRev							= aoe_disk.mFirmwareRev
	blkDevInfo.mIsSCSIdevice				= False
	blkDevInfo.mSCSIdevID					= ''
	blkDevInfo.mDriver						= 'aoe'
	blkDevInfo.mKind						= 'SAN'
	blkDevInfo.mTransport					= 'AOE'
	blkDevInfo.mTarget						= ''
	blkDevInfo.mTargetPort					= ''
	blkDevInfo.mTargetIP					= ''
	blkDevInfo.mTargetIQN					= ''
	blkDevInfo.mTargetTPGT					= -1
	blkDevInfo.mSpecificTransport			= aoe_disk.mInterface
	blkDevInfo.mCurState.mPathDMmapState	= pathDMmapState
	blkDevInfo.mCurState.mCheckerState		= checkerState

	blkDevInfo.mOperations					= AOE_BlkDev_Operations( blkDevInfo, aoe_disk )

	return 0

####################################################################################################
def fill_blkDevInfo_CCISS_device( vi_errorAndLog_obj, blkDevInfo, refresh ):
	"""
	This routine
	"""
#
	cciss_disk	= find_particular_cciss_disk( vi_errorAndLog_obj, blkDevInfo.mBlkDev, refresh )
	if not cciss_disk:
		return -1

	blkDevInfo.mDevPath						= cciss_disk.mDevPath
	blkDevInfo.mProduct						= cciss_disk.mModel
	blkDevInfo.mRev							= cciss_disk.mFirmwareRev
	blkDevInfo.mIsSCSIdevice				= False
	blkDevInfo.mSCSIdevID					= ''
	blkDevInfo.mDriver						= 'cciss'
	blkDevInfo.mKind						= 'LOCAL'
	blkDevInfo.mTransport					= 'SAS_cciss'
	blkDevInfo.mTarget						= ''
	blkDevInfo.mTargetPort					= ''
	blkDevInfo.mTargetIP					= ''
	blkDevInfo.mTargetIQN					= ''
	blkDevInfo.mTargetTPGT					= -1
	blkDevInfo.mSpecificTransport			= cciss_disk.mInterface
	blkDevInfo.mOperations					= CCISS_BlkDev_Operations( blkDevInfo, cciss_disk )

	blkDevInfo.mOldWWID						= cciss_disk.mOldWWID	# changed in Sodium

	return 0

####################################################################################################
def fill_blkDevInfo_loopback_device( vi_errorAndLog_obj, blkDevInfo, loopback_disk ):
	"""
	This routine fills out a BlkDevInfo_OBJ with information that a loopback device would contain.
	"""
#
	blkDevInfo.mDevPath						= loopback_disk.mDevPath
	blkDevInfo.mVendor						= 'loopback'
	blkDevInfo.mProduct						= loopback_disk.mSource
	blkDevInfo.mRev							= '1.0'
	blkDevInfo.mIsSCSIdevice				= False
	blkDevInfo.mSCSIdevID					= ''
	blkDevInfo.mDriver						= 'loopback'
	blkDevInfo.mKind						= 'LOCAL'
	blkDevInfo.mTransport					= 'loopback'
	blkDevInfo.mTarget						= ''
	blkDevInfo.mTargetIP					= ''
	blkDevInfo.mTargetPort					= ''
	blkDevInfo.mTargetIQN					= ''
	blkDevInfo.mTargetTPGT					= -1
	blkDevInfo.mSpecificTransport			= 'loopback'

	blkDevInfo.mOperations					= LoopBack_BlkDev_Operations( blkDevInfo, loopback_disk )

	return 0
####################################################################################################
def fill_blkDevInfo_non_scsi_device( vi_errorAndLog_obj, blkDevInfo, refresh ):
	"""
	This routine
	"""
#
	aoe_disk_LIST = find_aoe_disks( vi_errorAndLog_obj, refresh )
	for aoe_disk in aoe_disk_LIST:
		if aoe_disk.mBlkDev != blkDevInfo.mBlkDev:
			continue
		return fill_blkDevInfo_AOE_device( vi_errorAndLog_obj, blkDevInfo, aoe_disk )

	loop_disk_LIST = find_loopback_disks ( vi_errorAndLog_obj, refresh )
	for loop_disk in loop_disk_LIST:
		if loop_disk.mBlkDev != blkDevInfo.mBlkDev:
			continue
		return fill_blkDevInfo_loopback_device ( vi_errorAndLog_obj, blkDevInfo, loop_disk )

	vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' unknown non_scsi device %s !?!" % ( blkDevInfo.mBlkDev, blkDevInfo ) )
	return 1

####################################################################################################
def fill_blkDevInfo_scsi_device_basics( vi_errorAndLog_obj, blkDevInfo, refresh ):
	"""
	This routine
	"""
#
	blkdev_info_path = "/sys/block/%s/device/vendor" % ( blkDevInfo.mBlkDev )
	blkDevInfo.mVendor, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
#	if read_failed:
#		vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkDevInfo.mBlkDev, blkdev_info_path ) )
#		return 1

	blkdev_info_path = "/sys/block/%s/device/model" % ( blkDevInfo.mBlkDev )
	blkDevInfo.mProduct, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
#	if read_failed:
#		vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkDevInfo.mBlkDev, blkdev_info_path ) )
#		return 1

	blkdev_info_path = "/sys/block/%s/device/rev" % ( blkDevInfo.mBlkDev )
	blkDevInfo.mRev, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
#	if read_failed:
#		vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkDevInfo.mBlkDev, blkdev_info_path ) )
#		return 1

	return 0

####################################################################################################
def fill_blkDevInfo_scsi_device( vi_errorAndLog_obj, blkDevInfo, refresh ):
	"""
	This routine
	"""

#
	blkDevInfo.mIsSCSIdevice			= True

	blkdev_info_path					= "/sys/block/%s/device" % ( blkDevInfo.mBlkDev )
	try:
		blkdev_devLink					= os.readlink( blkdev_info_path )
	except:
		vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkDevInfo.mBlkDev, blkdev_info_path ) )
		return 1

	try:
		theSCSIid						= re.sub( '^.*/', '', blkdev_devLink )
		blkDevInfo.mSCSIdevID			= gen_SCSI_device_ID_OBJ_from_str(theSCSIid)
		if not blkDevInfo.mSCSIdevID:
			vi_errorAndLog_obj.vi_print_err( "blkdev '%s' bad devLink [%s] !?!" % ( blkDevInfo.mBlkDev, blkdev_devLink ) )
			return 1

		targetID		= "target%s:%s:%s" % ( blkDevInfo.mSCSIdevID.mHost, blkDevInfo.mSCSIdevID.mChannel, blkDevInfo.mSCSIdevID.mSCSIid  )
		devLink_match	= re.match( "[\./]*(.*)/host%s/(.*)%s/%s$" % ( blkDevInfo.mSCSIdevID.mHost, targetID, blkDevInfo.mSCSIdevID ), blkdev_devLink )

		blkdev_info_path = '/sys/' + devLink_match.group( 1 ) + '/driver'
		if os.path.exists( blkdev_info_path ):
			blkdev_drvrLink = os.readlink( blkdev_info_path )
			blkDevInfo.mDriver = re.sub( '^.*/', '', blkdev_drvrLink )
		else:
			blkDevInfo.mDriver = '???'

		blkDevInfo.mKind                 = 'LOCAL'
		blkDevInfo.mTransport            = 'LOCAL'
		blkDevInfo.mTarget               = ''
		blkDevInfo.mTargetIP             = ''
		blkDevInfo.mTargetPort           = ''
		blkDevInfo.mTargetIQN            = ''
		blkDevInfo.mTargetTPGT           = -1
		blkDevInfo.mSpecificTransport    = ''
		specificTransport                = devLink_match.group( 2 ).rstrip( '/' )

		if len(specificTransport) > 0:

			specificTransport_match = re.match( '^session(\d+)$', specificTransport )
			if specificTransport_match:
				iscsi_sessionNum              = int( specificTransport_match.group( 1 ) )

				blkDevInfo.mDriver            = 'iscsi_tcp'
				blkDevInfo.mKind              = 'SAN'
				blkDevInfo.mTransport         = 'iSCSI'
				blkDevInfo.mSpecificTransport = specificTransport

				openiscsi_session_info_DICT	= get_openiscsi_session_info_dict( vi_errorAndLog_obj, False )
				openISCSIstate				= openiscsi_session_info_DICT[ iscsi_sessionNum ]

				blkDevInfo.mTargetPort		= str(openISCSIstate.mTargetPort)
				blkDevInfo.mTarget			= openISCSIstate.mIQN
				blkDevInfo.mTargetIP		= openISCSIstate.mTargetIP
				blkDevInfo.mTargetIQN		= openISCSIstate.mIQN
				blkDevInfo.mTargetTPGT		= openISCSIstate.mTpgt
				blkDevInfo.mOperations		= openISCSI_BlkDev_Operations( blkDevInfo, openISCSIstate )

			specificTransport_match = re.match( "^rport-%s:%s-(\d+)$" % ( blkDevInfo.mSCSIdevID.mHost, blkDevInfo.mSCSIdevID.mChannel ), specificTransport )
			if specificTransport_match:
				rportNum                      = specificTransport_match.group( 1 )
				blkDevInfo.mKind              = 'SAN'
				blkDevInfo.mTransport         = 'FC'
				blkDevInfo.mSpecificTransport = specificTransport
				blkDevInfo.mTargetIQN		  = ''
				blkDevInfo.mTargetTPGT		  = -1
				blkdev_info_path = "/sys/class/fc_transport/%s/node_name" % ( targetID )
				blkDevInfo.mTarget, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
				if read_failed:
					vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkDevInfo.mBlkDev, blkdev_info_path ) )
					return 1

				blkdev_info_path = "/sys/class/fc_transport/%s/port_name" % ( targetID )
				blkDevInfo.mTargetPort, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
				if read_failed:
					vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkDevInfo.mBlkDev, blkdev_info_path ) )
					return 1

			# HBA->SAS disk are 'end_device-<SCSI Host Nr>:<Port ID>'. This is non-SAS SAN setup
			# HBA->SAS Expander->SAS disk are 'end_device-<SCSI Host Nr>:<SCSI Id>:<Port Id>'. This is SAS SAN setup.
			specificTransport_match = re.match( '.*end_device-(\d+):(\d+):(\d+)$', specificTransport )
			if specificTransport_match:
				end_device			= re.sub( '^.*end_device-', 'end_device-', specificTransport )
				blkdev_info_path	= "/sys/class/sas_device/%s/enclosure_identifier" % (end_device)
				blkDevInfo.mTarget, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
				if not read_failed:
					# Local HBA's address:
					blkdev_info_path	= "/sys/class/sas_device/%s/sas_address" % (end_device)
					blkDevInfo.mTargetPort, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
					if not read_failed:
						# If we fail with setting those values, we will just fail-back to the LOCAL storage view.
						blkDevInfo.mKind              = 'SAN'
						blkDevInfo.mTransport         = 'SAS'
						blkDevInfo.mSpecificTransport = specificTransport
						blkDevInfo.mTargetIQN		  = ''
						blkDevInfo.mTargetTPGT		  = -1

		else:
			if blkDevInfo.mDriver == 'qla4xxx':
				blkDevInfo.mKind				= 'SAN'
				blkDevInfo.mTransport			= 'iSCSI'
				blkDevInfo.mSpecificTransport	= 'HW_iSCSI'
				try:
					qla4xxx_state_info_DICT		= get_qla4xxx_info_dict( vi_errorAndLog_obj, refresh )
					qla4xxx_state				= qla4xxx_state_info_DICT[ blkDevInfo.mSCSIdevID.mSCSIid ]
					blkDevInfo.mTargetPort		= str(qla4xxx_state.mTargetPort)
					blkDevInfo.mTarget			= qla4xxx_state.mIQN
					blkDevInfo.mTargetIP		= qla4xxx_state.mTargetIP
					blkDevInfo.mTargetIQN		= qla4xxx_state.mIQN
					blkDevInfo.mTargetTPGT		= qla4xxx_state.mTpgt
				except KeyError:
					pass

#		vi_errorAndLog_obj.vi_print_dbg( 0, str(blkDevInfo.mSCSIdevID) + ' ' + blkDevInfo.mSpecificTransport + ' ' + blkDevInfo.mDriver + ' ' + blkDevInfo.mTransport + ' ' + blkDevInfo.mTarget + ' ' + blkDevInfo.mTargetPort )

	except:
#		vi_errorAndLog_obj.vi_croak_exception( traceback.format_exc() )
		vi_errorAndLog_obj.vi_print_err( "blkdev '%s' bad devLink [%s] !?!\n%s" % ( blkDevInfo.mBlkDev, blkdev_devLink, traceback.format_exc() ) )
		return 1

	if not blkDevInfo.mOperations:
		blkDevInfo.mOperations	= Std_BlkDev_Operations( blkDevInfo )

	return 0

####################################################################################################
def get_empty_blkDevInfo( vi_errorAndLog_obj, blkdev ):
	"""
	This routine
	"""
#
	blkdev_info_path = "/sys/block/%s" % ( blkdev )
	if not os.path.exists( blkdev_info_path ):
		return None

	blkDevInfo						= BlkDevInfo_OBJ()
	blkDevInfo.mBlkDev				= blkdev
	blkDevInfo.mDevPath				= "/dev/%s" % ( blkdev )
	blkDevInfo.mCurState			= BlkDev_CurState_OBJ()
	blkDevInfo.mCurState.mBlkDev	= blkdev

	blkdev_info_path = "/sys/block/%s/dev" % ( blkdev )
	blkDevInfo.mDevNums, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
	if read_failed:
		vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkdev, blkdev_info_path ) )
		return None

	blkdev_info_path = "/sys/block/%s/size" % ( blkdev )
	blkdev_sizeInSectors, read_failed = vi_utils.vi_read_text_value( vi_errorAndLog_obj, blkdev_info_path )
	if read_failed:
		vi_errorAndLog_obj.vi_print_wrn( "blkdev '%s' vanished [%s] !?!" % ( blkdev, blkdev_info_path ) )
		return None

	blkDevInfo.mCurState.mSizeInBytes	= int( blkdev_sizeInSectors ) * 512

	return blkDevInfo

####################################################################################################
def get_blkDevInfo( vi_errorAndLog_obj, blkdev, blkdevWWID, stripOffOperations, refresh ):
	"""
	This routine retrieves fully populated BlkDevInfo_OBJ object.
	"""
#
	blkDevInfo			= get_empty_blkDevInfo( vi_errorAndLog_obj, blkdev )
	if not blkDevInfo:
		return None

	blkDevInfo.mWWID	= blkdevWWID

	blkdev_info_path	= "/sys/block/%s/device" % ( blkdev )

    

	if not os.path.exists( blkdev_info_path ):
		if fill_blkDevInfo_non_scsi_device( vi_errorAndLog_obj, blkDevInfo, refresh ):
			return None
	else:
		if fill_blkDevInfo_scsi_device_basics( vi_errorAndLog_obj, blkDevInfo, refresh ):
			return None
		if fill_blkDevInfo_scsi_device( vi_errorAndLog_obj, blkDevInfo, refresh ):
			return None

	if stripOffOperations:
		blkDevInfo.mOperations	= None

	if 0 and not blkDevInfo.is_valid():
		vi_errorAndLog_obj.vi_print_err( "INTERNAL ERROR: blkdev '%s' invalid info %s" % ( blkdev, blkDevInfo ) )

	return blkDevInfo

####################################################################################################
def get_stats_for_blkDev( vi_errorAndLog_obj, blkDev ):
	"""
	This routine
	"""
#
	statPath	= "/sys/block/%s/stat" % ( blkDev )
	if not os.path.exists( statPath ):
		return None

	statDataLine, read_failed	= vi_utils.vi_read_text_value( vi_errorAndLog_obj, statPath )
	if read_failed:
		return None

	try:
		blkDevStats	= BlkDevStats_OBJ( statDataLine )
	except:
		vi_errorAndLog_obj.vi_print_err( "INTERNAL ERROR: bad statDataLine '%s' for %s" % ( statDataLine, blkdev ) )
		return None

	return blkDevStats

####################################################################################################
# END
