#!/usr/bin/python
####################################################################################################
#
#	Author:		Joshua Nicholas, Konrad Rzeszutek
#
"""
	Utilities for network specific operations.
"""

import os
import sys
import time
import re
import errno
import shlex

import vi_utils

from vi_misc		import *

####################################################################################################

__all__ = [	'NIC_state_OBJ'
		,	'Network_Route_OBJ'
		,	'vi_ping_test'
		,	'vi_extract_IP_addr'
		,	'vi_get_IP_addr_dict'
		,	'vi_get_MAC_addr_dict'
		,	'vi_get_NIC_dict'
		,	'vi_get_NIC_state'
		,	'vi_wait_for_NIC_RUNNING'
		,	'vi_reset_NIC_and_wait_for_RUNNING'
		,	'vi_terminate_dhcp'
		,	'vi_invoke_dhcp'
		,	'vi_get_route_list'
		,	'vi_is_subnet_for_IP_present'
		,	'vi_is_route_to_IP_present'
		,	'vi_get_route_list_for_IP'
		,	'vi_add_route_for_IP'
		,	'vi_del_route_for_IP'
		,	'vi_ensure_route_for_IP']

####################################################################################################
def _INTERNAL__ntoa_ipv4( val ):
	"""
	This routine
	"""
#
	if type(val) == int or type(val) == long:
		return "%s.%s.%s.%s" % ( val >> 24, (val & 0xff0000) >> 16, (val & 0xff00) >> 8, val & 0xff )

	return "%X" % ( val )

####################################################################################################

class NIC_state_OBJ:

	def __init__(self):
		self.mMAC				= None
		self.mInterface			= None
		self.mIsUP				= False
		self.mIsRUNNING			= False
		self.mMTU				= 1500
		self.mIPv4_addr			= None
		self.mIPv4_Bcast_addr	= None
		self.mIPv4_netmask		= None
		self.mIPv6_addr			= None

	def __repr__(self):
		return "mMAC: %s mInterface: %s mIsUP: %s mIsRUNNING: %s mMTU: %d mIPv4_addr: %s mIPv4_Bcast_addr: %s mIPv4_netmask: %s mIPv6: %s" % (
				self.mMAC
			,	self.mInterface
			,	self.mIsUP
			,	self.mIsRUNNING
			,	self.mMTU
			,	self.mIPv4_addr
			,	self.mIPv4_Bcast_addr
			,	self.mIPv4_netmask
			,	self.mIPv6_addr )

####################################################################################################

class Network_Route_OBJ:

	def __init__(self):
		self.mInterface		= None
		self.mIPv4_dest		= None
		self.mIPv4_netmask	= None
		self.mIPv4_gateway	= None

	def get_IPv4_dest_str(self):
		return _INTERNAL__ntoa_ipv4( self.mIPv4_dest )

	def get_IPv4_netmask_str(self):
		return _INTERNAL__ntoa_ipv4( self.mIPv4_netmask )

	def get_IPv4_gateway_str(self):
		return _INTERNAL__ntoa_ipv4( self.mIPv4_gateway )

	def __str__(self):
		return "mInterface: %s, mIPv4_dest: %X, mIPv4_gateway: %X, mIPv4_netmask: %X" % (
				self.mInterface
			,	self.mIPv4_dest
			,	self.mIPv4_gateway
			,	self.mIPv4_netmask )

	def __repr__(self):
		if self.mIPv4_netmask or self.mIPv4_gateway:
			return "%-16s %-16s %-16s %-4s" % ( self.get_IPv4_dest_str(), self.get_IPv4_gateway_str(), self.get_IPv4_netmask_str(), self.mInterface )
		return "Can't do IPv6 yet."

	def is_it_this_dest(self, IPv4_addr):
		"""
		Returns True if this route is for this destination.
		"""

		q = IPv4_addr.split(".")
		if len(q)==1: # It is IPv6..need to make it work.
			return False

		# Possibly get rid of lambda. Think about IPv6..
		ip_in_hex = reduce(lambda a,b: long(a)*256 + long(b), q)
		return self.mIPv4_dest == ip_in_hex

	def __cmp__(self, other):
		"""
		Overloading of standard __cmp__ function.
		"""
		if not isinstance( other, Network_Route_OBJ ):
			return -1
		if (self.mInterface < other.mInterface):
			return -1
		if (self.mIPv4_dest < other.mIPv4_dest):
			return -1
		if (self.mIPv4_netmask < other.mIPv4_netmask):
			return -1
		if (self.mInterface > other.mInterface):
			return 1
		if (self.mIPv4_dest > other.mIPv4_dest):
			return 1
		if (self.mIPv4_netmask > other.mIPv4_netmask):
			return 1
		return 0

####################################################################################################

DHCPCD_SCRIPT	=	"/etc/init.d/network_scripts/dhcpcd-hook"
DHCPCD_DEBUG	=	"-d"
DHCPCD_ARGS		=	"%s -t 999999 -C -N -Y -R -G -T -c %s" % ( DHCPCD_DEBUG, DHCPCD_SCRIPT )
DHCPCD_CONFIG	=	'/var/lib/dhcpcd'

####################################################################################################
def vi_ping_test(errAndLog_obj, ip, retries):
	"""
	Pings the specified host once per second. It will retry up to 'retries'
	if the ping is unsucessfull. If it gets a ping, it exits immediately with
	zero return code.
	Returns:
	Zero on success.
	"""
#
	while (retries):
		args = "ping -c %d %s" % (1, ip)
   		stdout_LIST, stderr_LIST, exitStatus, deltaTime = \
 			vi_utils.vi_exe_sync( errAndLog_obj, args, 2 )

		if exitStatus == 0:
			break

		retries = retries - 1
		time.sleep( 1.0 )

	del stdout_LIST
	del stderr_LIST
	del deltaTime

	return exitStatus

####################################################################################################
def vi_extract_IP_addr(IP_addr_str):
	"""
	We sanitize the input argument to remove extranous data such as:
	the port, "127.0.0.1:80" -> <"127.0.0.1","80">; invalid names, "192.168.0.300" -> "";
	(NOT YET)or handle IPv6 names, "fe80::200:ff:fe00:0/64" -> "fe80:0000:0200:00ff:fe00:0000"

	Arguments:
	 The IP address as a string.

	Returns:
	  None if unsuccesfull. Otherwise the proper IP address as string.
	"""

	IPv4_cmplRegEx = re.compile("\s*(\d+)\.(\d+)\.(\d+)\.(\d+)(:\d*|\s*)")

	returnIP = ""
	returnPort = ""
	IPv4_match = IPv4_cmplRegEx.match(IP_addr_str)
	if IPv4_match:
		for i,match in enumerate(IPv4_match.groups()):
			if (i==4):
				returnPort += re.sub(":","", match)
				break
			if (int(match) > 255) or (int(match) < 0):
				return None,None
			returnIP += match
			# 3 is the last tuple.
			if (i != 3):
				returnIP += '.'
		return returnIP, returnPort
	# .. and here we will handle IPv6...
	#IPv6_cmplRegEx = re.compile(".*(\S*):(\S*):(\S*):(\S*):(\S*):(\S*)")
	return None,None

####################################################################################################

global__cached__NIC__pattern_LIST	= None
global__cached__NIC_DICT			= None
global__cached__IP_to_NIC_DICT		= None
global__cached__MAC_to_NIC_DICT		= None
global__cached__IPv4_route_LIST		= None

####################################################################################################
def vi_get_NIC_dict( errAndLog_obj, all_interfaces, refresh ):
	"""
	Retrieves a dictionary of interfaces (key) and it's associated data as a NIC_state_OBJ.
	Example:
	'eth4' -> [NIC_state_obj]

	Returns:
	 a dictionary with the key value being the interface and the value the NIC_state_OBJ

	"""

	# bond0     Link encap:Ethernet  HWaddr 00:00:00:00:00:00
	#          BROADCAST MASTER MULTICAST  MTU:1500  Metric:1
	# and its bond:
	# eth6      Link encap:Ethernet  HWaddr 00:15:17:7B:BE:06
	#          UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
	#
	# IPv4 NIC:
	# eth4      Link encap:Ethernet  HWaddr 00:15:17:7B:BE:04
	#          inet addr:10.1.33.252  Bcast:10.1.33.255  Mask:255.255.255.0
	# LO NIC:
	# lo        Link encap:Local Loopback 
	#          inet addr:127.0.0.1  Mask:255.0.0.0

	# IPv6 NIC (We DO NOT SUPPORT IT YET):
	# eth0     Link encap:Ethernet  HWaddr 00:13:72:27:F4:A6
	#          inet6 addr: fe80::213:72ff:fe27:f4a6/64 Scope:Link
	def __handle_MAC( NIC_state_obj, val):

		NIC_state_obj.mMAC = str(val)

	def __handle_UP( NIC_state_obj, val):

		NIC_state_obj.mIsUP	= True

	def __handle_RUNNING( NIC_state_obj, val):

		NIC_state_obj.mIsRUNNING	= True

	def __handle_MTU(NIC_state_obj, val):

		NIC_state_obj.mMTU = int(val)

	def __handle_IPv4_addr( NIC_state_obj, val):

		NIC_state_obj.mIPv4_addr = str(val)

	def __handle_IPv4_Bcast_addr(NIC_state_obj, val):

		NIC_state_obj.mIPv4_Bcast_addr = str(val)

	def __handle_IPv4_netmask(NIC_state_obj, val):

		NIC_state_obj.mIPv4_netmask = str(val)

	def __handle_IPv6_addr(NIC_state_obj, val):

		NIC_state_obj.mIPv6_addr = str(val)

	########

	global global__cached__NIC_DICT
	global global__cached__NIC__pattern_LIST

	if not refresh and global__cached__NIC_DICT and not all_interfaces:
		return global__cached__NIC_DICT

	if not global__cached__NIC__pattern_LIST:
		pattern_DICT = {	".*\sHWaddr\s(\S+)"				:	__handle_MAC
				,			".*\s(UP)"						:	__handle_UP
				,			".*\s(RUNNING)"					:	__handle_RUNNING
				,			".*\sMTU:(\d+)"					:	__handle_MTU
				,			".*\sinet addr:(\S+)"			:	__handle_IPv4_addr
				,			".*\sBcast:(\S+)"				:	__handle_IPv4_Bcast_addr
				,			".*\sMask:(\S+)"				:	__handle_IPv4_netmask
				,			".*\sinet6 addr: (\S*) Scope"	:	__handle_IPv6_addr
				}
		# Construct the patterns
		global__cached__NIC__pattern_LIST = []
		for pat, handler in pattern_DICT.iteritems():
			global__cached__NIC__pattern_LIST.append((re.compile(pat), handler))

	NIC_RegEx = re.compile("^(\S+)\s+Link.*$")

	cmd_LIST	= [ 'ifconfig' ]
	if all_interfaces:
		cmd_LIST.append( '-a' )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( errAndLog_obj, cmd_LIST )

	interface_DICT		= {}
	cur_NIC_state		= None
	for input_line in stdout_LIST:
		errAndLog_obj.vi_print_dbg( 5, "L: [%s]" % ( input_line ) )
		NIC_match = NIC_RegEx.match(input_line)
		if NIC_match:
			cur_NIC_state = NIC_state_OBJ()
			cur_NIC_state.mInterface = NIC_match.group(1)
			interface_DICT[  cur_NIC_state.mInterface ] = cur_NIC_state
			errAndLog_obj.vi_print_dbg( 5, "I: %s" % (  cur_NIC_state.mInterface ))

		# .. and for the rest we do the matching if there is a current object.
		if cur_NIC_state:
			for pat, handler in global__cached__NIC__pattern_LIST:
				m = pat.match(input_line)
				if m:
					if callable(handler):
						data	= m.group(1)
						errAndLog_obj.vi_print_dbg( 5, "P: %s" % (  data ))
						handler(cur_NIC_state, data)
		# nothing matched.. ingest the next data

	if all_interfaces:
		return interface_DICT

	global__cached__NIC_DICT = interface_DICT
	return global__cached__NIC_DICT

####################################################################################################
def vi_get_NIC_state( vi_errorAndLog_obj, all_interfaces, givenInterfaceOrMACtoVerify ):
	"""
	This routine
	"""
#
	interface_DICT			= vi_get_NIC_dict( vi_errorAndLog_obj, all_interfaces, True )

	interface_cmpldRegEx	= re.compile( '^(eth|bond)(\d+)$' )
	interface_match			= interface_cmpldRegEx.match( givenInterfaceOrMACtoVerify )
	if interface_match:
		nic_state			= interface_DICT.get( givenInterfaceOrMACtoVerify, None )
		if nic_state:
			return nic_state

		interface_DICT	= vi_get_NIC_dict( vi_errorAndLog_obj, all_interfaces, True )
		nic_state		= interface_DICT.get( givenInterfaceOrMACtoVerify, None )
		if nic_state:
			return nic_state

		return None

	givenInterfaceOrMACtoVerify	= givenInterfaceOrMACtoVerify.upper()	# Deal with lowercase coming from config files
	possibles_LIST				= []

	for interface, nic_state in interface_DICT.iteritems():
		if nic_state.mMAC == givenInterfaceOrMACtoVerify:
			possibles_LIST.append( nic_state )

	if len( possibles_LIST ) == 1:
		return possibles_LIST[ 0 ]

	if len( possibles_LIST ) > 0:	# Find the bond
		for nic_state in possibles_LIST:
			if nic_state.mInterface.startswith( 'bond' ):
				return nic_state

	interface_DICT	= vi_get_NIC_dict( vi_errorAndLog_obj, all_interfaces, True )

	for interface, nic_state in interface_DICT.iteritems():
		if nic_state.mMAC == givenInterfaceOrMACtoVerify:
			possibles_LIST.append( nic_state )

	if len( possibles_LIST ) == 1:
		return possibles_LIST[ 0 ]

	if len( possibles_LIST ) > 0:	# Find the bond
		for nic_state in possibles_LIST:
			if nic_state.mInterface.startswith( 'bond' ):
				return nic_state

	return None

####################################################################################################
def vi_wait_for_NIC_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs ):
	"""
	This routine waits for the given NIC to be both UP and RUNNING
	"""
#
	nicIsUP			= False
	new_nic_state	= nic_state

	for cntr in range( timeoutInSecs ):

		interface_DICT	= vi_get_NIC_dict( vi_errorAndLog_obj, False, True )
		if not interface_DICT:
			vi_errorAndLog_obj.vi_print_log( "All NIC state unavailable :: %s" % ( nic_state ) )
			return errno.ENODEV

		new_nic_state	= interface_DICT[ nic_state.mInterface ]
		if not new_nic_state:
			vi_errorAndLog_obj.vi_print_log( "new NIC state unavailable :: %s" % ( nic_state ) )
			return errno.ENODEV

		if not new_nic_state.mIsUP:
			time.sleep( 1.0 )
			continue

		if not nicIsUP:
			nicIsUP	= True
			vi_errorAndLog_obj.vi_print_log( "NIC %s (%s) is UP" % ( nic_state.mInterface, nic_state.mMAC ) )

		if not new_nic_state.mIsRUNNING:
			time.sleep( 1.0 )
			continue

		vi_errorAndLog_obj.vi_print_log( "NIC %s (%s) is RUNNING" % ( nic_state.mInterface, nic_state.mMAC ) )
		return 0	# Success
	else:
		vi_errorAndLog_obj.vi_print_log( "NIC %s (%s) timed out : UP == %s RUNNING == %s" % ( nic_state.mInterface, nic_state.mMAC, new_nic_state.mIsUP, new_nic_state.mIsRUNNING ) )

	return errno.ETIMEDOUT

####################################################################################################
def vi_reset_NIC_and_wait_for_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs, optional_MTU_sizeInBytes=None ):
	"""
	This routine resets the given NIC and then waits for it to be both UP and RUNNING
	"""
#
	if vi_terminate_dhcp( vi_errorAndLog_obj, nic_state ):
		vi_errorAndLog_obj.vi_print_err( "Could not terminated DHCP : %s", nic_state )
		return errno.EBUSY

	# Reset the NIC IP
	cmd_LIST	= [ 'ifconfig', nic_state.mInterface, 'inet', '0' ]
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )

	# Now bring up the NIC
	if not optional_MTU_sizeInBytes or nic_state.mMTU == optional_MTU_sizeInBytes:
		cmd_LIST	= [ 'ifconfig', nic_state.mInterface, 'up' ]
	else:
		cmd_LIST	= [ 'ifconfig', nic_state.mInterface, 'mtu', str(optional_MTU_sizeInBytes), 'up' ]

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )
	if exitStatus != 0:
		vi_errorAndLog_obj.vi_print_err( "NIC failed to 'up' : %s", nic_state )
		return exitStatus

	return vi_wait_for_NIC_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs )

####################################################################################################
def vi_terminate_dhcp( vi_errorAndLog_obj, nic_state ):
	"""
	This routine terminates any dhcp that is running on the given interface
	"""
#
	ps_process_DICT	= vi_get_ps_process_dict( vi_errorAndLog_obj, True )

	for pid, ps_process in ps_process_DICT.iteritems():
		if not ps_process.mCmd_LIST[0].startswith( 'dhcpcd' ):
			continue

		for arg in ps_process.mCmd_LIST:
			if nic_state.mInterface != arg:
				continue

			vi_errorAndLog_obj.vi_print_log( "Stopping DHCP on %s : pid %d" % ( nic_state.mInterface, ps_process.mPID ) )
			if vi_utils.vi_terminate_process( vi_errorAndLog_obj, ps_process.mPID, True ):
				vi_errorAndLog_obj.vi_print_err( "FAILED to stop DHCP on %s : pid %d" % ( nic_state.mInterface, ps_process.mPID ) )
				return errno.EBUSY

	return 0

####################################################################################################
def vi_invoke_dhcp( vi_errorAndLog_obj, nic_state, timeoutInSecs ):
	"""
	This routine invokes dhcp to run for a given interface
	"""
#
	vi_errorAndLog_obj.vi_print_log( "Resetting interface: %s" % ( nic_state.mInterface ) )

	exitStatus	= vi_reset_NIC_and_wait_for_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs )
	if exitStatus != 0:
		return exitStatus

	vi_errorAndLog_obj.vi_print_log( "Starting DHCP on %s" % ( nic_state.mInterface ) )

	cmd_LIST	= [ 'dhcpcd' ]
	cmd_LIST.extend( shlex.split( DHCPCD_ARGS ) )
	cmd_LIST.append( nic_state.mInterface )

	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( vi_errorAndLog_obj, cmd_LIST, timeoutInSecs )
	if exitStatus != 0:
		vi_terminate_dhcp( vi_errorAndLog_obj, interface )
		return exitStatus

	exitStatus	= vi_wait_for_NIC_RUNNING( vi_errorAndLog_obj, nic_state, timeoutInSecs )
	if exitStatus != 0:
		vi_terminate_dhcp( vi_errorAndLog_obj, interface )
		return exitStatus

	return 0

####################################################################################################
def vi_get_IP_addr_dict( errAndLog_obj, refresh):
	"""
	Retrieves a dictionary of IP addresses available on the machine. The key is the IP address
	(string value), while the value is the list of NIC_state_obj associated with the IP.

	Example:
	'10.1.1.202 -> [NIC_state_obj, NIC_state_obj]
	'fe80::213:72ff:fe27:f4a6/64' -> [NIC_state_obj]

	"""
#
	global global__cached__IP_to_NIC_DICT

	if not refresh and global__cached__IP_to_NIC_DICT:
		return global__cached__IP_to_NIC_DICT

	interfaces_DICT	= vi_get_NIC_dict(errAndLog_obj, False, refresh)
	IP_addr_DICT	= {}
	for interface, NIC_state in interfaces_DICT.iteritems():

		if NIC_state.mIPv4_addr:
			interface_LIST = IP_addr_DICT.get(NIC_state.mIPv4_addr, None)
			if interface_LIST:
				interface_LIST.append(NIC_state)
			else:
				IP_addr_DICT[ NIC_state.mIPv4_addr ] = [NIC_state]

		if NIC_state.mIPv6_addr:
			interface_LIST = IP_addr_DICT.get(NIC_state.mIPv6_addr, None)
			if interface_LIST:
				interface_LIST.append(NIC_state)
			else:
				IP_addr_DICT[ NIC_state.mIPv6_addr ] = [NIC_state]

	global__cached__IP_to_NIC_DICT = IP_addr_DICT
	return global__cached__IP_to_NIC_DICT

####################################################################################################
def vi_get_MAC_addr_dict( errAndLog_obj, all_interfaces, refresh):
	"""
	Retrieves a dictionary of Media Access Control address (MAC) available on the machine.
	The key is the MAC address (string value), while the value is the interface for the MAC.

	Example:
	'00-00-00-00-00-00-00-00-00-00-00-00-00-00-00-00' -> 'tun0'
	'00:15:17:7B:BE:04' -> [eth4, eth5, bond0]
	"""
#
	global global__cached__IP_to_MAC_DICT

	if not refresh and global__cached__IP_to_MAC_DICT:
		return global__cached__IP_to_MAC_DICT

	interfaces_DICT = vi_get_NIC_dict(errAndLog_obj, all_interfaces, refresh)
	MAC_addr_DICT = {}
	for interface, NIC_state_obj in interfaces_DICT.iteritems():
		if NIC_state_obj.mMAC:
			interface_LIST= MAC_addr_DICT.get(NIC_state_obj.mMAC, None)
			if interface_LIST:
				interface_LIST.append(NIC_state_obj)
			else:
				MAC_addr_DICT [ NIC_state_obj.mMAC ] = [NIC_state_obj]

	global__cached__IP_to_MAC_DICT = MAC_addr_DICT
	return global__cached__IP_to_MAC_DICT

###################################################################################################
def vi_get_route_list(errAndLog_obj, refresh):
	"""
	Collect all the IPv4 network routes present on the machine.
	"""
#
	def convert_hex_str_to_big_endian( string_val ):
		"""
		Converts a little endian string value ("0021010A") to a big-endian
		numeric (long) value: A012100
		"""
		convertor = vi_utils.vi_strctFrmt__be__u32()
		long_val = long(string_val,16)
		return convertor.ConvertFromNativeEndianLayout(long_val)

	# IPv4:
	#cat /proc/net/route
	#Iface   Destination     Gateway         Flags   RefCnt  Use     Metric  Mask            MTU     Window  IRTT
	#eth0    058639D8        0164A8C0        0007    0       0       0       FFFFFFFF        1500    0       0  
	#eth12   00000000        01E6010A        0003    0       0       0       00000000        0       0       0  

	#cat /proc/net/ipv6_route
	#00000000000000000000000000000001 80 00000000000000000000000000000000 00 00000000000000000000000000000000 00000000 00000001 00000012 80200001       lo
	#fe800000000000000000000000000000 40 00000000000000000000000000000000 00 00000000000000000000000000000000 00000100 00000000 00000000 00000001     eth0

	global global__cached__IPv4_route_LIST

	if not refresh and global__cached__IPv4_route_LIST:
		return global__cached__IPv4_route_LIST

	# Collect IPv4 routes:
	buf_LIST, read_failed = vi_utils.vi_read_text_lines( errAndLog_obj, '/proc/net/route' )

	if read_failed:
		errAndLog_obj.vi_print_wrn( "Failed to read /proc/net/route!")
		return []

	# Kernel gives the data in little-endian. Must convert to network-endian (big-endian)
	IPv4_parse_route_RegEx = re.compile("^(\S+)\s(\S+)\s(\S+)\s\S+\s\d+\s\d+\s\d+\s(\S+)\s\d+\s\\d+\s\d+\S*$")

	route_LIST = []
	for route in buf_LIST:
		line = route.strip()
		parse_route_match = IPv4_parse_route_RegEx.match(line)
		errAndLog_obj.vi_print_dbg( 5, "F: %s" % ( line ) )
		if parse_route_match:
			route_obj = Network_Route_OBJ()
			route_obj.mInterface = parse_route_match.group(1)
			route_obj.mIPv4_dest = convert_hex_str_to_big_endian(parse_route_match.group(2))
			route_obj.mIPv4_gateway = convert_hex_str_to_big_endian(parse_route_match.group(3))
			route_obj.mIPv4_netmask = convert_hex_str_to_big_endian(parse_route_match.group(4))
			route_LIST.append(route_obj)

	# Collect IPv6 routes...
	#buf_LIST, read_failed = vi_utils.vi_read_text_lines( errAndLog_obj, '/proc/net/ipv6_route' )

	#if read_failed:
	#	errAndLog_obj.vi_print_wrn( "Failed to read /proc/net/route!")
	#else:
	#	IPv6_parse_route_RegEx = re.compile("^(\S+)\s+\S+\s(\S)\s+\S+\s+\S\s+\S+\s\S+\s\S+\s+\S+\s+(\S+)")

	global__cached__IPv4_route_LIST = route_LIST
	return global__cached__IPv4_route_LIST


###################################################################################################
def vi_is_subnet_for_IP_present(errAndLog_obj, refresh, ip):
	"""
	Finds out if there is a subnet specified for the passed in IP.

	Returns:
	True if it is.
	"""
#
	q = ip.split(".")
	if len(q)==1: # It is IPv6..need to make it work.
		return False

	# Possibly get rid of lambda. Think about IPv6..
	ip_in_hex = reduce(lambda a,b: long(a)*256 + long(b), q)

	route_LIST = vi_get_route_list(errAndLog_obj, refresh)
	for route_obj in route_LIST:
		if route_obj.mIPv4_netmask and route_obj.mIPv4_dest:
			# Logic is as follow:
			#  if IP & mask == subnet, then there is a subnet for it
			#  if (192.168.1.100 & 255.255.255.0) == 192.168.1.0
			errAndLog_obj.vi_print_dbg(4, "Interface: %s (%X & %X == %X)?" % (
					route_obj.mInterface
				,	ip_in_hex
				,	route_obj.mIPv4_netmask
				,	route_obj.mIPv4_dest))
			if ((ip_in_hex & route_obj.mIPv4_netmask) == route_obj.mIPv4_dest):
				return True

	return False

####################################################################################################
def vi_is_route_to_IP_present(errAndLog_obj, refresh, ip, NIC_state = None):
	"""
	Check if there is a host or network route for the specified IP. Optionally
	the NIC_state can be supplied to check if it is also on the appropiate interface.

	Returns:
	  True if there is and False otherwise.
	"""
	q = ip.split(".")
	if len(q)==1: # It is IPv6..need to make it work.
		return False

	# Possibly get rid of lambda. Think about IPv6..
	ip_in_hex = reduce(lambda a,b: long(a)*256 + long(b), q)

	route_LIST = vi_get_route_list(errAndLog_obj, refresh)
	for route_obj in route_LIST:
		if ip_in_hex  == route_obj.mIPv4_dest:
			if NIC_state:
				if NIC_state.mInterface != route_obj.mInterface:
					continue

			errAndLog_obj.vi_print_dbg(4,  "Interface: %s has route to %X" % (
					route_obj.mInterface
				,	ip_in_hex))
			return True

	return False

####################################################################################################
def vi_get_route_list_for_IP( errAndLog_obj, given_ip, doNOTincludeDefaultRoute, refresh ):
	"""
	Get the list of network route for the specified IP.

	Returns:
	  List of network routes
	"""
#
	q	= given_ip.split(".")
	if len(q) == 1: # It is IPv6..need to make it work.
		return []

	# Possibly get rid of lambda. Think about IPv6..
	ip_in_hex	= reduce(lambda a,b: long(a)*256 + long(b), q)

	ipRoute_LIST	= []
	route_LIST		= vi_get_route_list(errAndLog_obj, refresh)
	for route_obj in route_LIST:
		masked_ip_in_hex	= ip_in_hex & route_obj.mIPv4_netmask
#		errAndLog_obj.vi_print_dbg( 0, "IP: %s %X => %X R: %s" % ( given_ip, ip_in_hex, masked_ip_in_hex, route_obj ) )
		if masked_ip_in_hex != route_obj.mIPv4_dest:
			continue
		if doNOTincludeDefaultRoute and route_obj.mIPv4_netmask == 0:
			continue
		ipRoute_LIST.append( route_obj )

	return ipRoute_LIST

####################################################################################################
def vi_get_NIC_from( errAndLog_obj, refresh, all_interfaces, NIC_interface, MAC_addr):
	"""
	Returns the NIC interface of the supplied MAC address
	or the NIC interface (in which case it checks to make sure
	the NIC interface is correct).
	"""
#
	interfaces_DICT	= vi_get_NIC_dict(errAndLog_obj, all_interfaces, refresh)
	target_NIC		= None

	if NIC_interface:
		NIC_state_obj =  interfaces_DICT.get(NIC_interface, None)
		if NIC_state_obj:
			target_NIC = NIC_state_obj.mInterface

	if MAC_addr:
		for interface, NIC_state_obj in interfaces_DICT.iteritems():
			if NIC_state_obj.mMAC:
				# TODO: We only deal with IPv4 here...
				if NIC_state_obj.mMAC == MAC_addr and NIC_state_obj.mIPv4_addr:
					target_NIC=NIC_state_obj.mInterface
					break

	return target_NIC

####################################################################################################
def vi_add_route_for_IP(errAndLog_obj, refresh, IP_addr, NIC_interface = None, MAC_addr = None):
	"""
	Adds a host route for the specified IP address using either the MAC address or the
	NIC name. After figuring out what interface to use (if MAC address is supplied), the
	routine adds a route for the specified host as so:

	route add -host <IP_addr> dev <NIC_interface>

	Returns:
	Zero if succesfull
	"""
#
	target_NIC = vi_get_NIC_from(errAndLog_obj, refresh, False, NIC_interface, MAC_addr)
	if not target_NIC:
		return errno.EINVAL

	args = "/sbin/route add -host %s dev %s" % (IP_addr, target_NIC)
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( errAndLog_obj, args)
	if len(stdout_LIST) > 0:
		errAndLog_obj.vi_print_wrn( "".join(stdout_LIST) )
	if len(stderr_LIST) > 0:
		errAndLog_obj.vi_print_wrn( "".join(stderr_LIST) )

	del stdout_LIST
	del stderr_LIST
	del deltaTime

	return exitStatus

####################################################################################################
def vi_del_route_for_IP(errAndLog_obj, refresh, IP_addr, NIC_interface = None, MAC_addr = None):
	"""
	Deletes a host route for the specified IP address using either the MAC address or the
	NIC name. After figuring out what interface to use (if MAC address is supplied), the
	routine adds a route for the specified host as so:

	route del -host <IP_addr> dev <NIC_interface>

	Returns:
	Zero if succesfull
	"""
#
	target_NIC = vi_get_NIC_from(errAndLog_obj, refresh, False, NIC_interface, MAC_addr)
	if not target_NIC:
		return errno.EINVAL

	args = "/sbin/route del -host %s dev %s" % (IP_addr, target_NIC)
	stdout_LIST, stderr_LIST, exitStatus, deltaTime = vi_utils.vi_exe_sync( errAndLog_obj, args)
	if len(stdout_LIST) > 0:
		errAndLog_obj.vi_print_wrn( "".join(stdout_LIST) )
	if len(stderr_LIST) > 0:
		errAndLog_obj.vi_print_wrn( "".join(stderr_LIST) )

	del stdout_LIST
	del stderr_LIST
	del deltaTime

	return exitStatus

####################################################################################################
def vi_ensure_route_for_IP(errAndLog_obj, refresh, IP_addr, MAC_addr ):
	"""
	Ensures that a route to the specified IP address on the specified
	MAC address will be added (or updated if it existed before). If the
	MAC is for a QLogic iSCSI HBA it is ignored.

	Returns:
	Zero on success.
	"""

	rc = errno.ENODEV
	MAC_addr_DICT = vi_get_MAC_addr_dict(errAndLog_obj, False, refresh)
	NIC_state_obj_LIST = MAC_addr_DICT.get(MAC_addr, None)
	if not NIC_state_obj_LIST:
		errAndLog_obj.vi_print_wrn("Cannot set route for %s because the associated MAC address (%s) does not exist!" \
				% (IP_addr, MAC_addr))
		return rc

	route_to_del_LIST = []

	for NIC_state_obj in NIC_state_obj_LIST:
		if vi_is_route_to_IP_present(errAndLog_obj, refresh, IP_addr):
			# Be more specific. Check the _exact_ interface.
			if vi_is_route_to_IP_present(errAndLog_obj, refresh, IP_addr, NIC_state = NIC_state_obj):
				errAndLog_obj.vi_print_log("Route for %s exists already on %s (%s, %s). Skipping." % (IP_addr, NIC_state_obj.mInterface, NIC_state_obj.mMAC, MAC_addr))
				return 0

			# defer deletion of old routes _after_ adding a new route.
			# VIKR: This is IPv4 specific right now.
			route_obj_LIST = vi_get_route_list(errAndLog_obj, refresh)
			for route_obj in route_obj_LIST:
				if route_obj.is_it_this_dest(IP_addr) and route_obj not in route_to_del_LIST:
					route_to_del_LIST.append(route_obj)

	# .. and now add the route.
	for NIC_state_obj in NIC_state_obj_LIST:
		# VIKR, this is IPv4 specific right now
		if not NIC_state_obj.mIPv4_addr and NIC_state_obj.mIPv6_addr:
			errAndLog_obj.vi_print_wrn("Interface %s, MAC address: %s is IPv6, while we are adding a route (to %s) for IPv4!" %
				(NIC_state_obj.mInterface, MAC_addr, IP_addr))
			continue
		rc=vi_add_route_for_IP(errAndLog_obj, refresh, IP_addr, NIC_interface = NIC_state_obj.mInterface)
		if rc==0:
			errAndLog_obj.vi_print_log("Bound %s to interface %s, MAC address: %s." % (IP_addr, NIC_state_obj.mInterface, MAC_addr))
		else:
			errAndLog_obj.vi_print_wrn("Failed to bound %s to interface %s, MAC address: %s\nError: %d:%s!" % \
				(IP_addr, NIC_state_obj.mInterface, MAC_addr, rc, os.strerror(rc)))
			return rc

	# .. and delete the old routes.
	for route_obj in route_to_del_LIST:
		rc=vi_del_route_for_IP(errAndLog_obj, refresh, IP_addr, NIC_interface = route_obj.mInterface)
		if rc==0:
			errAndLog_obj.vi_print_log("Deleting old route for %s on %s." % \
					(IP_addr, route_obj.mInterface))
		else:
			errAndLog_obj.vi_print_wrn("Failed to delete route to %s on %s!\n%d:%s" % \
					(IP_addr, route_obj.mInterface, rc, os.strerror(rc)))
			return rc

	return rc

####################################################################################################
# END
