#!/usr/bin/python
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for dealing with error and debug logging
"""

import sys
import traceback

from vi_misc			import *
from vi_printControl	import *

####################################################################################################

__all__ = [	'VI_ErrAndLog__FOR_FLAG__ERR'
		,	'VI_ErrAndLog__FOR_FLAG__WRN'
		,	'VI_ErrAndLog__FOR_FLAG__E_W'
		,	'VI_ErrAndLog__FOR_FLAG__DBG'
		,	'VI_ErrAndLog__FOR_FLAG__LOG'
		,	'VI_ErrAndLog__FOR_FLAG__STD'
		,	'VI_ErrAndLog__FOR_FLAG__ALT'
		,	'VI_ErrAndLog__FOR_FLAG__ALL'
		,	'VI_ErrAndLog_OBJ'
		,	'VI_ErrAndLogNULL_OBJ'
		,	'VI_ErrAndLogRedirector_OBJ'
		,	'vi_create_std_ErrAndLogRedirector' ]

####################################################################################################

VI_ErrAndLog__FOR_FLAG__ERR	= 0x0000000FL
VI_ErrAndLog__FOR_FLAG__WRN	= 0x000000F0L
VI_ErrAndLog__FOR_FLAG__E_W	= 0x000000FFL
VI_ErrAndLog__FOR_FLAG__DBG	= 0x00000F00L
VI_ErrAndLog__FOR_FLAG__LOG	= 0x0000F000L
VI_ErrAndLog__FOR_FLAG__STD	= 0x000F0000L
VI_ErrAndLog__FOR_FLAG__ALT	= 0x00F00000L
VI_ErrAndLog__FOR_FLAG__ALL	= 0xFFFFFFFFL

####################################################################################################
class VI_ErrAndLog_OBJ:

	"""ABSTRACT class for subclasses:		VI_ErrAndLogNULL_OBJ
											VI_ErrAndLogRedirector_OBJ
	"""

	def set_debugLvl( self, debugLvl ):
		vi_raise_not_implemented()

	def get_debugLvl( self ):
		vi_raise_not_implemented()

	def get_num_of_errors( self ):
		vi_raise_not_implemented()

	def get_num_of_warnings( self ):
		vi_raise_not_implemented()

	def vi_set_on_crash_fcn( self, onCrashFcn, onCrashFcnData ):
		vi_raise_not_implemented()

	def vi_call_on_crash_fcn( self, crashTitleMsg, crashMsg ):
		vi_raise_not_implemented()

	def vi_stop( self, *itemsToPrint_TPL ):
		vi_raise_not_implemented()

	def vi_croak( self, extraMsg ):
		vi_raise_not_implemented()

	def vi_croak_exception( self, exceptionMsg, extraMsg=None ):
		vi_raise_not_implemented()

	def vi_print_str( self, vi_errAndLog_forFlag, strToPrint ):
		vi_raise_not_implemented()

	def vi_print_err( self, *itemsToPrint_TPL ):
		vi_raise_not_implemented()

	def vi_print_wrn( self, *itemsToPrint_TPL ):
		vi_raise_not_implemented()

	def vi_should_print_dbg( self, dbgLvlThreshold ):
		vi_raise_not_implemented()

	def vi_print_dbg( self, dbgLvlThreshold, *itemsToPrint_TPL ):
		vi_raise_not_implemented()

	def vi_print_log( self, *itemsToPrint_TPL ):
		vi_raise_not_implemented()

	def vi_print_std( self, *itemsToPrint_TPL ):
		vi_raise_not_implemented()

	def vi_print_alt( self, *itemsToPrint_TPL ):
		vi_raise_not_implemented()

####################################################################################################
class VI_ErrAndLogNULL_OBJ(VI_ErrAndLog_OBJ):

	"""Class provides a /dev/null stub for APIs that require a VI_ErrAndLog_OBJ when none is
	available.  WARNING:  This should be used very sparingly as errors (ie BUGS! ) will be lost.
	"""

	########################################

	def __init__( self ):
		self.mDbgLvl			= 0
		self.mNumOfErrs			= 0
		self.mNumOfWrns			= 0
		self.mOnCrashFcn		= None
		self.mOnCrashFcnData	= None

	def set_debugLvl( self, debugLvl ):
		self.mDbgLvl	= debugLvl

	def get_debugLvl( self ):
		return self.mDbgLvl

	def get_num_of_errors( self ):
		return self.mNumOfErrs

	def get_num_of_warnings( self ):
		return self.mNumOfWrns

	def vi_set_on_crash_fcn( self, onCrashFcn, onCrashFcnData ):
		self.mOnCrashFcn		= onCrashFcn
		self.mOnCrashFcnData	= onCrashFcnData

	def vi_call_on_crash_fcn( self, crashTitleMsg, crashMsg ):
		if self.mOnCrashFcn:
			return self.mOnCrashFcn( self, self.mOnCrashFcnData, crashTitleMsg, crashMsg )
		else:
			return 1

	def vi_stop( self, *itemsToPrint_TPL ):
		vi_exit( 0 )

	def vi_croak( self, extraMsg ):
		self.vi_call_on_crash_fcn( 'Croak', extraMsg )
		vi_exit( 0 )

	def vi_croak_exception( self, exceptionMsg, extraMsg=None ):
		self.vi_call_on_crash_fcn( 'Exception', exceptionMsg + extraMsg )
		vi_exit( 0 )

	def vi_print_str( self, vi_errAndLog_forFlag, strToPrint ):
		pass

	def vi_print_err( self, *itemsToPrint_TPL ):
		self.mNumOfErrs = self.mNumOfErrs + 1

	def vi_print_wrn( self, *itemsToPrint_TPL ):
		self.mNumOfWrns = self.mNumOfWrns + 1

	def vi_should_print_dbg( self, dbgLvlThreshold ):
		pass

	def vi_print_dbg( self, dbgLvlThreshold, *itemsToPrint_TPL ):
		pass

	def vi_print_log( self, *itemsToPrint_TPL ):
		pass

	def vi_print_std( self, *itemsToPrint_TPL ):
		print ' '.join( itemsToPrint_TPL )

	def vi_print_alt( self, *itemsToPrint_TPL ):
		pass

####################################################################################################
class VI_ErrAndLogRedirector_OBJ(VI_ErrAndLog_OBJ):

	"""Class for redirecting print() output to mutiple 'file's

	Arguments:	dst_DICT	DICT of str -> ( VI_ErrAndLog__FOR_FLAG..., files object or VI_PrintAccumulator_OBJ object )
							to redirect the output to

	Example:

	errAndLog = VI_ErrAndLogRedirector_OBJ( { 'stdErr' : ( VI_ErrAndLog__FOR_FLAG__ERR, sys.stderr ) } )
	errAndLog.vi_print_str( VI_ErrAndLog__FOR_FLAG__ERR, "just a single string" )
	errAndLog.vi_print_err( "whatever args I want to print ", myObjToPrint, 'yeah' )
	"""

	########################################
	def __init__( self, dst_DICT, debugLvl=0 ):

		self.mDbgLvl			= debugLvl
		self.mNumOfErrs			= 0
		self.mNumOfWrns			= 0
		self.mOnCrashFcn		= None
		self.mOnCrashFcnData	= None
		self.mOutLookup			= {}

		errPrntDst = {}
		wrnPrntDst = {}
		dbgPrntDst = {}
		logPrntDst = {}
		stdPrntDst = {}
		altPrntDst = {}
		for dstNameStr, flagDst_TPL in dst_DICT.iteritems():
			if type(flagDst_TPL) != tuple:
				raise ValueError, "given ( flag, dst ) %s is not a tuple !?!" % ( type(flagDst_TPL) )
			if len( flagDst_TPL ) != 2:
				raise ValueError, "given ( flag, dst ) TPL has bad len"
			vi_errAndLog_forFlag	= flagDst_TPL[ 0 ]
			dst						= flagDst_TPL[ 1 ]
			if type(vi_errAndLog_forFlag) != long:
				raise ValueError, "given ( flag, dst ) flag is not a long !?!"

			if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__ERR:
				errPrntDst[ dstNameStr ] = dst
			if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__WRN:
				wrnPrntDst[ dstNameStr ] = dst
			if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__DBG:
				dbgPrntDst[ dstNameStr ] = dst
			if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__LOG:
				logPrntDst[ dstNameStr ] = dst
			if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__STD:
				stdPrntDst[ dstNameStr ] = dst
			if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__ALT:
				altPrntDst[ dstNameStr ] = dst

			self.mOutLookup[ dstNameStr ] = dst

		self.mErrOut	= VI_PrintRedirector_OBJ( errPrntDst )
		self.mWrnOut	= VI_PrintRedirector_OBJ( wrnPrntDst )
		self.mDbgOut	= VI_PrintRedirector_OBJ( dbgPrntDst )
		self.mLogOut	= VI_PrintRedirector_OBJ( logPrntDst )
		self.mStdOut	= VI_PrintRedirector_OBJ( stdPrntDst )
		self.mAltOut	= VI_PrintRedirector_OBJ( altPrntDst )

	########################################
	def __repr__( self ):
		return "dL %s nE %s nW %s ocF %s ocFd %s eO %s wO %s dO %s lO %s sO %s aO %s" % (
						self.mDbgLvl
					,	self.mNumOfErrs
					,	self.mNumOfWrns
					,	self.mOnCrashFcn
					,	self.mOnCrashFcnData
					,	self.mErrOut
					,	self.mWrnOut
					,	self.mDbgOut
					,	self.mLogOut
					,	self.mStdOut
					,	self.mAltOut )

	########################################
	def augment_dst( self, vi_errAndLog_forFlag, dstNameStr, dst ):

		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__ERR:
			return self.mErrOut.augment_dst( dstNameStr, dst )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__WRN:
			return self.mWrnOut.augment_dst( dstNameStr, dst )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__DBG:
			return self.mDbgOut.augment_dst( dstNameStr, dst )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__LOG:
			return self.mLogOut.augment_dst( dstNameStr, dst )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__STD:
			return self.mStdOut.augment_dst( dstNameStr, dst )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__ALT:
			return self.mAltOut.augment_dst( dstNameStr, dst )

		return True

	########################################
	def remove_dst( self, vi_errAndLog_forFlag, dstNameStr ):

		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__ERR:
			return self.mErrOut.remove_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__WRN:
			return self.mWrnOut.remove_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__DBG:
			return self.mDbgOut.remove_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__LOG:
			return self.mLogOut.remove_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__STD:
			return self.mStdOut.remove_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__ALT:
			return self.mAltOut.remove_dst( dstNameStr )

		return True

	########################################
	def get_dst( self, vi_errAndLog_forFlag, dstNameStr ):

		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__ERR:
			return self.mErrOut.get_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__WRN:
			return self.mWrnOut.get_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__DBG:
			return self.mDbgOut.get_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__LOG:
			return self.mLogOut.get_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__STD:
			return self.mStdOut.get_dst( dstNameStr )
		if vi_errAndLog_forFlag == VI_ErrAndLog__FOR_FLAG__ALT:
			return self.mAltOut.get_dst( dstNameStr )

		return None

	########################################
	def close_dst_files( self ):

		self.mErrOut.close_dst_files()
		self.mWrnOut.close_dst_files()
		self.mDbgOut.close_dst_files()
		self.mLogOut.close_dst_files()
		self.mStdOut.close_dst_files()
		self.mAltOut.close_dst_files()

	########################################
	def set_debugLvl( self, debugLvl ):

		self.mDbgLvl	= debugLvl

	########################################
	def get_debugLvl( self ):

		return self.mDbgLvl

	########################################
	def get_num_of_errors( self ):

		return self.mNumOfErrs

	########################################
	def get_num_of_warnings( self ):

		return self.mNumOfWrns

	########################################
	def vi_set_on_crash_fcn( self, onCrashFcn, onCrashFcnData ):

		self.mOnCrashFcn		= onCrashFcn
		self.mOnCrashFcnData	= onCrashFcnData

	########################################
	def vi_call_on_crash_fcn( self, crashTitleMsg, crashMsg ):

		if self.mOnCrashFcn:
			return self.mOnCrashFcn( self, self.mOnCrashFcnData, crashTitleMsg, crashMsg )
		else:
			return 1

	########################################
	def vi_stop( self, *itemsToPrint_TPL ):

		calling_frame	= sys._getframe( 1 )
		caller_name		= calling_frame.f_code.co_name
		caller_lineno	= calling_frame.f_lineno

		outputBuffer	 = "\n\nStopping @ %s() line %s :" % ( caller_name, caller_lineno )
		outputBuffer	+= vi_TPL_to_str( itemsToPrint_TPL )

		self.vi_print_str( VI_ErrAndLog__FOR_FLAG__ERR, outputBuffer + '\n' )

		vi_exit( 0 )

	########################################
	def vi_croak( self, extraMsg ):

		if not extraMsg:
			extraMsg	= ''

		outputBuffer	= "\nERROR:  %d %d\n        %s\nTraceback (most recent call last):\n" % \
							( self.mNumOfErrs, self.mNumOfWrns, extraMsg )

		stackTraceMsg_LIST = traceback.format_stack()
		stackTraceMsg_LIST.pop()
		for stackTraceMsg in stackTraceMsg_LIST:
			outputBuffer = outputBuffer + stackTraceMsg

		self.mNumOfErrs	= self.mNumOfErrs + 1
		self.mErrOut.vi_print_str( outputBuffer )

		exitValue	= self.vi_call_on_crash_fcn( 'Croak', outputBuffer )
		vi_exit( exitValue )

	########################################
	def vi_croak_exception( self, exceptionMsg, extraMsg=None ):

		if not extraMsg:
			extraMsg = ''

		outputBuffer	= "\nERROR:  %d %d (exception)\n        %s\n%s" % \
							( self.mNumOfErrs, self.mNumOfWrns, extraMsg, exceptionMsg )

		self.mNumOfErrs	= self.mNumOfErrs + 1
		self.mErrOut.vi_print_str( outputBuffer )

		exitValue = self.vi_call_on_crash_fcn( 'Exception', outputBuffer )
		vi_exit( exitValue )

	########################################
	def vi_print_str( self, vi_errAndLog_forFlag, strToPrint ):

		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__ERR:
			self.mErrOut.vi_print_str( strToPrint )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__WRN:
			self.mWrnOut.vi_print_str( strToPrint )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__DBG:
			self.mDbgOut.vi_print_str( strToPrint )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__LOG:
			self.mLogOut.vi_print_str( strToPrint )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__STD:
			self.mStdOut.vi_print_str( strToPrint )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__ALT:
			self.mAltOut.vi_print_str( strToPrint )

	########################################
	def vi_add_str( self, vi_errAndLog_forFlag, strToAdd ):

		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__ERR:
			self.mErrOut.vi_add_str( strToAdd )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__WRN:
			self.mWrnOut.vi_add_str( strToAdd )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__DBG:
			self.mDbgOut.vi_add_str( strToAdd )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__LOG:
			self.mLogOut.vi_add_str( strToAdd )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__STD:
			self.mStdOut.vi_add_str( strToAdd )
		if vi_errAndLog_forFlag & VI_ErrAndLog__FOR_FLAG__ALT:
			self.mAltOut.vi_add_str( strToAdd )

	########################################
	def vi_print_err( self, *itemsToPrint_TPL ):

		calling_frame	= sys._getframe( 1 )
		caller_name		= calling_frame.f_code.co_name
		caller_lineno	= calling_frame.f_lineno

		outputBuffer  = "ERROR %d @ %s %s() " % ( self.mNumOfErrs, caller_lineno, caller_name )
		outputBuffer += vi_TPL_to_str( itemsToPrint_TPL )

		self.mNumOfErrs	= self.mNumOfErrs + 1
		self.mErrOut.vi_print_str( outputBuffer )

	########################################
	def vi_print_wrn( self, *itemsToPrint_TPL ):

		calling_frame	= sys._getframe( 1 )
		caller_name		= calling_frame.f_code.co_name
		caller_lineno	= calling_frame.f_lineno

		outputBuffer  = "WARNING %d @ %s %s() " % ( self.mNumOfWrns, caller_lineno, caller_name )
		outputBuffer += vi_TPL_to_str( itemsToPrint_TPL )

		self.mNumOfWrns	= self.mNumOfWrns + 1
		self.mWrnOut.vi_print_str( outputBuffer )

	########################################
	def vi_add_wrn( self, *itemsToPrint_TPL ):

		calling_frame	= sys._getframe( 1 )
		caller_name		= calling_frame.f_code.co_name
		caller_lineno	= calling_frame.f_lineno

		outputBuffer  = "WARNING %d @ %s %s() " % ( self.mNumOfWrns, caller_lineno, caller_name )
		outputBuffer += vi_TPL_to_str( itemsToPrint_TPL )

		self.mNumOfWrns	= self.mNumOfWrns + 1
		self.mWrnOut.vi_add_str( outputBuffer )

	########################################
	def vi_should_print_dbg( self, dbgLvlThreshold ):

		if self.mDbgLvl < int( dbgLvlThreshold ):
			return False
		return True

	########################################
	def vi_print_dbg( self, dbgLvlThreshold, *itemsToPrint_TPL ):

		if self.mDbgLvl < int( dbgLvlThreshold ):
			return

		self.mDbgOut.vi_print_str( vi_TPL_to_str( itemsToPrint_TPL ) )

	########################################
	def get_out_dst( self, dstNameStr ):
		return self.mOutLookup.get( dstNameStr, None )

	########################################
	def vi_print_log( self, *itemsToPrint_TPL ):

		self.mLogOut.vi_print_str( vi_TPL_to_str( itemsToPrint_TPL ) )

	########################################
	def vi_print_std( self, *itemsToPrint_TPL ):

		self.mStdOut.vi_print_str( vi_TPL_to_str( itemsToPrint_TPL ) )

	########################################
	def vi_print_alt( self, *itemsToPrint_TPL ):

		self.mAltOut.vi_print_str( vi_TPL_to_str( itemsToPrint_TPL ) )

####################################################################################################
def vi_create_std_ErrAndLogRedirector( debugLvl=0 ):
	"""
	This routine initializes and complex global variables
	"""
#
	errAndLog_dst_DICT =	{	'stderr'	:	(	VI_ErrAndLog__FOR_FLAG__ERR
												|	VI_ErrAndLog__FOR_FLAG__WRN
												|	VI_ErrAndLog__FOR_FLAG__DBG
												,	sys.stderr
												)
#							,	'log'		:	(	VI_ErrAndLog__FOR_FLAG__LOG
#												,	open( '/dev/null', 'a', 0 )
#												)
							,	'stdout'	:	(	VI_ErrAndLog__FOR_FLAG__STD
												,	sys.stdout
												)
							}

	return VI_ErrAndLogRedirector_OBJ( errAndLog_dst_DICT, debugLvl )

####################################################################################################
# END
