#!/usr/bin/python
####################################################################################################
#
#	Author:		Joshua Nicholas
#
"""Utilities for accumulating and redirecting print output
"""

import os
import re
import sys
import time
import stat

from vi_misc	import *

####################################################################################################

__all__ = [	'VI_PrintAccumulator_OBJ'
		,	'VI_PrintRedirector_OBJ'
		,	'vi_TPL_to_str'
		,	'vi_rototillLogs'
		,	'vi_generate_log_filename'
		,	'vi_create_simple_PrintRedirector' ]

####################################################################################################

global_uniquifier = 0

####################################################################################################
class VI_Print_OBJ:

	"""ABSTRACT class for subclasses:		VI_PrintAccumulator_OBJ
											VI_PrintRedirector_OBJ
	"""

	def vi_print_str( self, strToPrint ):
		vi_raise_not_implemented()

	def vi_print( self, *itemsToPrint_TPL ):
		vi_raise_not_implemented()

####################################################################################################
class VI_PrintAccumulator_OBJ(VI_Print_OBJ):

	"""Class for accumulating print() output into a list

	Arguments:	maxNumOfAccumulations	max number of accumulations before droping from the head
				preserveAtHeadCnt		number of accumulations to preverse at the head of the list

		NOTE:	maxNumOfAccumulations == -1					denotes save everything
				maxNumOfAccumulations == preserveAtHeadCnt	denotes drop new item when full

	Example:
	
	prntAccum = VI_PrintAccumulator_OBJ( 10, 5 )
	prntAccum.vi_print_str( "just a single string" )
	prntAccum.vi_print( "whatever args I want to print ", myObjToPrint, 'yeah' )
	myStr = prntAccum.vi_join( '\n' )
	myCpyOfAccum_LIST = prntAccum.vi_get_accumulation()
	print prntAccum
	prntAccum.clear()
	"""

	########################################
	def __init__( self, maxNumOfAccumulations = -1, preserveAtHeadCnt = -1 ):

		if maxNumOfAccumulations >= 0:
			if preserveAtHeadCnt > maxNumOfAccumulations:
				raise ValueError, "preserveAtHeadCnt %d > %d maxNumOfAccumulations" % \
						( preserveAtHeadCnt, maxNumOfAccumulations )

		self.mPreserveAtHeadCnt		= preserveAtHeadCnt
		self.mMaxNumOfAccumulations	= maxNumOfAccumulations
		self.mAccumulation_LIST		= []

	########################################
	def __repr__( self ):

		return self.vi_join( '' )

	########################################
	def vi_add_str( self, strToPrint ):

		if len( self.mAccumulation_LIST ) + 1 > self.mMaxNumOfAccumulations:
			if self.mPreserveAtHeadCnt == self.mMaxNumOfAccumulations:
				return
			del( self.mAccumulation_LIST[ self.mPreserveAtHeadCnt ] )

		self.mAccumulation_LIST.append( str(strToPrint) )

	########################################
	def vi_add( self, *itemsToPrint_TPL ):

		self.vi_add_str( vi_TPL_to_str( itemsToPrint_TPL ) )

	########################################
	def vi_print_str( self, strToPrint ):

		if len( self.mAccumulation_LIST ) + 1 > self.mMaxNumOfAccumulations:
			if self.mPreserveAtHeadCnt == self.mMaxNumOfAccumulations:
				return
			del( self.mAccumulation_LIST[ self.mPreserveAtHeadCnt ] )

		self.mAccumulation_LIST.append( str(strToPrint) )

	########################################
	def vi_print( self, *itemsToPrint_TPL ):

		self.vi_print_str( vi_TPL_to_str( itemsToPrint_TPL ) )

	########################################
	def vi_join( self, strToJoinWith ):

		return strToJoinWith.join( self.mAccumulation_LIST )

	########################################
	def vi_get_accumulation( self ):

		return self.mAccumulation_LIST[:]

	########################################
	def vi_clear_accumulation( self ):

		self.mAccumulation_LIST = []

####################################################################################################
class VI_PrintRedirector_OBJ(VI_Print_OBJ):

	"""Class for redirecting print() output to mutiple 'file's
	
	Arguments:	dst_DICT	DICT of str -> (files objects or VI_PrintAccumulator_OBJ objects)
							to redirect the output to
	
	Example:
	
	prntRedir = VI_PrintRedirector_OBJ( { 'stdErr' : sys.stderr } )
	prntRedir.vi_print_str( "just a single string" )
	prntRedir.vi_print( "whatever args I want to print ", myObjToPrint, 'yeah' )
	"""

	########################################
	def augment_dst( self, dstNameStr, dst ):

		if not isinstance( dstNameStr, str ):
			return True

		if self.mDst_FILE_DICT.has_key( dstNameStr ):
			del self.mDst_FILE_DICT[ dstNameStr ]

		if self.mDst_ACUM_DICT.has_key( dstNameStr ):
			del self.mDst_FILE_DICT[ dstNameStr ]

		if isinstance( dst, file ):
			self.mDst_FILE_DICT[ dstNameStr ]	= dst
			return False

		if isinstance( dst, VI_PrintAccumulator_OBJ ):
			self.mDst_ACUM_DICT[ dstNameStr ]	= dst
			return False

		return True	# given dst MUST be file instance OR VI_PrintAccumulator_OBJ instance

	########################################
	def remove_dst( self, dstNameStr ):

		if self.mDst_FILE_DICT.has_key( dstNameStr ):
			del self.mDst_FILE_DICT[ dstNameStr ]
			return False

		if self.mDst_ACUM_DICT.has_key( dstNameStr ):
			del self.mDst_FILE_DICT[ dstNameStr ]
			return False

		return True

	########################################
	def get_dst( self, dstNameStr ):

		dst	= self.mDst_FILE_DICT.get( dstNameStr, None )
		if dst:
			return dst

		dst	= self.mDst_ACUM_DICT.get( dstNameStr, None )
		if dst:
			return dst

		return None

	########################################
	def close_dst_files( self ):

		for dstNameStr, dst in self.mDst_FILE_DICT.iteritems():
			if dst != sys.stdout and dst != sys.stderr:
				dst.close()

		del self.mDst_FILE_DICT
		self.mDst_FILE_DICT = {}

	########################################
	def __init__( self, dst_DICT ):

		self.mDst_FILE_DICT	= {}
		self.mDst_ACUM_DICT	= {}

		for dstNameStr, dst in dst_DICT.iteritems():
			if not isinstance( dstNameStr, str ):
				raise ValueError, "given dst key MUST be a string"

			if self.augment_dst( dstNameStr, dst ):
				raise ValueError, "given dst MUST be file instance OR VI_PrintAccumulator_OBJ instance"

	########################################
	def __repr__( self ):

		return "dst: F: %s %s A: %s %s" % (	len( self.mDst_FILE_DICT )
										,	' '.join( self.mDst_FILE_DICT.iterkeys() )
										,	len( self.mDst_ACUM_DICT )
										,	' '.join( self.mDst_ACUM_DICT.iterkeys() ) )


	########################################
	def vi_print_str( self, strToPrint ):

		for dst in self.mDst_FILE_DICT.itervalues():
			try:
				print >> dst, str(strToPrint),
				dst.flush()
			except:
				pass
             
		for dst in self.mDst_ACUM_DICT.itervalues():
			dst.vi_print_str( strToPrint )

	########################################
	def vi_print( self, *itemsToPrint_TPL ):

		self.vi_print_str( vi_TPL_to_str( itemsToPrint_TPL ) )

	########################################
	def vi_add_str( self, strToAdd ):

		for dst in self.mDst_ACUM_DICT.itervalues():
			dst.vi_add_str( strToAdd )

	########################################
	def vi_add( self, *itemsToAdd_TPL ):

		self.vi_add_str( vi_TPL_to_str( itemsToAdd_TPL ) )

####################################################################################################
def vi_TPL_to_str( items_TPL ):
	"""
	This routine converts a TPL of items into a single str

	NOTE:	IF the last item in the TPL is    None
			THEN the newline is omitted
	"""
#
	lstIdx_ex = len( items_TPL )
	if lstIdx_ex > 0:

		if items_TPL[lstIdx_ex-1] != None:
			appendNewline = True
		else:
			if lstIdx_ex == 1:
				return ''

			appendNewline = False
			lstIdx_ex = lstIdx_ex - 1

		strToPrint	= str( items_TPL[0] )
		curIdx		= 1
		while curIdx < lstIdx_ex:
			strToPrint	= strToPrint + ' ' + str( items_TPL[curIdx] )
			curIdx		= curIdx + 1

		if appendNewline:
			return strToPrint + '\n'

		return strToPrint

	return '\n'

####################################################################################################
def _INTERNAL__rototill_logFile_TPL__compare( logFile_TPL__A, logFile_TPL__B ):
	"""
	This routine helps vi_rototillLogs() sort log files
	"""
#
	if logFile_TPL__A[0] > logFile_TPL__B[0]:
		return 1
	if logFile_TPL__A[0] < logFile_TPL__B[0]:
		return -1
	return 0

####################################################################################################
def vi_rototillLogs( logDirectory, baseLogFileName, numToPreserveAtHead, numToPreserveAtTail ):
	"""
	This routine examines the log files in the given logDirectory with the given baseLogFileName
	and unlinks based on the last modification time - preserving those based on 

	Arguments:	numToPreserveAtHead		Number to preserve regardless of age at the head
				numToPreserveAtTail		Number to preserve regardless of age at the tail

	Return:		Number of logs unlinked or negative the number of logs that could NOT be unlinked.
	"""
#
	if not os.path.exists( logDirectory ):
		return

	if numToPreserveAtHead < 0:
		numToPreserveAtHead	= 0
	if numToPreserveAtTail < 0:
		numToPreserveAtTail	= 0

	logfileFilter_cmpldRegEx	= re.compile( "^%s" % ( baseLogFileName ) )

	logFile_TPL_LIST = []
	for potentialLogFile in os.listdir( logDirectory ):
		if logfileFilter_cmpldRegEx.match( potentialLogFile ):
			logFilePath				= "%s/%s" % ( logDirectory, potentialLogFile )
			lastModificationTime	= os.stat( logFilePath )[stat.ST_MTIME]
			logFile_TPL_LIST.append( ( lastModificationTime, logFilePath ) )

	if len( logFile_TPL_LIST ) < ( numToPreserveAtHead + numToPreserveAtTail ):
		return 0

	logFile_TPL_LIST.sort( _INTERNAL__rototill_logFile_TPL__compare )

	numFailedToUnlink	= 0
	curIdx				= numToPreserveAtHead
	lstIdx				= len( logFile_TPL_LIST ) - numToPreserveAtTail
	while ( curIdx < lstIdx ):
		try:
			while ( curIdx < lstIdx ):
				logFilePath = logFile_TPL_LIST[curIdx][1]
				curIdx = curIdx + 1		# Increment now in case of exception later
				os.unlink( logFilePath )
		except:
			numFailedToUnlink = numFailedToUnlink - 1
			pass

	if numFailedToUnlink:
		return numFailedToUnlink
	if lstIdx > 0:
		return lstIdx
	return 0

####################################################################################################
def vi_generate_log_filename( baseLogFileName ):
	"""
	This routine returns a unique log filename

	Arguments:	baseLogFileName		base for log filename

	Return:		Unique log filename
	"""
#
	global global_uniquifier

	uniqueNum			= global_uniquifier
	global_uniquifier	= global_uniquifier + 1
	uniqueStr			= time.strftime( "%Y%m%d%H%M%S" )
	return "%s__%d_%s" % ( baseLogFileName, uniqueNum, uniqueStr )

####################################################################################################
def vi_create_simple_PrintRedirector( logFilePath, appendVStruncate ):
	"""
	This routine returns a logFilePath for the given logFilePath

	Arguments:	logFilePath		given log file path

	Return:		logFilePath
	"""
#
	if appendVStruncate:
		open_mode = 'a'
	else:
		open_mode = 'w'

	newLog_FILE = None
	try:
		newLog_FILE = open( logFilePath, open_mode, 0 )		# UNbuffered
	except:
		return None

	return VI_PrintRedirector_OBJ( { 'log' : newLog_FILE } )

####################################################################################################
# END
