#!/usr/bin/python
#####################################################################################################
#
#	Author:		Konrad Rzeszutek
#
"""
	Utilities for parsing the iscsi_portal_list.xml file. The format of such file is
	as follow:
	<?xml version="1.0" encoding="UTF-8"?>
	<ISCSIportalList>
	  	<ISCSIportal>10.1.32.230:3260</ISCSIportal>

		<BoundISCSIportal>
			<Interface>00:15:17:7B:BE:04</Interface>
			<Portal>10.1.33.245:3260</Portal>
		</BoundISCSIportal>

		....
	  </ISCSIportalList>

	There are two types of data structures extracted from this XML when using this library:
	 a) an list of <IP>:<Port> strings, and
	 b) a dictionary, wherein the key is the MAC address and the associated value
	    is the list of the portal strings (["<IP>:<Port>", "<IP>:<Port>", ...])

"""

import vi_utils

import os
import traceback

####################################################################################################

__all__ = [	'vi_iscsi_get_portal_list'
		,'vi_iscsi_get_portal_list_assoc_to_MAC_addr']

####################################################################################################

global__cached__iscsi_portal_LIST					= None
global__cached__iscsi_bounded_portal_DICT				= None
global__cached__iscsi_portal_xml_file					= None

####################################################################################################
def _parse_XML_file(vi_errAndLog_obj, portal_xml_file):
	"""
	Parse the XML file and return two data types:
	 <LIST> - list of portals of format ["<IP>:<Port>", "<IP>:<Port>", ... ]
	 <DICT> - dictionary of key <MAC> with the value being a list of portal strings in the
		  format of ["<IP>:<Port>", "<IP>:<Port>", .. ]

	The return data types can also be empty.
	"""
	class dataParser:
		mPortal_LIST		= []
		mBound_to_MAC_addr_DICT	= {}
		UNDEFINED_LIST_SEL 	= 0
		PORTAL_LIST_SEL		= 1
		INTERFACE_SEL		= 2
		PORTAL_SEL		= 3
		mSelector		= UNDEFINED_LIST_SEL
		mTempInterface		= None
		def __init__(self):
			pass

		def  _handle_start(self, name, attrs):
			if name == 'ISCSIportal':
				self.mSelector = self.PORTAL_LIST_SEL
			if name == 'Interface':
				self.mSelector = self.INTERFACE_SEL
			if name == 'Portal':
				self.mSelector = self.PORTAL_SEL
	    		#print 'START: %s (sel: %d)' % ( repr(name), self.mSelector)

		def _handle_end(self, name):
			pass

		def _handle_data(self, data):
			if str(data).isspace():
				return

	    		#print 'DATA: %s (sel: %d)' % ( repr(data), self.mSelector)

			if self.mSelector == self.PORTAL_LIST_SEL:
				self.mPortal_LIST.append(str(data))
			if self.mSelector == self.INTERFACE_SEL:
				self.mTempInterface = str(data)
			if self.mSelector == self.PORTAL_SEL and self.mTempInterface:
				#print "DICT: in %s for %s" % (data, self.mTempInterface)
				portal_LIST = self.mBound_to_MAC_addr_DICT.get(self.mTempInterface, None)
				if portal_LIST:
					# Additional bounded portals for the MAC address.
					portal_LIST.append( str(data) )
				else:
					self.mBound_to_MAC_addr_DICT [ self.mTempInterface ] = [str(data)]
				self.mTempInterface = None

	if not os.path.exists( portal_xml_file):
		vi_errAndLog_obj.vi_print_wrn( "XML Portal file: '%s' does not exist" % (portal_xml_file))
		return [],{}

	buf, read_failed = vi_utils.vi_read_text_value(vi_errAndLog_obj, portal_xml_file)
	if read_failed:
		vi_errAndLog_obj.vi_print_wrn( "XML Portal file: '%s' could not read" % (portal_xml_file))
		return [],{}

	d = dataParser()
	ExpatParser = None
	try:
		import xml.parsers.expat
		if not hasattr(xml.parsers.expat, "ParserCreate"):
			raise exception
		#NOTE: To use the older parser, just raise an exception..
		#raise exception
	except:
		ExpatParser = None
	else:
		ExpatParser = True
		p = xml.parsers.expat.ParserCreate()
		p.StartElementHandler = d._handle_start
		p.EndElementHandler = d._handle_end
		p.CharacterDataHandler = d._handle_data
		try:
			p.Parse(buf, 0)
		except:
			# This can happen if we have garbage at the end of the file.
			#vi_errAndLog_obj.vi_print_err( "Problems extracting XML file: %s\n%s" % ( portal_xml_file , traceback.format_exc()) )
			pass
		try:
			p.Parse("", 1) # EOF
		except:
			pass
		del p # circular references...

	if not ExpatParser:
		class MyParser:
			def __init__(self, d):
				import xmllib
				if xmllib.XMLParser not in MyParser.__bases__:
					MyParser.__bases__ = (xmllib.XMLParser,)
				self.unknown_starttag = d._handle_start
				self.handle_data = d._handle_data
				self.handle_cdata = d._handle_data
				self.unknown_endtag = d._handle_end
				try:
					xmllib.XMLParser.__init__(self, accept_utf8=1)
				except:
					pass

		p = MyParser(d)
		try:
			p.feed(buf)
		except:
			pass

		try:
			p.close()
		except:
			pass
		del p

	return d.mPortal_LIST, d.mBound_to_MAC_addr_DICT

####################################################################################################
def vi_iscsi_get_portal_list( vi_errAndLog_obj, portal_xml_file, refresh = False):
	"""
	Gets the list of iSCSI portals. The list is of string values in the format of
	"<IP>:<TCP Port>".

	Arguments:
	 portal_xml_file is the absolute path to the XML file
	 refresh specifies whether to use a cache copy of data or refresh from scratch. However
	 if the portal_xml_file specified is different from a previous run the cache data will be
	 evicted automatically.

	Returns:
	 Either a list of "<IP>:<TCP>" strings or an empty list.

	"""
	global global__cached__iscsi_portal_xml_file
	global global__cached__iscsi_portal_LIST
	global global__cached__iscsi_bounded_portal_DICT

	if portal_xml_file is global__cached__iscsi_portal_xml_file and not refresh and global__cached__iscsi_portal_LIST:
		return global__cached__iscsi_portal_LIST

	portal_LIST, bound_DICT = _parse_XML_file(vi_errAndLog_obj, portal_xml_file)

	global__cached__iscsi_portal_xml_file = portal_xml_file
	global__cached__iscsi_portal_LIST = portal_LIST
	global__cached__iscsi_bounded_portal_DICT = bound_DICT
	return global__cached__iscsi_portal_LIST

####################################################################################################
def vi_iscsi_get_portal_list_assoc_to_MAC_addr(vi_errAndLog_obj, portal_xml_file, refresh = False):
	"""
	Gets the dictionary of iSCSI portals associated with a specific MAC address.
	The dictionay key is the string value of type  "<MAC>" and the value a list of
	strings in the format of "<IP>:<TCP Port>".

	Arguments:
	 portal_xml_file is the absolute path to the XML file
	 refresh specifies whether to use a cache copy of data or refresh from scratch. However
	 if the portal_xml_file specified is different from a previous run the cache data will be
	 evicted automatically.

	Returns:
	 Either a dictionary with the <MAC> key and value a list of "<IP>:<TCP>" strings or an empty dictionary.
	"""

	global global__cached__iscsi_portal_xml_file
	global global__cached__iscsi_portal_LIST
	global global__cached__iscsi_bounded_portal_DICT

	if portal_xml_file is global__cached__iscsi_portal_xml_file and not refresh and global__cached__iscsi_bounded_portal_DICT:
		return global__cached__iscsi_bounded_portal_DICT

	portal_LIST, bound_DICT = _parse_XML_file(vi_errAndLog_obj, portal_xml_file)

	global__cached__iscsi_portal_xml_file = portal_xml_file
	global__cached__iscsi_portal_LIST = portal_LIST
	global__cached__iscsi_bounded_portal_DICT = bound_DICT
	return global__cached__iscsi_bounded_portal_DICT


